#ifndef ENCRYPT_DECRYPT_PACKET_H_INCLUDED
#define ENCRYPT_DECRYPT_PACKET_H_INCLUDED
#include "random.h"

static const u32 max_data_length = 32768; // Data field can't be longer then this
static const u64 max_skip_counter = 65536; // If the counter jumped forward more then this, then packet is invalid


// Packet structure

static const u32 packet_length_offset   = 0; // u32
static const u32 packet_counter_offset  = 4 + packet_length_offset;  // u64
static const u32 packet_data_offset     = 8 + packet_counter_offset; // u8[length]
//static const u32 packet_mic_offset      = u32 at the end

static const u32 packet_overhead_bytes = 16; // u64 counter  + u32 length + u32 MIC

// Encrypted packet length = data length + 4 (length field) + 8(counter) + 4(MIC) = len + 16


class ENCRYPT_DECRYPT_PACKET
{
KEYSTREAM_GENERATOR keystream; // keystream generator 

u32 mic_iv[2];  // 64 bit IV for mic
u32 mic_key[4]; // 128-bit key for mic

u32 max_data_len;  
u32 compute_mic(u8 const *packet);
  
public:
     ENCRYPT_DECRYPT_PACKET(u32 packet_max_data_length_bytes = max_data_length);
     ~ENCRYPT_DECRYPT_PACKET();
     void reset();
     void init(u32 const *key_data); // Initialize encryption/decryption with u32[8] key (256 bits)
     u32 encrypt(u8 *dest, u8 const *src, const u32 src_len_bytes);  // Packet encryption, returns length of encrypted block
     u32 decrypt(u8 *dest, u8 const *src, const u32 src_len_bytes);  // Packet decryption, returns length of decrypted block (length -> 0 = invalid block)
     u64 total_rx_bytes();
     u64 dropped_rx_bytes();

                
};

#endif

