#include "types.h"
#include "fast_modulo.h"
#include "large_math.h"
#include "fast_power_2.h"
#include <stdio.h>
//#include "main.h"

// a = (2^ea) mod P 
// b = (2^eb) mod P
// result = a b mod P
//  
u32 test_mod_mul_2(u32 a, u32 eb, u32 p)
{
u64 tmp = a;


 
while(eb--)
 {
 tmp <<= 1;
 if(tmp >= p) tmp -= p; 
 }
 

return (u32)tmp;
}

u32 test_fast_pow_2(u32 x, u32 p)
{
u32 result;


//this works


u32 tmp_pow = 1;
u32 xx = x;
result = 1;


const u32 pow_2[4] = { 1, 2, 4, 8 }; // 2^x

for(u32 ci = 16; ci > 0; ci--)
 {
 if(tmp_pow != 1)
  {   
  for(u32 cj = 0; cj < 2; cj++) tmp_pow = (u32)((tmp_pow*(u64)tmp_pow)%p);  
  }

 u32 j = 0x03 & (xx >> 30);
 xx <<= 2;

 if(j != 0) tmp_pow = (u32)((tmp_pow*(u64)pow_2[j])%p);   
 }

result = tmp_pow;



//----------------------------
// Can't understand this
/*
const u32 pow_2[2] = { 2, 8 }; // 2 in odd orders (1, 3) 
u32 result;
u32 xx = x;
u32 y = 16;

for(s32 i = 16; i > 0; i--)
 { 
 
 u32 jj = 0x03  & (xx >> 30); // 2 upper bits of exponent
 xx <<= 2;  
 u32 j = jj >> 1; // odd exponents only
 u32 e1 = 

 for(s32 j=0; j < e[i+l]-e[i]; j++) y = (u32)((y*(u64)y)%p);

 
 y = (u32)((y*(u64)pow_2[j])%p);
 } 

for(u32 j=0; j < e[0]; j++) y = (u32)((y*(u64)y)%p); 
result = y;
*/


//-- dumb algorithm -

u32 direct = 1;
for(u32 ci = 0; ci < x; ci++) 
 {
 direct = (2*direct)%p;
 }

printf("\n result = %u direct = %u", result, (u32)direct);

return result;
}

//----------------------------------------
// Fast calculation of (2^x) mod P
// (About 30% faster then simple square - multiply Montgomery algorithm
//

FAST_POW_MOD_2::FAST_POW_MOD_2(u32 length_x32)
{
len = length_x32;
monpro = new MONPRO(len);

tmp0 = new u32[len+1];
tmp1 = new u32[len+1];
tmp2 = new u32[len+1];
tmp3 = new u32[len+1];

for(u32 ci = 0; ci < 256; ci++) precalculated_x[ci] = new u32[len];
}

FAST_POW_MOD_2::~FAST_POW_MOD_2()
{
for(u32 ci = 256; ci > 0; ci--) delete[] precalculated_x[ci-1];

delete[] tmp3;
delete[] tmp2;
delete[] tmp1;
delete[] tmp0;

delete monpro;
}
//-------------------------------------    
// calculate 2^x mod P
// Function processes 8 bit of exponent at time, this accelerates it by ~30%
// compared to simple Montgomery square - multiply algorithm
//
void FAST_POW_MOD_2::fast_pow_mod_2(u32 *result, u32 const* x, u32 const *p)
{

monpro->monpro_init(p);

// Montgomerize powers of 2 from 0 to 255

zero(tmp0, len);
tmp0[len] = 1;   // r = 2^n

copy(tmp1, p, len);
tmp1[len] = 0;

multi_div(tmp3, tmp2, tmp0, tmp1, len+1);

// Initialize 2^0 == 1

copy(precalculated_x[0], tmp2, len);

// Also initialize result to 1

copy(result, tmp2, len);

// Compute 2^x mod P, x = 1...255

copy(tmp1, p, len);
tmp1[len] = 0;

for(u32 ci = 1; ci < 256; ci++)
 {
 multi_shift_left(tmp2, len + 1); // multiply by 2

 if(!multi_sub(tmp0, tmp2, tmp1, len + 1)) copy(tmp2, tmp0, len + 1); // modulo P
 
 copy(precalculated_x[ci], tmp2, len); 
 }

//---------------------------------------------------------
// Square - multiply loop does 8 bits of exponent at time
// this boosts the efficiency by ~ 30%
//
for(u32 ci = (len*32); ci > 0; ci -= 8) // 8 bit steps
 {
 for(u32 cj = 0; cj < 8; cj++) // squaring loop
  { 
  monpro->monpro(result, result, result);
  }
 
 u32 i = ci - 1;   
 u32 idx = (x[i>>5] >> (i & 0x18)) & 0xFF; // 8 bit
 
 monpro->monpro(result, result, precalculated_x[idx]);
 }

// Covert result back to numbers 

monpro->demontgomerize(result);
}


//==================== <END> =======================
