// Sofie German precalculated primes 

#include "types.h"
#include <stdio.h>
#include "large_math.h"
#include "simple_test.h"
#include "primes.h"
#include "random.h"

// List

typedef struct 
{
void * next_prime_found;
u32 *prime;    
} PRIME_FOUND;



// Fetch random prime from predefined LUT

void get_p(u32 *p)
{
u32 rnd[2]; 
cheap_random_number(rnd, 2);

u64 r64 = rnd[0] | (((u64)rnd[1]) << 32);
u32 n = (u32)(r64%number_of_primes_4096());    

copy(p, prime_4096_ptr(n), 128); 
}

//---------------------------------------------
// Verify new primes against known primes
//
u32 add_verify_prime_file(char const* prime_file_dest, char const* prime_file_src)
{
const u32 miller_rabin_rounds = 64; // Paranoid     
const u32 prime_len_x32 = 128;    
const u32 prime_len_bytes = 4*prime_len_x32; // 4096 bit

u32 p[prime_len_x32];
u32 q[prime_len_x32];
    

FILE *src = fopen(prime_file_src, "rb");
if(src == NULL) return 0;


MILLER_RABIN_MEMORY *mrm = Miller_Rabin_Init(prime_len_x32);

fseek(src, 0, SEEK_END);
u64 file_len = ftell(src);
fseek(src, 0, SEEK_SET);

u32 N = 0;
u32 M = 0;

PRIME_FOUND *first_prime_found_ptr = NULL;
PRIME_FOUND *prev_prime_found_ptr = NULL;


for(u64 pos = 0; pos < file_len; pos += prime_len_bytes)
 {
 N++;   
 printf("\n Prime: %u", N);
 fread(p, prime_len_bytes, 1, src);

 int found = 0;
 for(u32 ci = 0; ci < number_of_primes_4096(); ci++)
  {
  if(0 == multi_compare(p,  prime_4096_ptr(ci) , prime_len_x32))
   {
   found = 1;
   break;
   }
  }

 if(!found)
  {
  PRIME_FOUND *ptr = first_prime_found_ptr;

  while(ptr != NULL)
   {
   if(0 == multi_compare(p, ptr->prime, prime_len_x32))
    {
    found = 1;
    break;
    }
   
   ptr = (PRIME_FOUND *)ptr->next_prime_found;
   }
  }

  
 if(found)
  {
  printf(" Already Known");
  continue;
  }

 // Size/oddness check

 if(!(p[prime_len_x32-1] & 0x80000000))
  {
  printf(" Not full size");
  continue;
  }

 if(0x03 != (p[0] & 0x03))
  {
  printf(" Not a safe prime");
  continue;
  }


 // Run Miller Rabin test on P and Q
 
 copy(q, p, prime_len_x32);
 multi_shift_right(q, prime_len_x32);
 
 int ret_value = 1; 
 
 printf("\n");
 for(u32 ci = 0; ci < miller_rabin_rounds; ci++)
  {
  printf(" Miller Rabin Test Round: %u %c", ci + 1, 0x0d);
  
  ret_value = MRT_core(p, mrm, prime_len_x32);  // random witness 
  if(!ret_value) break;

  ret_value = MRT_core(q, mrm, prime_len_x32);  // random witness
  if(!ret_value) break;  
  }
 
 if(!ret_value)
  {
  printf(" Failed MRT");
  continue;
  } 

 M++;
 printf(" OK (M = %u)", M);

 
 PRIME_FOUND *prime_found = new PRIME_FOUND;
 prime_found->prime = new u32[prime_len_x32];
 prime_found->next_prime_found = NULL;
 copy(prime_found->prime, p, prime_len_x32);

 if(first_prime_found_ptr == NULL) first_prime_found_ptr = prime_found;
 else prev_prime_found_ptr->next_prime_found = prime_found; 
  
 prev_prime_found_ptr = prime_found;
 }


Miller_Rabin_UnInit(mrm);    
fclose(src);


if(first_prime_found_ptr == NULL) return 0;

prev_prime_found_ptr = first_prime_found_ptr;

//--------------------
FILE *dest = fopen(prime_file_dest,"wt");

fprintf(dest, "// This is generated file. Do not edit.\n");
fprintf(dest, "// It contains precalculated %u bit safe primes.\n\n", prime_len_x32 *32);

fprintf(dest, "#include \"types.h\" \n\n");
fprintf(dest, "static const u32 prime%u[%u] =  \n",prime_len_x32*32, prime_len_x32*(M + number_of_primes_4096()));
fprintf(dest, "{");

for(u32 cj = 0; cj < number_of_primes_4096(); cj++)
 {
 for(u32 ci = 0; ci < prime_len_x32; ci++)
  {
  if(!(ci & 0x07)) fprintf(dest, "\n");
  fprintf(dest, "0x%08x,", prime_4096_ptr(cj)[ci]);
  }
 }

for(u32 cj = 0; cj < M; cj++)
 {
 u32  *ptr = prev_prime_found_ptr->prime; 

 for(u32 ci = 0; ci < prime_len_x32; ci++)
  {
  if(!(ci & 0x07)) fprintf(dest, "\n");
  
  if((cj == (M-1)) && (ci == (prime_len_x32 - 1))) fprintf(dest, "0x%08x", ptr[ci]);
  else fprintf(dest, "0x%08x,", ptr[ci]);
  }
 
 PRIME_FOUND *tmp_ptr = (PRIME_FOUND *)prev_prime_found_ptr->next_prime_found;
 delete[] prev_prime_found_ptr->prime;
 delete prev_prime_found_ptr;
 prev_prime_found_ptr = tmp_ptr;
 }


fprintf(dest, "\n};\n\n");

fprintf(dest, "u32 number_of_primes_4096(void)\n{\n");
fprintf(dest, "return %u;\n}\n\n",M + number_of_primes_4096());

fprintf(dest, "u32 const * prime_4096_ptr(u32 n)\n{\n");
fprintf(dest, "return prime4096 + n*128; \n}\n");

fclose(dest);


//--------------------

return M;
}
//===================== <END> ===============
