#ifndef CONTROL_H
#define CONTROL_H

#include "network.h"

typedef struct MESSAGE_STR{
    TCHAR* text;
    u32 msg_num;
    u32 send_cnt;
	LONG min_pos;
	LONG max_pos;
    struct MESSAGE_STR* next;
} MESSAGE;


enum {
    CTRL_ST_IDLE,
    CTRL_ST_REJECTED,
    CTRL_ST_CONNECTION,
    CTRL_ST_ABORT,
    CTRL_ST_CONNECTED,
    CTRL_ST_DIALING,
    CTRL_ST_IN_VOICE,
    CTRL_ST_LAST
};
//----------------------------------------------------------------
// 0 is reserved for connection request
static const u32 CMD_START_CONNECTION = 1;
static const u32 CMD_CONNECTED = 2;
static const u32 CMD_SHOW_MESSAGE = 3;
static const u32 CMD_MESSAGE_RECEIVED = 4;
static const u32 CMD_HANG_UP = 5;
static const u32 CMD_HANG_DOWN = 6;
static const u32 CMD_IN_VOICE = 7;
static const u32 CMD_VOICE = 8;
static const u32 CMD_TYPING = 9;
static const u32 CMD_REJECT = 10;

static const u32 CONTR_MAX_VOICE_BUF = 14;

class CONTR_PROT {
	CRITICAL_SECTION cs;
	HWND mainWnd;
	HWND parent;
    u32 state;
    MESSAGE* msg;
    u32 send_msg_num;
    u32 rec_msg_num;
    u32 rec_msg_flag;
    u32 voice_cnt;
	u32* send_voice_buf;
    u8* rec_data[CONTR_MAX_VOICE_BUF];
    u32 rec_size[CONTR_MAX_VOICE_BUF];
	bool voice_buf_filled;
	u32 total_pack_num;
	u32 voice_pack_num;
	u32 voice_pack_proc;
	DWORD type_ticks;

    void processMessages(void);

public:
    CONTR_PROT(NODE* node, HWND mWnd, HWND parent);
    ~CONTR_PROT();
    NODE nd;
    void sendMessage(TCHAR* message,LONG len,LONG posMin, LONG posMax);
    void sendVoice(u8* data, u32 size);
    u8* getVoice(u32* size);
    void dial(void);
    void hang_down(void);
    void accept(void);
    void reject(void);
    void process(void);
    void received(u32* packet, u32 size);
    u32 getState(void) {return state;}
    bool isPendingMessages(void) {return msg != NULL;}
	bool buf_is_filled() {return voice_buf_filled;};
	void sendTyping(void);
};

#endif // CONTROL_H
