#pragma once
//---------------------------------------------------------------------------
#if !defined(_WIN32) && !defined(_LINUX)
#error "OS not specified!"
#endif
//---------------------------------------------------------------------------
#include "types.h"
#include <stdio.h>
#include <time.h>
//---------------------------------------------------------------------------
#ifdef _LINUX
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#define FO_MODE_R			"r"
#define FO_MODE_W			"w"
#define FO_MODE_RW			"rw"
#define FO_MODE_A			"a"
#define FO_MODE_RB			"rb"
#define FO_MODE_WB			"wb"
#define FO_MODE_RWB			"rwb"
#define FO_MODE_AB			"ab"

#define FO_RDONLY			O_RDONLY
#define FO_WRONLY			O_WRONLY
#define FO_RDWR				O_RDWR
#define FO_CREAT			O_CREAT
#define FO_TRUNC			O_TRUNC

#define cp_fopen(f,n,m)		(*f = fopen(n,m))
#define cp_fgets			fgets
#define cp_fputs			fputs
#define cp_fscanf			fscanf
#define cp_fprintf			fprintf

#define cp_getwd			getcwd
#define cp_setwd			chdir
#define cp_mkdir(p)			mkdir(p,0777)

typedef int					fd_t;
#define INVALID_FD			(-1)

#elif defined(_WIN32)

#include <tchar.h>
#include <windows.h>

#ifdef _UNICODE
#define FO_MODE_R			_T("r, ccs=UTF-8")
#define FO_MODE_W			_T("w, ccs=UTF-8")
#define FO_MODE_RW			_T("rw, ccs=UTF-8")
#define FO_MODE_A			_T("a, ccs=UTF-8")
#else
#define FO_MODE_R			_T("r")
#define FO_MODE_W			_T("w")
#define FO_MODE_RW			_T("rw")
#define FO_MODE_A			_T("a")
#endif
#define FO_MODE_RB			_T("rb")
#define FO_MODE_WB			_T("wb")
#define FO_MODE_RWB			_T("rwb")
#define FO_MODE_AB			_T("ab")

#define FO_RDONLY			(1<<0)
#define FO_WRONLY			(1<<1)
#define FO_RDWR				(FO_RDONLY|FO_WRONLY)
#define FO_CREAT			(1<<2)
#define FO_TRUNC			(1<<3)

#define cp_fopen			!_tfopen_s
#define cp_fgets			_fgetts
#define cp_fputs			_fputts
#define cp_fscanf			_ftscanf
#define cp_fprintf			_ftprintf

#define cp_getwd(dir,sz)	GetCurrentDirectory(sz,dir)
#define cp_setwd			SetCurrentDirectory
#define cp_mkdir(p)			CreateDirectory(p,0)

typedef HANDLE				fd_t;
#define INVALID_FD			(HANDLE)(-1)

#endif
//---------------------------------------------------------------------------
#define cp_fread			fread
#define cp_fwrite			fwrite
#define cp_feof				feof
#define cp_fseek			fseek
#define cp_fclose			fclose
//---------------------------------------------------------------------------
#define FS_BEGIN			0
#define FS_CURRENT			1
#define FS_END				2
//---------------------------------------------------------------------------
fd_t cp_open(pctstr_t fname,int mode);
int cp_read(fd_t file,void *data,u32 szdata);
int cp_write(fd_t file,const void *data,u32 szdata);
int cp_close(fd_t file);
s64 cp_seek(fd_t file,s64 offs,int wh);
#define cp_tell(file)		cp_seek(file,0,FS_CURRENT);
//---------------------------------------------------------------------------
#define FF_DIR				(1<<0)
#define FF_DEVICE			(1<<1)
#define FF_LINK				(1<<2)
//---------------------------------------------------------------------------
#define MAXFILENAME			260
//---------------------------------------------------------------------------
typedef struct {
	tchar_t name[MAXFILENAME];
	time_t atime;
	time_t mtime;
	time_t ctime;
	u64 size;
	u32 flags;
} fileinfo_t;
//---------------------------------------------------------------------------
#if (CLOCKS_PER_SEC == 1000)
#define CLK2MS(clk)		(clk)
#else
#define CLK2MS(clk)		((clk)*CLOCKS_PER_SEC)/1000
#endif
//---------------------------------------------------------------------------
int cp_scandir(pctstr_t dir,pctstr_t pattern,fileinfo_t **result);
int cp_subdirs(pctstr_t dir,fileinfo_t **result);
//---------------------------------------------------------------------------
