#include <windows.h>
#include <commctrl.h>
#include <Windowsx.h>
#include <Windowsx.h>
#include <Shlobj.h>
#include "lists.h"
#include "main.h"
#include "log.h"
#include "resource.h"
//----------------------------------------------------------------------------
HWND g_NetList = NULL;
HWND g_KadList = NULL;
HWND g_DownloadsList = NULL;
//----------------------------------------------------------------------------
static void WndSizing(HWND hWnd, WPARAM wParam, LPRECT newWndRect)
{
	RECT oldWndRect;
	RECT oldClientRect;
	const LONG lMinWidth = 400;
	const LONG lMinHeight = 300;
	//
	GetWindowRect(hWnd, &oldWndRect);
	GetClientRect(hWnd, &oldClientRect);
	//
	oldClientRect.right -= oldWndRect.right - oldWndRect.left - (newWndRect->right - newWndRect->left);		// calculate new client width
	oldClientRect.bottom -= oldWndRect.bottom - oldWndRect.top - (newWndRect->bottom - newWndRect->top);	// calculate new client height
	// adjust width
	if (oldClientRect.right < lMinWidth){
		switch (wParam)
		{
		case WMSZ_BOTTOMLEFT:
		case WMSZ_LEFT:
		case WMSZ_TOPLEFT:
			newWndRect->left -= lMinWidth - oldClientRect.right;
			break;
			//
		default:
			newWndRect->right += lMinWidth - oldClientRect.right;
		}
	}
	// adjust height
	if (oldClientRect.bottom < lMinHeight){
		switch (wParam)
		{
		case WMSZ_TOPLEFT:
		case WMSZ_TOP:
		case WMSZ_TOPRIGHT:
			newWndRect->top -= lMinHeight - oldClientRect.bottom;
			break;
			//
		default:
			newWndRect->bottom += lMinHeight - oldClientRect.bottom;
		}
	}
}
//----------------------------------------------------------------------------
static void WndSize(HWND hWnd)
{
	RECT rc;
	GetClientRect(hWnd,&rc);

	if (GetDlgItem(hWnd,IDC_FIND))
		MoveWindow(GetDlgItem(hWnd,IDC_USERS),0,40,rc.right,rc.bottom-40,TRUE);
	else
		MoveWindow(GetDlgItem(hWnd, IDC_USERS), 0, 0, rc.right, rc.bottom, TRUE);
}
//----------------------------------------------------------------------------
static const LPTSTR status_str[NET_STATE_MAX] = {TEXT("Idle"),TEXT("Pinging"),TEXT("Ses requesting"),TEXT("Ses confirmation"),TEXT("Ping requested"),TEXT("Ses requested"),TEXT("Wait confirmation")};
//----------------------------------------------------------------------------
void GetUserStatus(CONNECTION* usr, LPTSTR str)
{
	LPCTSTR status;
	if (usr->established) status = TEXT("Connected");
	else if (usr->est) status = status_str[usr->est->state];
	else status = TEXT("Not connected");
	_stprintf_s(str,100,TEXT("%s %d"),status,usr->send_cnt);
}
//----------------------------------------------------------------------------
static void UpdateNetList(HWND hWnd)
{
	TCHAR str[100];
	HWND hUsers = GetDlgItem(hWnd,IDC_USERS);
	NETWORK* net = getNet();
    CONNECTION* ptr;
    int cnt = 0;
	int cur_sel = ListView_GetSelectionMark(hUsers);
	u32 sel_id[NETWORK_ID_LENGTH];
	int num = ListView_GetItemCount(hUsers);
	if (!net) return;
	if (!ListView_GetSelectedCount(hUsers))
	{
		cur_sel = -1;
	}
	if (cur_sel >= 0)
	{
		ListView_GetItemText(hUsers,cur_sel,4,str,100);
		str2id(str,sel_id);
	}
	for(ptr = net->get_first_node(); ptr; ptr = net->get_next_node(ptr))
	{
		LVITEM item = {0};
		
		item.mask = LVIF_TEXT|LVIF_IMAGE;
		item.pszText = (LPTSTR)ptr->userInfo;
		item.iImage = isTrustUser(ptr->nd.id)?0:2;
		item.iItem = cnt++;
		if (num < cnt) item.iItem = ListView_InsertItem(hUsers,&item);
		else ListView_SetItem(hUsers,&item);
		item.mask = LVIF_TEXT;
		item.iSubItem = 1;
		item.pszText = str;
		ip2str(ptr->nd.ip,str);
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 2;
		_stprintf_s(str,100,TEXT("%d"),ptr->nd.port);
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 3;
		GetUserStatus(ptr,str);
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 4;
		id2str(ptr->nd.id,str);
		ListView_SetItem(hUsers,&item);

		if ((cur_sel >= 0) && isIdIdentical(sel_id,ptr->nd.id))
		{
			ListView_SetItemState(hUsers,item.iItem,LVIS_SELECTED,LVIS_SELECTED);
			ListView_SetSelectionMark(hUsers,item.iItem);
		}
	}
	// clear all other items if exist
	while(cnt < (num = ListView_GetItemCount(hUsers)))
	{
		ListView_DeleteItem(hUsers,num-1);
	}
	ListView_RedrawItems(hUsers,0,cnt);
}
//----------------------------------------------------------------------------
static void ListInit(HWND hWnd)
{
	LVITEM item = {0};
	LVCOLUMN col;
	HWND hUsers = GetDlgItem(hWnd,IDC_USERS);
	ListView_SetExtendedListViewStyle(hUsers,WS_EX_CLIENTEDGE|LVS_EX_FULLROWSELECT|LVS_EX_AUTOSIZECOLUMNS|LVS_EX_GRIDLINES);
	// create columns
	col.mask = LVCF_WIDTH|LVCF_TEXT;
	col.pszText = TEXT("Name");
	col.cx = 100;
	ListView_InsertColumn(hUsers,0,&col);
	col.pszText = TEXT("IP");
	ListView_InsertColumn(hUsers,1,&col);
	col.pszText = TEXT("Port");
	ListView_InsertColumn(hUsers,2,&col);
	col.pszText = TEXT("Status");
	ListView_InsertColumn(hUsers,3,&col);
	col.pszText = TEXT("ID");
	col.cx = 500;
	ListView_InsertColumn(hUsers,4,&col);
	// add image list
	HIMAGELIST hLarge = ImageList_Create(GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), ILC_COLOR32, 3, 1);
	HIMAGELIST hSmall = ImageList_Create(GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), ILC_COLOR32, 3, 1);
	HICON hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_ONLINE));
	ImageList_AddIcon(hLarge, hiconItem);
	ImageList_AddIcon(hSmall, hiconItem);
	DestroyIcon(hiconItem);
	hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_OFFLINE));
	ImageList_AddIcon(hLarge, hiconItem);
	ImageList_AddIcon(hSmall, hiconItem);
	DestroyIcon(hiconItem);
	hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_UNAUTH));
	ImageList_AddIcon(hLarge, hiconItem);
	ImageList_AddIcon(hSmall, hiconItem);
	DestroyIcon(hiconItem);

	ListView_SetImageList(hUsers,hLarge,LVSIL_NORMAL);
	ListView_SetImageList(hUsers,hSmall,LVSIL_SMALL);
	SetTimer(hWnd,1,1000,NULL);
	WndSize(hWnd);
}
//----------------------------------------------------------------------------
static void ListDestroy(HWND hWnd)
{
	KillTimer(hWnd,1);
}
//----------------------------------------------------------------------------
void GetUserStatus(DHT* usr, LPTSTR str)
{
	LPCTSTR status;
	if (usr->flags&KAD_DHT_FLAG_LOST) status = TEXT("Lost");
	else if (usr->flags&KAD_DHT_FLAG_REPLIED) status = TEXT("Responded");
	else if (usr->flags&KAD_DHT_FLAG_PINGED) status = TEXT("Pinged");
	else status = TEXT("Unknown");
	_stprintf_s(str,100,TEXT("%s %d"),status,usr->ping_count);
}
//----------------------------------------------------------------------------
static void UpdateKadList(HWND hWnd)
{
	TCHAR str[100];
	HWND hUsers = GetDlgItem(hWnd,IDC_USERS);
	KADEMLIA* kad = getKad();
	NETWORK* net = getNet();
    DHT* ptr;
    int cnt = 0;
	int cur_sel = ListView_GetSelectionMark(hUsers);
	u32 sel_id[NETWORK_ID_LENGTH];
	TCHAR user_info[user_info_max_length_bytes/sizeof(TCHAR)];
	int num = ListView_GetItemCount(hUsers);
	if (!kad) return;
	if (!ListView_GetSelectedCount(hUsers))
	{
		cur_sel = -1;
	}
	if (cur_sel >= 0)
	{
		ListView_GetItemText(hUsers,cur_sel,4,str,100);
		str2id(str,sel_id);
	}
	for (u32 buck = 0; buck < NETWORK_ID_LENGTH*32; buck++)
	{
		LVITEM item;
		ZeroMemory(&item,sizeof(LVITEM));
		item.mask = LVIF_TEXT|LVIF_IMAGE;
		item.pszText = str;
		item.iImage = -1;
		item.iItem = cnt++;
		_stprintf_s(str,100,TEXT("Group %d"),buck);
		if (num < cnt) item.iItem = ListView_InsertItem(hUsers,&item);
		else ListView_SetItem(hUsers,&item);
		item.mask = LVIF_TEXT;
		item.iSubItem = 1;
		item.pszText = TEXT("");
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 2;
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 3;
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 4;
		ListView_SetItem(hUsers,&item);
		
		for(ptr = kad->get_first_node(buck); ptr; ptr = kad->get_next_node(ptr))
		{
			item.mask = LVIF_TEXT|LVIF_IMAGE;
			net->get_user_info(&ptr->node,user_info);
			item.pszText = user_info;
			if (isTrustUser(ptr->node.id))
			{
				item.iImage = (net->get_user_status(ptr->node.id) == USER_STATUS_ONLINE) ? 0 : 1;
			}
			else
				item.iImage = 2;
			item.iItem = cnt++;
			item.iSubItem = 0;
			if (num < cnt) item.iItem = ListView_InsertItem(hUsers,&item);
			else ListView_SetItem(hUsers,&item);
			item.mask = LVIF_TEXT;
			item.iSubItem = 1;
			item.pszText = str;
			ip2str(ptr->node.ip,str);
			ListView_SetItem(hUsers,&item);
			item.iSubItem = 2;
			_stprintf_s(str,100,TEXT("%d"),ptr->node.port);
			ListView_SetItem(hUsers,&item);
			item.iSubItem = 3;
			GetUserStatus(ptr, str);
			ListView_SetItem(hUsers,&item);
			item.iSubItem = 4;
			id2str(ptr->node.id,str);
			ListView_SetItem(hUsers,&item);

			if ((cur_sel >= 0) && isIdIdentical(sel_id,ptr->node.id))
			{
				ListView_SetItemState(hUsers,item.iItem,LVIS_SELECTED,LVIS_SELECTED);
				ListView_SetSelectionMark(hUsers,item.iItem);
			}
		}
	}
	// clear all other items if exist
	while(cnt < (num = ListView_GetItemCount(hUsers)))
	{
		ListView_DeleteItem(hUsers,num-1);
	}
	ListView_RedrawItems(hUsers,0,cnt);
}
//----------------------------------------------------------------------------
static void ActivateChat(HWND hWnd, int iItem)
{
	TCHAR str[100];
	NODE nd;
	ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),iItem,4,str,100);
	if (str2id(str,nd.id))
	{
		OpenChat(&nd);
	}
}
//----------------------------------------------------------------------------
static void userMenu(HWND hWnd,LPNMITEMACTIVATE ntf)
{
	NODE nd;
	TCHAR str[100]=TEXT("");
	HMENU menu = CreatePopupMenu();
	MENUITEMINFO item;

	ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),ntf->iItem,4,str,100);
	if (!str2id(str,nd.id)) return;
	ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),ntf->iItem,1,str,100);
	str2ip(str,&nd.ip);
	ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),ntf->iItem,2,str,100);
	nd.port = (u16)wcstol(str,NULL,10);

	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_TYPE|MIIM_ID;
	item.fType = MFT_STRING;
	item.wID = IDM_ADDSAVED;
	item.dwTypeData = TEXT("Add to favorite list");
	InsertMenuItem(menu,1,TRUE,&item);
	item.wID = IDM_TRANSFERFILE;
	item.dwTypeData = TEXT("Send file");
	InsertMenuItem(menu, 2, TRUE, &item);
	item.wID = IDM_PING;
	item.dwTypeData = TEXT("Ping");
	InsertMenuItem(menu, 3, TRUE, &item);
	item.wID = IDM_INFO;
	item.dwTypeData = TEXT("Info");
	InsertMenuItem(menu, 4, TRUE, &item);
	ClientToScreen(GetDlgItem(hWnd, IDC_USERS), &ntf->ptAction);
	switch (TrackPopupMenu(menu,TPM_RETURNCMD,ntf->ptAction.x,ntf->ptAction.y,0,hWnd,NULL))
	{
	case IDM_ADDSAVED:
		appendLike(&nd);
		break;
	case IDM_TRANSFERFILE:
		FileTransfer(hWnd, &nd);
		break;
	case IDM_PING:
		getKad()->pingNode(&nd);
		break;
	case IDM_INFO:
		ShowUserInfo(hWnd,&nd);
		break;
	}
}
//----------------------------------------------------------------------------
INT_PTR CALLBACK NetListProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG: ListInit(hWnd); return TRUE;
    case WM_DESTROY: ListDestroy(hWnd); return 0;
	case WM_SIZE: WndSize(hWnd); return 0;
	case WM_SIZING: WndSizing(hWnd,wParam,(LPRECT)lParam); return TRUE;
	case WM_CLOSE: ShowWindow(hWnd,SW_HIDE); return FALSE;
	case WM_TIMER: UpdateNetList(hWnd);
	case WM_NOTIFY:
	{
		LPNMITEMACTIVATE ntf = (LPNMITEMACTIVATE)lParam;
		if (ntf && ntf->hdr.idFrom == IDC_USERS)
		{
			switch (ntf->hdr.code)
			{
			case NM_DBLCLK:
				ActivateChat(hWnd,ntf->iItem);
				break;
			case NM_RCLICK:
				userMenu(hWnd,ntf);
				break;
			}
		}
		return FALSE;
	}
	}
	return FALSE;
}
//----------------------------------------------------------------------------
INT_PTR CALLBACK KadListProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG: ListInit(hWnd); return TRUE;
    case WM_DESTROY: ListDestroy(hWnd); return 0;
	case WM_SIZE: WndSize(hWnd); return 0;
	case WM_SIZING: WndSizing(hWnd, wParam, (LPRECT)lParam); return TRUE;
	case WM_CLOSE: ShowWindow(hWnd,SW_HIDE); return FALSE;
	case WM_TIMER: UpdateKadList(hWnd);
	case WM_NOTIFY:
	{
		LPNMITEMACTIVATE ntf = (LPNMITEMACTIVATE)lParam;
		if (ntf && ntf->hdr.idFrom == IDC_USERS)
		{
			switch (ntf->hdr.code)
			{
			case NM_DBLCLK:
				ActivateChat(hWnd,ntf->iItem);
				break;
			case NM_RCLICK:
				userMenu(hWnd,ntf);
				break;
			}
		}
		return FALSE;
	}
	}
	return FALSE;
}
//----------------------------------------------------------------------------
size_t _find_slash(tchar_t* str)
{
	size_t i = 0;
	for (i = cp_strlen(str); i--;)
	{
		if (str[i] == '\\') break;
	}
	return i+1;
}
//----------------------------------------------------------------------------
VOID UpdateDownloads(HWND hWnd)
{
	TCHAR str[100];
	HWND hUsers = GetDlgItem(hWnd,IDC_USERS);
	USERS* usr = getUserList();
    int cnt = 0;
	int cur_sel = ListView_GetSelectionMark(hUsers);
	int num = ListView_GetItemCount(hUsers);
	if (!usr) return;
	for(download_t *ptr = usr->get_downloads(); ptr; ptr = ptr->next)
	{
		LVITEM item = {0};
		
		item.mask = LVIF_TEXT;
//		item.pszText = (LPTSTR)&ptr->file[_find_slash(ptr->file)];
		item.pszText = ptr->file;
		item.iItem = cnt++;
		if (num < cnt) item.iItem = ListView_InsertItem(hUsers,&item);
		else ListView_SetItem(hUsers,&item);

		item.mask = LVIF_TEXT;
		item.iSubItem = 1;
		item.pszText = ptr->user;
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 2;
		time2str(ptr->time,str);
		item.pszText = str;
		ListView_SetItem(hUsers,&item);

		if (cur_sel == cnt)
		{
			ListView_SetItemState(hUsers,item.iItem,LVIS_SELECTED,LVIS_SELECTED);
			ListView_SetSelectionMark(hUsers,item.iItem);
		}
	}
	// clear all other items if exist
	while(cnt < (num = ListView_GetItemCount(hUsers)))
	{
		ListView_DeleteItem(hUsers,num-1);
	}
	ListView_RedrawItems(hUsers,0,cnt);
}
//----------------------------------------------------------------------------
static void OpenFile(HWND hWnd, int num)
{
	int cnt = 0;
	USERS* usr = getUserList();
	for(download_t *ptr = usr->get_downloads(); ptr; ptr = ptr->next)
	{
		if (cnt == num)
		{
			ShellExecute(hWnd,NULL,ptr->file,NULL,NULL,SW_SHOWNORMAL);
			break;
		}
		cnt++;
	}
}
//----------------------------------------------------------------------------
static void ShowFile(HWND hWnd, int num)
{
	int cnt = 0;
	USERS* usr = getUserList();
	for(download_t *ptr = usr->get_downloads(); ptr; ptr = ptr->next)
	{
		if (cnt == num)
		{
			PIDLIST_ABSOLUTE pidl = ILCreateFromPath(ptr->file);
			if(pidl) {
				SHOpenFolderAndSelectItems(pidl,0,0,0);
				ILFree(pidl);
			}
			break;
		}
		cnt++;
	}
}
//----------------------------------------------------------------------------
static void downloadsMenu(HWND hWnd,LPNMITEMACTIVATE ntf)
{
	HMENU menu = CreatePopupMenu();
	MENUITEMINFO item;
	USERS *usr = getUserList();

	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_TYPE|MIIM_ID;
	item.fType = MFT_STRING;
	item.wID = IDM_OPEN;
	item.dwTypeData = TEXT("Open");
	InsertMenuItem(menu,1,TRUE,&item);
	item.wID = IDM_SHOW_FOLDER;
	item.dwTypeData = TEXT("Show in folder");
	InsertMenuItem(menu, 2, TRUE, &item);
	item.wID = IDM_DELETE;
	item.dwTypeData = TEXT("Delete");
	InsertMenuItem(menu,3,TRUE,&item);
	item.wID = IDM_DELETE_ALL;
	item.dwTypeData = TEXT("Delete All");
	InsertMenuItem(menu,4,TRUE,&item);
	ClientToScreen(GetDlgItem(hWnd, IDC_USERS), &ntf->ptAction);
	switch (TrackPopupMenu(menu,TPM_RETURNCMD,ntf->ptAction.x,ntf->ptAction.y,0,hWnd,NULL))
	{
	case IDM_OPEN:
		OpenFile(hWnd,ntf->iItem);
		break;
	case IDM_SHOW_FOLDER:
		ShowFile(hWnd,ntf->iItem);
		break;
	case IDM_DELETE:
		usr->remove_download(ntf->iItem);
		break;
	case IDM_DELETE_ALL:
		while(usr->get_downloads()) usr->remove_download(0);
		break;
	}
	UpdateDownloads(hWnd);
}
//----------------------------------------------------------------------------
static void DownloadsInit(HWND hWnd)
{
	HWND hUsers = GetDlgItem(hWnd,IDC_USERS);
	ListView_SetExtendedListViewStyle(hUsers,WS_EX_CLIENTEDGE|LVS_EX_FULLROWSELECT|LVS_EX_AUTOSIZECOLUMNS|LVS_EX_GRIDLINES);
	LVITEM item = {0};
	LVCOLUMN col;
	ListView_SetExtendedListViewStyle(hUsers,WS_EX_CLIENTEDGE|LVS_EX_FULLROWSELECT|LVS_EX_AUTOSIZECOLUMNS|LVS_EX_GRIDLINES);
	// create columns
	col.mask = LVCF_WIDTH|LVCF_TEXT;
	col.pszText = TEXT("File name");
	col.cx = 300;
	ListView_InsertColumn(hUsers,0,&col);
	col.pszText = TEXT("From");
	col.cx = 100;
	ListView_InsertColumn(hUsers,1,&col);
	col.pszText = TEXT("Date");
	ListView_InsertColumn(hUsers,2,&col);
	WndSize(hWnd);
}
//----------------------------------------------------------------------------
static void DownloadsDestroy(HWND hWnd)
{
}
//----------------------------------------------------------------------------
INT_PTR CALLBACK DownloadsProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG: DownloadsInit(hWnd); return TRUE;
    case WM_DESTROY: DownloadsDestroy(hWnd); return 0;
	case WM_SIZE: WndSize(hWnd); return 0;
	case WM_SIZING: WndSizing(hWnd, wParam, (LPRECT)lParam); return TRUE;
	case WM_CLOSE: ShowWindow(hWnd,SW_HIDE); return FALSE;
	case WM_NOTIFY:
	{
		LPNMITEMACTIVATE ntf = (LPNMITEMACTIVATE)lParam;
		if (ntf && ntf->hdr.idFrom == IDC_USERS)
		{
			switch (ntf->hdr.code)
			{
			case NM_DBLCLK:
				OpenFile(hWnd,ntf->iItem);
				break;
			case NM_RCLICK:
				downloadsMenu(hWnd,ntf);
				break;
			}
		}
		return FALSE;
	}
	case WM_UPDATE_PROGRESS:
		UpdateDownloads(hWnd);
		return TRUE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
void CreateListsDialog(HWND hWnd)
{
	g_NetList = CreateDialog(GetModuleHandle(NULL),TEXT("Network"),hWnd,NetListProc);
	g_KadList = CreateDialog(GetModuleHandle(NULL),TEXT("Kademlia"),hWnd,KadListProc);
	g_DownloadsList = CreateDialog(GetModuleHandle(NULL),TEXT("Downloads"),hWnd,DownloadsProc);
}
//----------------------------------------------------------------------------
static int iequals_lower(LPCTSTR str1, LPCTSTR str2, size_t num)
{
	for (size_t i = 0; (i < num) && str1[i] && str2[i]; i++)
	{
		int diff = tolower(str1[i]) - tolower(str2[i]);
		if (diff) return diff;
	}
    return 0;
}
//----------------------------------------------------------------------------
static bool names_match(LPTSTR name1, LPTSTR name2)
{
	int len1 = (int)cp_strlen(name1);
	int len2 = (int)cp_strlen(name2);
	int pos1[KAD_MAX_HASHES-2];
	int pos2[KAD_MAX_HASHES-2];
	int i,j;

	if (!iequals_lower(name1, name2, min(len1,len2))) return true;
	for (int i = 0; i < KAD_MAX_HASHES-2; i++)
	{
		pos1[i] = -1;
		pos2[i] = -1;
	}
	for (i = 1, j = 0; (i < len1) && (j < KAD_MAX_HASHES-2); i++)
	{
		if ((name1[i] == ' ') && (name1[i-1] != ' '))
		{
			pos1[j++] = i;
		}
	}
	for (i = 1, j = 0; (i < len2) && (j < KAD_MAX_HASHES-2); i++)
	{
		if ((name2[i] == ' ') && (name2[i-1] != ' '))
		{
			pos2[j++] = i;
		}
	}
	for (i = 0; i < KAD_MAX_HASHES-2; i++)
	{
		int start1 = i?pos1[i-1]+1:0;
		size_t part_len1 = ((pos1[i] < 0)?len1:pos1[i]) - start1;
		for (j = 0; j < KAD_MAX_HASHES-2; j++)
		{
			int start2 = j?pos2[j-1]+1:0;
			size_t part_len2 = ((pos2[j] < 0)?len2:pos2[j])-start2;

			if (!iequals_lower(&name1[start1], &name2[start2], min(part_len1,part_len2))) return true;
		}
		if (pos1[i] < 0) break;
	}
	return false;
}
//----------------------------------------------------------------------------
static void UpdateFindList(HWND hWnd, LPTSTR name, u32* id, u32 ip)
{
	TCHAR str[100];
	HWND hUsers = GetDlgItem(hWnd, IDC_USERS);
	NETWORK* net = getNet();
	KADEMLIA* kad = getKad();
	CONNECTION* ptr;
	DHT* dht;
	int cnt = 0;
	int cur_sel = ListView_GetSelectionMark(hUsers);
	u32 sel_id[NETWORK_ID_LENGTH];
	int num = ListView_GetItemCount(hUsers);
	TCHAR user_info[user_info_max_length_bytes / sizeof(TCHAR)];
	if (!net || !kad) return;
	if (!ListView_GetSelectedCount(hUsers))
	{
		cur_sel = -1;
	}
	if (cur_sel >= 0)
	{
		ListView_GetItemText(hUsers, cur_sel, 4, str, 100);
		str2id(str, sel_id);
	}
	for (ptr = net->get_first_node(); ptr; ptr = net->get_next_node(ptr))
	{
		LVITEM item = { 0 };

		if ((name && !id && !ip && (cp_strlen(name) > 0) && !names_match(name, (LPTSTR)ptr->userInfo)) || (id && !ip && !isIdIdentical(ptr->nd.id, id)) || (ip && (ip != ptr->nd.ip))) continue;

		item.mask = LVIF_TEXT | LVIF_IMAGE;
		item.pszText = (LPTSTR)ptr->userInfo;
		item.iImage = isTrustUser(ptr->nd.id) ? 0 : 2;
		item.iItem = cnt++;
		if (num < cnt) item.iItem = ListView_InsertItem(hUsers, &item);
		else ListView_SetItem(hUsers, &item);
		item.mask = LVIF_TEXT;
		item.iSubItem = 1;
		item.pszText = str;
		ip2str(ptr->nd.ip, str);
		ListView_SetItem(hUsers, &item);
		item.iSubItem = 2;
		_stprintf_s(str, 100, TEXT("%d"), ptr->nd.port);
		ListView_SetItem(hUsers, &item);
		item.iSubItem = 4;
		id2str(ptr->nd.id, str);
		ListView_SetItem(hUsers, &item);

		if ((cur_sel >= 0) && isIdIdentical(sel_id, ptr->nd.id))
		{
			ListView_SetItemState(hUsers, item.iItem, LVIS_SELECTED, LVIS_SELECTED);
			ListView_SetSelectionMark(hUsers, item.iItem);
		}
	}
	// from Kad
	if (id)
	{
		for (u32 buck = 0; buck < NETWORK_ID_LENGTH * 32; buck++)
		{
			for (dht = kad->get_first_node(buck); dht; dht = kad->get_next_node(dht))
			{
				LVITEM item = { 0 };

				if (!isIdIdentical(dht->node.id, id) && (dht->node.ip != ip)) continue;

				item.mask = LVIF_TEXT | LVIF_IMAGE;
				net->get_user_info(&dht->node, user_info);
				item.pszText = user_info;
				if (isTrustUser(dht->node.id))
				{
					item.iImage = (net->get_user_status(dht->node.id) == USER_STATUS_ONLINE) ? 0 : 1;
				}
				else
					item.iImage = 2;
				item.iItem = cnt++;
				item.iSubItem = 0;
				if (num < cnt) item.iItem = ListView_InsertItem(hUsers, &item);
				else ListView_SetItem(hUsers, &item);
				item.mask = LVIF_TEXT;
				item.iSubItem = 1;
				item.pszText = str;
				ip2str(dht->node.ip, str);
				ListView_SetItem(hUsers, &item);
				item.iSubItem = 2;
				_stprintf_s(str, 100, TEXT("%d"), dht->node.port);
				ListView_SetItem(hUsers, &item);
				item.iSubItem = 4;
				id2str(dht->node.id, str);
				ListView_SetItem(hUsers, &item);

				if ((cur_sel >= 0) && isIdIdentical(sel_id, dht->node.id))
				{
					ListView_SetItemState(hUsers, item.iItem, LVIS_SELECTED, LVIS_SELECTED);
					ListView_SetSelectionMark(hUsers, item.iItem);
				}
			}
		}
	}
	// clear all other items if exist
	while (cnt < (num = ListView_GetItemCount(hUsers)))
	{
		ListView_DeleteItem(hUsers, num - 1);
	}
	ListView_RedrawItems(hUsers, 0, cnt);
}
//----------------------------------------------------------------------------
static void StartSearch(HWND hWnd, LPTSTR name)
{
	KADEMLIA* kad = getKad();
	NETWORK* net = getNet();
	u32* id = (u32*)LocalAlloc(LPTR, sizeof(u32)*NETWORK_ID_LENGTH);
	NODE nd;
	if (GetProp(hWnd, TEXT("id"))) LocalFree(RemoveProp(hWnd, TEXT("id")));
	LogMessage(TEXT("Find"), name);
	if (NETWORK_ID_LENGTH == _stscanf_s(name, TEXT("%08X %08X %08X %08X %08X %08X %08X %08X"), &id[0], &id[1], &id[2], &id[3], &id[4], &id[5], &id[6], &id[7]))
	{
		SetProp(hWnd, TEXT("id"), id);
		LogMessage(TEXT("Find"), TEXT("User ID typed"));
		if (kad->findNode(id, &nd))
		{
//			selectUser(hWnd, id);
			LogMessage(TEXT("Find"), TEXT("User found"));
		}
	}
	if (4 == _stscanf_s(name, TEXT("%d.%d.%d.%d"), &id[0], &id[1], &id[2], &id[3]))
	{
		u32 ip = (id[0]<<24)+(id[1]<<16)+(id[2]<<8)+id[3];
		SetProp(hWnd, TEXT("ip"), (HANDLE)ip);
		LogMessage(TEXT("Find"), TEXT("User IP typed"));
		if (kad->findNode(ip, &nd))
		{
//			selectUser(hWnd, id);
			LogMessage(TEXT("Find"), TEXT("User found"));
		}
	}
	else
	{
		LogMessage(TEXT("Find"), TEXT("User name typed"));
		if (net->findUserByName((u8*)name, (u16)(_tcslen(name)*sizeof(TCHAR)), id))
		{
			LogMessage(TEXT("Find"), TEXT("User found"));
		}
		kad->findNode(name);
		LocalFree(id);
	}
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK FindDialog(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetProp(hWnd, TEXT("str"), (HANDLE)LocalAlloc(LPTR, user_info_max_length_bytes + sizeof(TCHAR)));
		Edit_LimitText(GetDlgItem(hWnd, IDC_FIND), user_info_max_length_bytes / sizeof(TCHAR));
		ListInit(hWnd);
		return TRUE;
	case WM_DESTROY:
		ListDestroy(hWnd);
		LocalFree(RemoveProp(hWnd, TEXT("str")));
		if (GetProp(hWnd, TEXT("id")))
		{
			LocalFree(RemoveProp(hWnd, TEXT("id")));
		}
		RemoveProp(hWnd,TEXT("ip"));
		return 0;
	case WM_CLOSE:
		DestroyWindow(hWnd);
		return TRUE;
	case WM_SIZE: WndSize(hWnd); return 0;
	case WM_SIZING: WndSizing(hWnd, wParam, (LPRECT)lParam); return TRUE;
	case WM_TIMER: UpdateFindList(hWnd, (LPTSTR)GetProp(hWnd, TEXT("str")), (u32*)GetProp(hWnd, TEXT("id")), (u32)GetProp(hWnd, TEXT("ip"))); return 0;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			GetDlgItemText(hWnd, IDC_FIND, (LPTSTR)GetProp(hWnd, TEXT("str")), user_info_max_length_bytes / sizeof(TCHAR)+sizeof(TCHAR));
			StartSearch(hWnd, (LPTSTR)GetProp(hWnd, TEXT("str")));
			return TRUE;
		}
		return FALSE;
	case WM_NOTIFY:
	{
		LPNMITEMACTIVATE ntf = (LPNMITEMACTIVATE)lParam;
		if (ntf && ntf->hdr.idFrom == IDC_USERS)
		{
			switch (ntf->hdr.code)
			{
			case NM_DBLCLK:
				ActivateChat(hWnd, ntf->iItem);
				break;
			case NM_RCLICK:
				userMenu(hWnd, ntf);
				break;
			}
		}
		return FALSE;
	}
	}
	return FALSE;
}
//----------------------------------------------------------------------------
void findUser(HWND hWnd)
{
	CreateDialog(GetModuleHandle(NULL), TEXT("Find"), hWnd, FindDialog);
}
//----------------------------------------------------------------------------
