#include <windows.h>
#include <Windowsx.h>
#include <commctrl.h>
#include <tchar.h>
#include <stdio.h>
#include <Share.h>

#include "resource.h"
#include "settings.h"
#include "network.h"
#include "main.h"
#include "build.h"
//----------------------------------------------------------------------------
HWND g_logWnd = NULL;
//----------------------------------------------------------------------------
typedef struct {
	FILE* F;
	CRITICAL_SECTION cs;
} params_t;
static params_t *g_par;
//----------------------------------------------------------------------------
static void LogSize(HWND hWnd)
{
	RECT rc;
	GetClientRect(hWnd,&rc);
	MoveWindow(GetDlgItem(hWnd,IDC_LOG),0,0,rc.right,rc.bottom,TRUE);
}
//----------------------------------------------------------------------------
static void ApplySettings(HWND hWnd,settings_t* set)
{
	SYSTEMTIME time;
	TCHAR datetime[100];
	GetLocalTime(&time);
	_stprintf_s(datetime,100,TEXT("%02d/%02d/%02d %02d:%02d:%02d"),time.wYear%100,time.wMonth,time.wDay,time.wHour,time.wMinute,time.wSecond);
	params_t *par = (params_t*)GetProp(hWnd,TEXT("par"));
	EnterCriticalSection(&par->cs);
	if (par->F) {
		_ftprintf_s(par->F,TEXT("\n\nLog normaly finished at %s\n\n"),datetime);
		fclose(par->F);
		par->F = NULL;
	}
	if (set->enableLogFile)
	{
		if (!(par->F = _tfsopen(set->logFileName,TEXT("a"), _SH_DENYWR)))
		{
			LeaveCriticalSection(&par->cs);
			MessageBox(hWnd,TEXT("Can't create log file"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
		}
		else
		{
			_ftprintf_s(par->F,TEXT("\n\nLog started at %s, ZAS version %s\n\n"),datetime,TEXT(BUILD_NUM));
			LeaveCriticalSection(&par->cs);
		}
	}
	else LeaveCriticalSection(&par->cs);
}
//----------------------------------------------------------------------------
static void LogInit(HWND hWnd, settings_t* set)
{
	LVCOLUMN col;
	params_t *par = (params_t*)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(params_t));
	SetProp(hWnd,TEXT("par"),par);
	g_par = par;
	HWND hLog = GetDlgItem(hWnd,IDC_LOG);
	ListView_SetExtendedListViewStyle(hLog,LVS_EX_FULLROWSELECT|/*LVS_EX_AUTOSIZECOLUMNS|*/LVS_EX_GRIDLINES);
	// create columns
	col.mask = LVCF_WIDTH|LVCF_TEXT;
	col.cx = 150;
	col.pszText = TEXT("Time");
	ListView_InsertColumn(hLog,0,&col);
	col.cx = 100;
	col.pszText = TEXT("Source");
	ListView_InsertColumn(hLog,1,&col);
	col.cx = 800;
	col.pszText = TEXT("Message");
	ListView_InsertColumn(hLog,2,&col);
	g_logWnd = hWnd;
	InitializeCriticalSection(&par->cs);
	ApplySettings(hWnd,set);
	LogSize(hWnd);
}
//----------------------------------------------------------------------------
static void LogDestroy(HWND hWnd)
{
	params_t *par = (params_t*)RemoveProp(hWnd,TEXT("par"));
	g_par = NULL;
	DeleteCriticalSection(&par->cs);
	if (par->F)
	{
		SYSTEMTIME time;
		GetLocalTime(&time);
		_ftprintf_s(par->F,TEXT("\n\nLog normaly finished at %02d/%02d/%02d %02d:%02d:%02d\n\n"),time.wYear%100,time.wMonth,time.wDay,time.wHour,time.wMinute,time.wSecond);
		 fclose(par->F);
	}
	HeapFree(GetProcessHeap(),0,par);
}
//----------------------------------------------------------------------------
INT_PTR CALLBACK LogProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		LogInit(hWnd,(settings_t*)lParam);
		return TRUE;
		//
    case WM_DESTROY:
		LogDestroy(hWnd);
		return 0;
		//
	case WM_CLOSE:
		ShowWindow(hWnd,SW_HIDE);
		return TRUE;
		//
	case WM_SIZE:
		LogSize(hWnd);
		return TRUE;
		//
	case WM_APPLY_LOG_SETTINGS:
		ApplySettings(hWnd,(settings_t*)lParam);
		return TRUE;
	}
	return FALSE;
}
//-----------------------------------------------------------------------
void LogMessage(LPTSTR module, LPTSTR message)
{
	TCHAR datetime[100];
	SYSTEMTIME time;
	LVITEM item = {0};
	params_t *par = (params_t*)GetProp(g_logWnd,TEXT("par"));
	PDWORD res;

	if (!IsWindow(g_logWnd)) return;
	if (message[0] == 0xCCCC)
		return;

	GetLocalTime(&time);
	_stprintf_s(datetime,100,TEXT("%02d/%02d/%02d %02d:%02d:%02d"),time.wYear%100,time.wMonth,time.wDay,time.wHour,time.wMinute,time.wSecond);

	item.mask = LVIF_TEXT;
	item.pszText = datetime;
	if (SendMessageTimeout(GetDlgItem(g_logWnd,IDC_LOG),LVM_GETITEMCOUNT,0,0,0,100,(PDWORD_PTR)&item.iItem) &&
		SendMessageTimeout(GetDlgItem(g_logWnd,IDC_LOG),LVM_INSERTITEM,0,(LPARAM)&item,0,100,(PDWORD_PTR)&res))
	{
		item.iItem = (int)res;
		item.iSubItem = 1;
		item.pszText = module;
		SendMessageTimeout(GetDlgItem(g_logWnd,IDC_LOG),LVM_SETITEM,0,(LPARAM)&item,0,100,(PDWORD_PTR)&res);
		item.iSubItem = 2;
		item.pszText = message;
		SendMessageTimeout(GetDlgItem(g_logWnd,IDC_LOG),LVM_SETITEM,0,(LPARAM)&item,0,100,(PDWORD_PTR)&res);
		ListView_Scroll(GetDlgItem(g_logWnd,IDC_LOG),0,20);
	}

	if (par->F)
	{
		EnterCriticalSection(&par->cs);
		_ftprintf_s(par->F,TEXT("%s - %s:\t%s\n"),datetime,module,message);
		LeaveCriticalSection(&par->cs);
	}
	
	if (par->F) fflush(par->F);
}
//-----------------------------------------------------------------------
static const LPTSTR types[SESSION_MAX_PACK+1] = {_T("encoded"),_T("ses_request"),_T("ses_response"),_T("ping_req"),_T("ping_resp"),_T("name_req"),_T("name_resp"),_T("ses_confirm"),_T("ses_terminate"),_T("busy"),_T("reject"),_T("confirmed"),_T("UNKNOWN")};
//-----------------------------------------------------------------------
void LogLowLevel(bool tx, u32 ip, u16 port, const u32* id, u32 type, u32 size)
{
	TCHAR datetime[100];
	SYSTEMTIME time;
	u32 t = type;

	if (g_par && g_par->F && id)
	{
		GetLocalTime(&time);
		_stprintf_s(datetime,TEXT("%02d/%02d/%02d %02d:%02d:%02d"),time.wYear%100,time.wMonth,time.wDay,time.wHour,time.wMinute,time.wSecond);
		if (t > SESSION_MAX_PACK) t = SESSION_MAX_PACK;
		EnterCriticalSection(&g_par->cs);
		_ftprintf_s(g_par->F,TEXT("%s:\t%s\t%3d.%3d.%3d.%3d:%5d %08X %08X %08X %08X %08X %08X %08X %08X type %d (%s) size %d\n"),datetime,tx?_T("TX"):_T("RX"),FIRST_IPADDRESS(ip),SECOND_IPADDRESS(ip),THIRD_IPADDRESS(ip),FOURTH_IPADDRESS(ip),port,
				id[0],id[1],id[2],id[3],id[4],id[5],id[6],id[7],type,types[t],size);
		fflush(g_par->F);
		if (getLogSize() && (ftell(g_par->F) > (int)getLogSize()*1024*1024)) // log file exceeds the limit
		{
			TCHAR new_name[MAX_PATH];
			LPCTSTR log_name = get_log_file_name();
			fclose(g_par->F);
			_stprintf_s(new_name,TEXT("%s.bak"),log_name);
			MoveFile(log_name,new_name);
			g_par->F = _tfsopen(log_name,TEXT("w"), _SH_DENYWR);
		}
		LeaveCriticalSection(&g_par->cs);
	}
}
//-----------------------------------------------------------------------
void LogLowLevelEnc(bool tx, const u32* data)
{
	if (g_par && g_par->F)
	{
		EnterCriticalSection(&g_par->cs);
		_ftprintf_s(g_par->F,TEXT("%s: data 0x%08X 0x%08X\n"),tx?_T("TX"):_T("RX"),data[0],data[1]);
		fflush(g_par->F);
		LeaveCriticalSection(&g_par->cs);
	}
}
//-----------------------------------------------------------------------
void LogString(LPCTSTR msg)
{
	if (g_par && g_par->F)
	{
		EnterCriticalSection(&g_par->cs);
		_ftprintf_s(g_par->F,TEXT("%s\n"),msg);
		fflush(g_par->F);
		LeaveCriticalSection(&g_par->cs);
	}
}
//-----------------------------------------------------------------------
