#include <winsock2.h>
#include <windows.h>
#include <commctrl.h>
#include <tchar.h>
#include <Windowsx.h>

#include "resource.h"
#include "settings.h"
#include "password.h"
#include "network.h"
#include "kademlia.h"
#include "chat.h"
#include "build.h"
#include "upnp.h"
#include "log.h"
#include "users.h"
#include "lists.h"
#include "main.h"
#include "file_transfer.h"
//----------------------------------------------------------------------------
static LPTSTR module = TEXT("Main");
//----------------------------------------------------------------------------
typedef struct {
	HWND hWnd;
	NODE nd;
	chat_t* ch;
} chatwnd_t;
//----------------------------------------------------------------------------
enum {
	STATE_ONLINE,
	STATE_AWAY,
	STATE_AWAY_MANUAL
};
//----------------------------------------------------------------------------
typedef struct {
	CRITICAL_SECTION cs;
	u8 password[password_max_length];
	u8 passwordMix[password_max_length];
	u32 passwordLength;
	u32 passwordTimeOut;
	HANDLE thPassword;
	HANDLE eStopThreadPassword;
	SECRET_DATA sdata;
	KADEMLIA* kad;
	NETWORK* net;
	USERS* usr;
	FILE_TRANSFER* ft;
	CRITICAL_SECTION chat_cs;
	chatwnd_t* chats;
	u32 chats_num;
	HWND hMainWnd;
	HWND hInfo;
	UINT wiNum;
	UINT woNum;
	UINT volume;
	UINT mic;
	UINT state;
	settings_t *set;
	CRITICAL_SECTION sig_cs;
} instance_t;
//----------------------------------------------------------------------------
static instance_t* inst; // must be static for access without message transfer
static HANDLE gMutex;
//----------------------------------------------------------------------------
KADEMLIA* getKad(void)
{
	return inst->kad;
}
//----------------------------------------------------------------------------
NETWORK* getNet(void)
{
	return inst->net;
}
//----------------------------------------------------------------------------
SECRET_DATA* getSecretData(void)
{
	return &inst->sdata;
}
//----------------------------------------------------------------------------
FILE_TRANSFER* getFileTransfer(void)
{
	return inst->ft;
}
//----------------------------------------------------------------------------
USERS* getUserList(void)
{
	return inst->usr;
}
//----------------------------------------------------------------------------
bool alowAnyone(void)
{
	return inst->set->allowAnyone!=0;
}
//----------------------------------------------------------------------------
bool saveChat(void)
{
	return inst->set->saveChat!=0;
}
//----------------------------------------------------------------------------
void getWaveNums(UINT *wiNum, UINT *woNum)
{
	if (wiNum) *wiNum = inst->wiNum;
	if (woNum) *woNum = inst->woNum;
}
//----------------------------------------------------------------------------
void getVolumeMic(UINT* vol, UINT* mic)
{
	if (vol) *vol = inst->volume;
	if (mic) *mic = inst->mic;
}
//----------------------------------------------------------------------------
UINT getRingVolume(void)
{
	return inst->set->ring_volume;
}
//----------------------------------------------------------------------------
UINT getLogSize(void)
{
	return inst->set->log_size;
}
//----------------------------------------------------------------------------
void setVolume(WORD value)
{
	inst->volume = value;
}
void setMic(WORD value)
{
	inst->mic = value;
}
bool isAGCEnabled(void)
{
	return inst->set->mic_agc!=0;
}
//----------------------------------------------------------------------------
void apply_AGC(int state)
{
	for (int cnt = 0; cnt < (int)inst->chats_num; cnt++)
	{
		if (IsWindow(inst->chats[cnt].hWnd) && PostMessage(inst->chats[cnt].hWnd,WM_SET_AGC,state,0))
			return;
	}
}
//----------------------------------------------------------------------------
LPCTSTR get_log_file_name(void)
{
	return inst->set->logFileName;
}
//----------------------------------------------------------------------------
void getFileTransferPath(LPTSTR FileName)
{
	settings_t *set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	cp_strcchcpy(FileName, MAX_PATH, set->ftPath);
	cp_strcchcpy(&FileName[cp_strlen(FileName)], MAX_PATH - cp_strlen(FileName), TEXT("\\"));
}
//----------------------------------------------------------------------------
bool acceptFileTransfer(void)
{
	settings_t *set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	return (set->reciveFiles == 1);
}
//----------------------------------------------------------------------------
void SetWaveParam(settings_t* set)
{
	inst->wiNum = WAVE_MAPPER;
	for (UINT i = 0; i < waveInGetNumDevs(); i++)
	{
		WAVEINCAPS cap;
		if (MMSYSERR_NOERROR == waveInGetDevCaps(i,&cap,sizeof(cap)))
		{
			if (!cp_strcmp(set->microphone,cap.szPname))
			{
				inst->wiNum = i;
				break;
			}
		}
	}
	inst->woNum = WAVE_MAPPER;
	for (UINT i = 0; i < waveOutGetNumDevs(); i++)
	{
		WAVEOUTCAPS cap;
		if (MMSYSERR_NOERROR == waveOutGetDevCaps(i,&cap,sizeof(cap)))
		{
			if (!cp_strcmp(set->speaker,cap.szPname))
			{
				inst->woNum = i;
				break;
			}
		}
	}
	inst->volume = set->volume;
	inst->mic = set->mic;
	for (u32 c = 0; c < inst->chats_num; c++)
	{
		PostMessage(inst->chats[c].hWnd, WM_RESTART_WAVE, 0,0);
	}
}
//----------------------------------------------------------------------------
DWORD WINAPI SignatureThread(void* s)
{
	settings_t* set = (settings_t*)s;
	while (!TryEnterCriticalSection(&inst->sig_cs)) Sleep(500);
	save_file_signature(set->likesFileName);
	LeaveCriticalSection(&inst->sig_cs);
	return 0;
}

//----------------------------------------------------------------------------
DWORD WINAPI PasswordThread(void* param)
{
	instance_t* inst = (instance_t*)param;
	settings_t *set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	while (WAIT_TIMEOUT == WaitForSingleObject(inst->eStopThreadPassword, 60000))
	{
		EnterCriticalSection(&inst->cs);
		if (!inst->passwordTimeOut && inst->passwordLength && set->passwordTime) // reset password
		{
			ZeroMemory(inst->password,sizeof(inst->password));
			ZeroMemory(inst->passwordMix,sizeof(inst->password));
			inst->passwordLength = 0;
			LeaveCriticalSection(&inst->cs);
			LogMessage(module,TEXT("Password Cleared"));
		}
		else
		{
			u32 i;
			if (inst->passwordTimeOut)
			{
				inst->passwordTimeOut--;
			}
			for (i = 0; i < inst->passwordLength; i++)
			{
				inst->password[i] ^= inst->passwordMix[i];
			}
			cheap_random_number(inst->passwordMix, (inst->passwordLength+3)/4);
			for (i = 0; i < inst->passwordLength; i++)
			{
				inst->password[i] ^= inst->passwordMix[i];
			}
			LeaveCriticalSection(&inst->cs);
		}
	}
	return 0;
}
//----------------------------------------------------------------------------
void setPassword(u8* password,u32 pass_len,settings_t* set)
{
	inst->passwordTimeOut = set->passwordTime;
	inst->passwordLength = pass_len;
	cheap_random_number(inst->passwordMix, pass_len);
	for (u32 i = 0; i < inst->passwordLength; i++)
	{
		inst->password[i] = password[i]^inst->passwordMix[i];
	}
}
//----------------------------------------------------------------------------
u32 getPassword(u8* password)
{
	u32 pass_len = 0;
	u32 i;
	if (!inst->sdata.data_is_initialized) return 0;
	// try to get remembered password
	EnterCriticalSection(&inst->cs);
	if (inst->passwordLength)
	{
		for (i = 0; i < inst->passwordLength; i++)
		{
			password[i] = inst->password[i]^inst->passwordMix[i];
		}
		pass_len = inst->passwordLength;
	}
	LeaveCriticalSection(&inst->cs);

	if (pass_len) 
	{
		LogMessage(module,TEXT("Get password: stored value"));
		return pass_len; // password exists
	}

	u32 cnt = 5;
	while(cnt--)
	{
		settings_t *set = (settings_t*)GetProp(inst->hMainWnd,TEXT("set"));
		if (set)
		{
			int rem = set->rememberPassword;
			pass_len = PasswordRequest(inst->hMainWnd,password,&rem);
			if (!pass_len) return 0;
			if (check_password(&inst->sdata.encoded_secret_key,password,pass_len))
			{
				EnterCriticalSection(&inst->cs);
				setPassword(password,pass_len,set);
				LeaveCriticalSection(&inst->cs);
				set->rememberPassword = rem;
				LogMessage(module,TEXT("Get password: user is requested"));
				return pass_len;
			}
			else MessageBox(inst->hMainWnd,TEXT("Password doesn't match key\nTry again"),TEXT("Password"),MB_OK|MB_ICONEXCLAMATION);
		}
	}
	MessageBox(inst->hMainWnd,TEXT("Too many attempts to enter the password\nPROGRAM WILL BE CLOSED"),TEXT("Error"),MB_OK|MB_ICONERROR);
	LogMessage(module,TEXT("USER ENTER WRONG PASSWORD MANY TIMES PROGRAM WILL BE CLOSED"));
	DestroyWindow(inst->hMainWnd);
	return 0;
}
//----------------------------------------------------------------------------
void SendReject(NODE* nd)
{
	PACKET_FOR_SEND pack;
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.buffer = &CMD_REJECT;
	pack.buf_len = sizeof(CMD_REJECT);
	inst->net->send_packet(&pack);
}
//----------------------------------------------------------------------------
void parsePacket(NET_PACKET* pac)
{
	u32 i;
    switch(pac->buffer[0])
    {
    case NETWORK_PROT_KADEMLIA:
		if (inst->kad)
		{
			inst->kad->packetReceived(&pac->buffer[1],pac->buf_len-sizeof(u32),pac->nd,inst->usr);
		}
        break;
    case NETWORK_PROT_CONTROL:
		if ((pac->buf_len >= 2 * sizeof(u32)) && !pac->buffer[1]) // connection request
		{
			SendMessage(inst->hMainWnd, WM_REQUEST_CONNECTION, 0, (LPARAM)pac->nd);
			break;
		}
		EnterCriticalSection(&inst->chat_cs);
		for (i = 0; i < inst->chats_num; i++)
		{
			if (inst->chats[i].hWnd && (isIdIdentical(pac->nd->id,inst->chats[i].nd.id) || !isNonZeroID(inst->chats[i].nd.id))) break;
		}
		if (i >= inst->chats_num) // create new chat
		{
			LeaveCriticalSection(&inst->chat_cs);
			// have to create new window in appropriate thread!!!
			SendMessage(inst->hMainWnd,WM_CREATECHAT,0,(LPARAM)pac->nd);
			break;
		}
		else
			LeaveCriticalSection(&inst->chat_cs);
		if (!isNonZeroID(inst->chats[i].nd.id))
		{
			memcpy(inst->chats[i].nd.id,pac->nd->id,NETWORK_ID_LENGTH*sizeof(u32));
			SendMessage(inst->chats[i].hWnd,WM_SETID,0,(LPARAM)&pac->nd->id);
		}
		if (IsWindow(inst->chats[i].hWnd) && inst->chats[i].ch) inst->chats[i].ch->contr->received(&pac->buffer[1],pac->buf_len-sizeof(u32));
		//SendMessage(inst->chats[i].hWnd,WM_PROCESSDATA,pac->buf_len-sizeof(u32),(LPARAM)&pac->buffer[1]);
        break;
	case NETWORK_PROT_FILE_TRANSFER:
		inst->ft->parse_packet(&pac->buffer[1], pac->buf_len - sizeof(u32), pac->nd, inst->set->acceptFT);
		break;
	default:
	    // another protocols here
		break;
	}
}

//----------------------------------------------------------------------------
void RequestConnection(NODE* nd)
{
	u32 cmd = 0;
	PACKET_FOR_SEND pack;
	pack.buffer = &cmd;
	pack.buf_len = sizeof(u32);
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_CONTROL;
	getNet()->send_packet(&pack);
}
//----------------------------------------------------------------------------
bool isTrustUser(u32* id)
{
	like_t* ptr;
	for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		if (isIdIdentical(id,ptr->nd.id)) return ptr->trust;
	}
	return false;
}
//----------------------------------------------------------------------------
void SetTrustUser(HWND hWnd, u32* id, bool trust)
{
	like_t* ptr;
	for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		if (isIdIdentical(id, ptr->nd.id))
		{
			settings_t* set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
			if (trust)
			{
				TCHAR id_str[100];
				TCHAR ip_str[20];
				TCHAR msg[1000];
				id2str(ptr->nd.id, id_str);
				ip2str(ptr->nd.ip, ip_str);
				_stprintf_s(msg, 1000, TEXT("Do you want to authorize following user:\nName: %s\nOriginal name: %s\nID: %s\nIP: %s\nPort: %d"), ptr->name, ptr->name_orig, id_str, ip_str, ptr->nd.port);
				if (IDYES != MessageBox(hWnd, msg, TEXT("Question"), MB_YESNO | MB_ICONQUESTION))
				{
					break;
				}
			}
			ptr->trust = trust;
			inst->usr->save_likes(set->likesFileName);
			CloseHandle(CreateThread(NULL,0,SignatureThread,set,0,NULL));
			break;
		}
	}
}
//----------------------------------------------------------------------------
void appendLike(NODE* nd)
{
	settings_t* set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	TCHAR UserName[200];
	inst->net->get_user_info(nd, UserName);
	inst->usr->append_like(nd, UserName, UserName, false);
	inst->usr->save_likes(set->likesFileName);
	CloseHandle(CreateThread(NULL,0,SignatureThread,set,0,NULL));
}
//----------------------------------------------------------------------------
void removeLike(u32* id)
{
	settings_t* set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	inst->usr->remove_like(id);
	inst->usr->save_likes(set->likesFileName);
	CloseHandle(CreateThread(NULL,0,SignatureThread,set,0,NULL));
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK RenameDialogProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	like_t* ptr;

	switch (uMsg)
	{
	case WM_INITDIALOG:
		ptr = (like_t*)lParam;
		SetProp(hWnd, TEXT("like"), ptr);
		if (ptr->name_orig) SetDlgItemText(hWnd, IDC_USER_ORIG_NAME, ptr->name_orig);
		if (ptr->name) SetDlgItemText(hWnd, IDC_USER_NAME, ptr->name);
		SetFocus(GetDlgItem(hWnd, IDC_USER_NAME));
		return TRUE;
	case WM_DESTROY:
		RemoveProp(hWnd, TEXT("like"));
		return 0;
	case WM_CLOSE:
		EndDialog(hWnd, 0);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			{
				int len = GetWindowTextLength(GetDlgItem(hWnd, IDC_USER_NAME));
				ptr = (like_t*)GetProp(hWnd, TEXT("like"));
				if (ptr->name) delete[] ptr->name;
				ptr->name = new tchar_t[len + 1];
				GetDlgItemText(hWnd, IDC_USER_NAME, ptr->name, len+1);
				EndDialog(hWnd, 0);
				return TRUE;
			}
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
void renameLike(HWND hWnd, u32* id)
{
	like_t* ptr;
	for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		if (isIdIdentical(id, ptr->nd.id))
		{
			DialogBoxParam(GetModuleHandle(NULL), TEXT("RenameDialog"), hWnd, RenameDialogProc, (LPARAM)ptr);
			settings_t* set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
			inst->usr->save_likes(set->likesFileName);
			CloseHandle(CreateThread(NULL,0,SignatureThread,set,0,NULL));
			break;
		}
	}
}
//----------------------------------------------------------------------------
static void updateTable(HWND hWnd)
{
	static int min_cnt = 20;
	like_t* ptr;
	TCHAR str[100];
	HWND hUsers = GetDlgItem(hWnd,IDC_USERS);
    int cnt = 0;
	int cur_sel = ListView_GetSelectionMark(hUsers);
	int num = ListView_GetItemCount(hUsers);
	u32 sel_id[NETWORK_ID_LENGTH];
	if (!ListView_GetSelectedCount(hUsers))
	{
		cur_sel = -1;
	}
	if (cur_sel >= 0)
	{
		ListView_GetItemText(hUsers,cur_sel,4,str,100);
		str2id(str,sel_id);
	}
	for(ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		LVITEM item = {0};

		inst->net->get_ip(&ptr->nd);
			
		item.mask = LVIF_TEXT|LVIF_IMAGE;
		if (ptr->name) item.pszText = ptr->name;
		else  item.pszText = ptr->name_orig;
		if (isTrustUser(ptr->nd.id))
		{
			u32 status = inst->net->get_user_status(ptr->nd.id);
			switch(status)
			{
			case USER_STATUS_ONLINE: item.iImage = 0; break;
			case USER_STATUS_AWAY: item.iImage = 3; break;
			default: item.iImage = 1; break;
			}
		}
		else
		{
			item.iImage = 2;
		}
		item.iItem = cnt++;
		if (num < cnt) item.iItem = ListView_InsertItem(hUsers,&item);
		else ListView_SetItem(hUsers,&item);
		item.mask = LVIF_TEXT;
		item.iSubItem = 1;
		item.pszText = str;
		ip2str(ptr->nd.ip,str);
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 2;
		_stprintf_s(str,100,TEXT("%d"),ptr->nd.port);
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 4;
		id2str(ptr->nd.id,str);
		ListView_SetItem(hUsers,&item);

		if ((cur_sel >= 0) && isIdIdentical(sel_id,ptr->nd.id))
		{
			ListView_SetItemState(hUsers,item.iItem,LVIS_SELECTED,LVIS_SELECTED);
			ListView_SetSelectionMark(hUsers,item.iItem);
		}
		if (!min_cnt)
		{
			inst->kad->pingNode(&ptr->nd);
		}
	}
	// clear all other items if exist
	while(cnt < (num = ListView_GetItemCount(hUsers)))
	{
		ListView_DeleteItem(hUsers,num-1);
	}
	ListView_RedrawItems(hUsers,0,cnt);
	if (!min_cnt) min_cnt = 30;
	else min_cnt--;
	u32 away_time = ((settings_t*)GetProp(hWnd,TEXT("set")))->away_timeout;
	if (away_time && (inst->state != STATE_AWAY_MANUAL))
	{
		LASTINPUTINFO lii;
		lii.cbSize = sizeof(LASTINPUTINFO);
		GetLastInputInfo(&lii);
		DWORD diff = (GetTickCount() - lii.dwTime)/60000;
		if ((inst->state == STATE_ONLINE) && (diff >= away_time))
		{
			for (cnt = 0; cnt < (int)inst->chats_num; cnt++)
			{
				if (IsWindow(inst->chats[cnt].hWnd) && SendMessage(inst->chats[cnt].hWnd,WM_ISTALKING,0,0))
					return;
			}
			inst->state = STATE_AWAY;
			inst->net->set_away(true);
			SetWindowText(hWnd,TEXT("ZAS Telephone - away"));
		}
		else if ((inst->state == STATE_AWAY) && (diff < away_time))
		{
			inst->state = STATE_ONLINE;
			inst->net->set_away(false);
			SetWindowText(hWnd,TEXT("ZAS Telephone - online"));
		}
	}
}
//----------------------------------------------------------------------------
VOID TrayMenu(HWND hWnd)
{
	HMENU menu = CreatePopupMenu();
	MENUITEMINFO item;
	POINT pt;
	UINT cmd;
	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_TYPE|MIIM_ID;
	item.fType = MFT_STRING;
	item.wID = IDM_ACTIVATE;
	item.dwTypeData = TEXT("ZAS Telephone");
	InsertMenuItem(menu, 0, TRUE, &item);
	item.wID = IDM_ABOUT;
	item.dwTypeData = TEXT("About");
	InsertMenuItem(menu,1,TRUE,&item);
	item.wID = IDM_EXIT;
	item.dwTypeData = TEXT("Exit");
	InsertMenuItem(menu, 2, TRUE, &item);
	GetCursorPos(&pt);
	cmd = TrackPopupMenu(menu,TPM_RETURNCMD,pt.x,pt.y,0,hWnd,NULL);
	switch (cmd)
	{
	case IDM_ACTIVATE:
		ShowWindow(hWnd, SW_NORMAL);
		SetForegroundWindow(hWnd);
		break;
	default:
		PostMessage(hWnd,WM_COMMAND,cmd,0);
		break;
	}
}
//----------------------------------------------------------------------------
static void userMenu(HWND hWnd,LPNMITEMACTIVATE ntf)
{
	TCHAR str[100];
	HMENU menu = CreatePopupMenu();
	MENUITEMINFO item;
	NODE nd;
	settings_t* set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));

	ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),ntf->iItem,4,str,100);
	if (!str2id(str,nd.id)) return;
	ListView_GetItemText(GetDlgItem(hWnd, IDC_USERS), ntf->iItem, 1, str, 100);
	str2ip(str, &nd.ip);
	ListView_GetItemText(GetDlgItem(hWnd, IDC_USERS), ntf->iItem, 2, str, 100);
	nd.port = (u16)wcstol(str, NULL, 10);

	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_TYPE|MIIM_ID;
	item.fType = MFT_STRING;
	if (!isTrustUser(nd.id))
	{
		item.wID = IDM_ADDTRUST;
		item.dwTypeData = TEXT("Authorize");
		InsertMenuItem(menu, 0, TRUE, &item);
	}
	else
	{
		item.wID = IDM_DEAUTH;
		item.dwTypeData = TEXT("Deauthorize");
		InsertMenuItem(menu, 0, TRUE, &item);
	}
	item.wID = IDM_REMOVESAVED;
	item.dwTypeData = TEXT("Remove from list");
	InsertMenuItem(menu,1,TRUE,&item);
	item.wID = IDM_REQUEST_CONNECTION;
	item.dwTypeData = TEXT("Request Connection");
	InsertMenuItem(menu, 2, TRUE, &item);
	item.wID = IDM_RENAME;
	item.dwTypeData = TEXT("Rename");
	InsertMenuItem(menu, 3, TRUE, &item);
	item.wID = IDM_TRANSFERFILE;
	item.dwTypeData = TEXT("Send file");
	InsertMenuItem(menu, 4, TRUE, &item);
	item.wID = IDM_INFO;
	item.dwTypeData = TEXT("Info");
	InsertMenuItem(menu, 5, TRUE, &item);
	ClientToScreen(GetDlgItem(hWnd, IDC_USERS), &ntf->ptAction);
	switch (TrackPopupMenu(menu,TPM_RETURNCMD,ntf->ptAction.x,ntf->ptAction.y,0,hWnd,NULL))
	{
	case IDM_ADDTRUST:
		SetTrustUser(hWnd,nd.id,true);
		break;
	case IDM_DEAUTH:
		SetTrustUser(hWnd,nd.id,false);
		break;
	case IDM_REMOVESAVED:
		inst->usr->remove_like(nd.id);
		inst->usr->save_likes(set->likesFileName);
		CloseHandle(CreateThread(NULL,0,SignatureThread,set,0,NULL));
		break;
	case IDM_RENAME:
		renameLike(hWnd,nd.id);
		break;
	case IDM_TRANSFERFILE:
		FileTransfer(hWnd,&nd);
		break;
	case IDM_REQUEST_CONNECTION:
		RequestConnection(&nd);
		break;
	case IDM_INFO:
		ShowUserInfo(hWnd,&nd);
		break;
	}
}
//----------------------------------------------------------------------------
static void selectUser(HWND hWnd, u32* id)
{
	TCHAR str[100];
	u32 item_id[NETWORK_ID_LENGTH];
	for (int i = 0; i < ListView_GetItemCount(GetDlgItem(hWnd,IDC_USERS)); i++)
	{
		ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),i,4,str,100);
		if (!str2id(str,item_id)) return;
		if (isIdIdentical(id,item_id))
		{
			ListView_SetItemState(GetDlgItem(hWnd,IDC_USERS),i,LVIS_SELECTED,LVIS_SELECTED);
			ListView_SetSelectionMark(GetDlgItem(hWnd,IDC_USERS),i);
			break;
		}
	}
}
//----------------------------------------------------------------------------
bool isInLikes(const NODE* nd)
{
	like_t* ptr;
	for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		if (isIdIdentical(nd->id, ptr->nd.id))
		{
			return true;
		}
	}
	return false;
}
//----------------------------------------------------------------------------
void OpenChat(NODE* nd)
{
	u32 i;
	settings_t* set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	if (!set->allowAnyone && !isInLikes(nd))
	{
		SendReject(nd);
		return;
	}
	// try to find same window
	EnterCriticalSection(&inst->chat_cs);
	for (i = 0; i < inst->chats_num; i++)
	{
		if (inst->chats[i].hWnd && isIdIdentical(nd->id,inst->chats[i].nd.id))
		{
			SetForegroundWindow(inst->chats[i].hWnd);
			LeaveCriticalSection(&inst->chat_cs);
			return;
		}
	}
	// try to find empty space
	for (i = 0; i < inst->chats_num; i++)
	{
		if (!inst->chats[i].hWnd) break;
	}
	if (i >= inst->chats_num)
	{
		// allocate space
		inst->chats_num += 20;
		chatwnd_t* ch = (chatwnd_t*)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,inst->chats_num*sizeof(chatwnd_t));
		if (ch && inst->chats)
		{
			CopyMemory(ch, inst->chats, (inst->chats_num-20)*sizeof(chatwnd_t));
		}
		if (inst->chats) HeapFree(GetProcessHeap(),0,inst->chats);
		inst->chats = ch;
	}
	LeaveCriticalSection(&inst->chat_cs);
	if (!isNonZeroID(nd->id) || inst->net->findNode(nd->id,nd) || inst->kad->findNode(nd->id,nd))
	{
		inst->chats[i].nd = *nd;
		inst->chats[i].hWnd = StartChat(inst->hMainWnd,nd);
		inst->chats[i].ch = (chat_t*)GetProp(inst->chats[i].hWnd,TEXT("chat"));
	}
}
//----------------------------------------------------------------------------
static void ActivateChat(HWND hWnd, int iItem)
{
	TCHAR str[100];
	NODE nd;
	ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),iItem,4,str,100);
	if (str2id(str,nd.id))
	{
		OpenChat(&nd);
	}
}
//----------------------------------------------------------------------------
static void DeleteChat(HWND hWnd, HWND chat)
{
	u32 i;
	EnterCriticalSection(&inst->chat_cs);
	for (i = 0; i < inst->chats_num; i++)
	{
		if (inst->chats[i].hWnd == chat)
		{
			inst->chats[i].hWnd = NULL;
			break;
		}
	}
	LeaveCriticalSection(&inst->chat_cs);
}
//----------------------------------------------------------------------------
static void ShowAbout(HWND hWnd)
{
const LPTSTR about = TEXT("ZAS Telephone ") TEXT(VER_TYPE) TEXT(" ") TEXT(VER_NUM) TEXT(" ") TEXT(BUILD_NUM)
#ifdef _WIN64
					TEXT(" 64-Bit ")
#else
					TEXT(" 32-Bit ")
#endif
					TEXT(DATE_NOW) TEXT("\r\n\r\n")
					TEXT("To all Freedom loving people of the World. ")
					TEXT("We do our best to provide you private communication in the upcoming darkness of the Orwellian age. ")
					TEXT("The software binaries are free and always be free for any use. The source code is included only for review purpose. ")
					TEXT("You are not allowed to modify any part of this software or source code or use it for your products unless we give you explicit permission.\r\n\r\n")
					TEXT("The software is provided as is. There are no warranties or liabilities of any kind.\r\n\r\n")
					TEXT("Report bugs and suggestions to:\r\n\r\n")
					TEXT("ZAS Telephone Team\r\n")
					TEXT("zas@zas-comm.ru\r\n\r\n")
					
TEXT("Key Type:  DSA/ELG\r\n")
TEXT("Algorithm: 3DES\r\n")
TEXT("\r\n")
TEXT("-----BEGIN PGP PUBLIC KEY BLOCK-----\r\n")
TEXT("\r\n")
TEXT("mQGiBFLi8QwRBADN266rAA/91Lo9LZg8s7o5iR9xaM5Pi6dKLlGCHqsNoejuESRh\r\n")
TEXT("zTbussbnKDk2Py4FKpQBZJdLvWcJoFOjJw+Tz5zkAsn9dlSwebMw1r3GF3m1gU1w\r\n")
TEXT("jRRgGWMP09FG8IRiPo8ZC0cDVziH9CaTg7hrhaQWoSYfiCfSI0iISeN7MQCg/zdY\r\n")
TEXT("l7IqjjtG1iBO/ZmXtYui04ED/2nAAmU2IW5tY3FuAyKBgwVPrfqUl8+aSGI3Enk6\r\n")
TEXT("woIchGEd1Coq557Z8DGnUEHn8yoJVrMQCiYlro+BAqKOUn+Vm2kugYfcV8yDXLEl\r\n")
TEXT("rmW/VAksZ5VLZ+BCqf0pJI/MR9PINBuN7tQIHp7OmxQCx6TG2IhVYsrsXfvtY5iP\r\n")
TEXT("CAtwBACg2ZKPPpvUrllloXhFFEXQ7kLv6BzCGDoevEk7AtYM9eNrp590EtB4Bl6J\r\n")
TEXT("2A0TFh2MsEYDoCTc7EI1Cj5EoPj2biEBBDJ/DUQOHR58fHsz90LHZgo3fYmfc0Qu\r\n")
TEXT("kXQ1I5GCsBqrm7VNGF7e8jOt5hpMFGQMgkPkoUV/8Lj2iTdh1rQVWkFTIDx6YXNA\r\n")
TEXT("emFzLWNvbW0ucnU+iQBLBBARAgALBQJS4vEMBAsCAwEACgkQK4UXmS+niMG6SgCd\r\n")
TEXT("GYZnTxNlLXG20OHnfnDm/CwA8pEAoKhgsL4kgYgXqEA4SsEMLR9imjUQuQQNBFLi\r\n")
TEXT("8ecQEADrITWmV9/bhDr0UU1ssWbtRkV/5E1f28DM52ZjHOOtPROteSAiMcVG0AOq\r\n")
TEXT("1JAoOXf4+2qdou7foXkUYjmFdLZlk9oOSNXP57/Ll11YOyN7e917EJ7ZPd5/AXu9\r\n")
TEXT("TbF3KrPX74Fs5SX6dxuZxGkAxDKtlx08lAFKp8yWRpLD8DnKSfH0wx29L3tblpdI\r\n")
TEXT("x5u5onpwe1SbCMWVbCvwRArUrY+EVnCvvcTTJKlAs9Qrw3Ufn2VMHwCn2rVyHMy6\r\n")
TEXT("e6N3/denVqMzoXOjBYh1h8G9mMZimQ/mFmXAls6S8HsevmFYbXpkXv6g1+0z9Uap\r\n")
TEXT("TpoH5b65OJXRvYKupUwXkJe2KP/CLVGOREUiYisJofP/at/m5+kaPFcXw1ylz1kH\r\n")
TEXT("1HCDerQ+ka4TPILsiK3xhX3FQHHd3xIJHhE10B9mt5XVV9RcOKB5Cx6XYOSoQq3P\r\n")
TEXT("S8XJpcvHoQLa6j/4fVXgr20pGR+vx95bXD178UM7+debFpwW/lAis5RTt9sj0rFH\r\n")
TEXT("45W7SFkIgf70tPnYW8LiNpl8x+c394Eu9PZfcji2ymatD+VXga6mDrnioquAehKZ\r\n")
TEXT("C13TRKJMsbJd6/596bJQpATZ4EQ5VK7g43JFp/gMk8fywn8QsqcowocQZgPdfSZ7\r\n")
TEXT("L+wf0Z3j/aqqwINWRBw1FZoTReSltN2fMHk0xAQA9ZI6Cm13fwACAhAAn41qFqO6\r\n")
TEXT("lkfR91hQejaS5n5bq1+tAz+PcaQFnyD7+CQ/zvkhK+i2M3ahGP7J5MPkaBRbFdxT\r\n")
TEXT("PvOtaW9xliNP1vh8X5Fn1iNo+5KkohXzFOUevRF61lMR4KeBX46FAUH4LXxGJU7c\r\n")
TEXT("npJAbzKZvY+khX+Rwdmy28j12ryjaAU6s7cpHokjh2wRmy030yxrIrKNzHQWgSYX\r\n")
TEXT("5460QGoXQM9qps+Bxm0j9NHZyKrICAxJGY06JLUVy6FdMsUstT2bfTHpmekiOMQl\r\n")
TEXT("q889lXIpu9TOs+T9aKrCWi9T7TiV5M7oGqux6DNS8OqAihZVkaHA5HczyBqjVA9y\r\n")
TEXT("zA8svc1KvjTO+lFrofmH/cnfwZUazqlDXHOPi+psJ915LPxr1pSDAPCvB/qQT2td\r\n")
TEXT("o1dE6j+CrrmXJMoU/gmoFmLmDlFDNgvwCBDSpCFTYFpGd2rPHyqAUiBd5iiH9wZB\r\n")
TEXT("ge87nzrrdA0jTxop6WJ1TfF0OQW26vmMmeU6FcM3pk9GqRqDkam8HqjxIh5929v7\r\n")
TEXT("/41HqeYGn6zmGj/Oj0d0QKj6mgUSin+LotZxEGGPQU5ZgV3QVKyFoWaZ+Em5MX6u\r\n")
TEXT("kLwZyNr5O3CJuvJYvBbsfQaIF/YlL6V1H5ilDG+naqulatL7fXqQ/3r6F3IoDiAa\r\n")
TEXT("CppVKRxXNKloBfjMRryanTZXMR/8LW/wsISJAEUEGBECAAYFAlLi8ecACgkQK4UX\r\n")
TEXT("mS+niMGwWwCcC3rEZ5jNa1uPZD1iFZXGhPGpINIAmOV6u6q3zguMl+1wEdddLmHt\r\n")
TEXT("hEk=\r\n")
TEXT("=1JLW\r\n")
TEXT("-----END PGP PUBLIC KEY BLOCK-----\r\n");

	HIMAGELIST IL = ImageList_Create(128,128,ILC_COLOR32,1,1);
	ImageList_AddIcon(IL,LoadIcon(GetModuleHandle(NULL),MAKEINTRESOURCE(IDI_ABOUT)));
	SetProp(hWnd,TEXT("image"),IL);

	SetDlgItemText(hWnd,IDC_ABOUT_TEXT,about);

	SetFocus(GetDlgItem(hWnd, IDOK));
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK AboutDialog(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		ShowAbout(hWnd);
		return FALSE;
    case WM_DESTROY:
		ImageList_Destroy((HIMAGELIST)RemoveProp(hWnd,TEXT("image")));
		return 0;
    case WM_CLOSE:
		EndDialog(hWnd,0);
		return TRUE;

    case WM_PAINT:
	{
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(hWnd, &ps);
		ImageList_Draw((HIMAGELIST)GetProp(hWnd,TEXT("image")),0,hdc,2,40,ILD_NORMAL);
        EndPaint(hWnd, &ps);
		return FALSE;
	}
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			EndDialog(hWnd,0);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
static void MainWndSizing(HWND hWnd, WPARAM wParam, LPRECT newWndRect)
{
	RECT oldWndRect;
	RECT oldClientRect;
	const LONG lMinWidth = 200;
	const LONG lMinHeight = 300;
	//
	GetWindowRect(hWnd, &oldWndRect);
	GetClientRect(hWnd, &oldClientRect);
	//
	oldClientRect.right -= oldWndRect.right - oldWndRect.left - (newWndRect->right - newWndRect->left);		// calculate new client width
	oldClientRect.bottom -= oldWndRect.bottom - oldWndRect.top - (newWndRect->bottom - newWndRect->top);	// calculate new client height
	// adjust width
	if (oldClientRect.right < lMinWidth){
		switch (wParam)
		{
		case WMSZ_BOTTOMLEFT:
		case WMSZ_LEFT:
		case WMSZ_TOPLEFT:
			newWndRect->left -= lMinWidth - oldClientRect.right;
			break;
			//
		default:
			newWndRect->right += lMinWidth - oldClientRect.right;
		}
	}
	// adjust height
	if (oldClientRect.bottom < lMinHeight){
		switch (wParam)
		{
		case WMSZ_TOPLEFT:
		case WMSZ_TOP:
		case WMSZ_TOPRIGHT:
			newWndRect->top -= lMinHeight - oldClientRect.bottom;
			break;
			//
		default:
			newWndRect->bottom += lMinHeight - oldClientRect.bottom;
		}
	}
}
//----------------------------------------------------------------------------
static void RequestConnectionInit(HWND hWnd, NODE* nd)
{
	TCHAR msg[1000];
	TCHAR id_str[100];
	TCHAR ip_str[20];
	TCHAR UserName[200];
	NODE* node = (NODE*)LocalAlloc(LPTR, sizeof(NODE));
	memcpy(node, nd, sizeof(NODE));
	SetProp(hWnd, TEXT("node"), node);
	id2str(nd->id, id_str);
	ip2str(nd->ip, ip_str);
	inst->net->get_user_info(nd, UserName);
	_stprintf_s(msg, 1000, TEXT("Do you accept connection request from user:\nName: %s\nID: %s\nIP: %s\nPort: %d"), UserName, id_str, ip_str, nd->port);
	SetDlgItemText(hWnd, IDC_USER_NAME, msg);
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK ConnectTo(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	NODE* nd = (NODE*)GetProp(hWnd,TEXT("nd"));
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetProp(hWnd,TEXT("nd"),(HANDLE)lParam);
		return TRUE;
	case WM_DESTROY:
		SendDlgItemMessage(hWnd,IDC_IP,IPM_GETADDRESS,0,(LPARAM)&nd->ip);
		nd->port = GetDlgItemInt(hWnd,IDC_PORT,NULL,FALSE);
		return 0;
	case WM_CLOSE:
		EndDialog(hWnd,0);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			EndDialog(hWnd,0);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
static void connectto(HWND hWnd)
{
	NODE nd = {0};
	DialogBoxParam(GetModuleHandle(NULL),TEXT("ConnectTo"),hWnd,ConnectTo,(LPARAM)&nd);
	if (nd.ip) OpenChat(&nd);
}
//----------------------------------------------------------------------------
void GetMyName(LPTSTR name)
{
	if (inst->sdata.data_is_initialized)
	{
		memcpy(name,inst->sdata.certificate.user_info,inst->sdata.certificate.user_info_length_bytes);
		name[inst->sdata.certificate.user_info_length_bytes/sizeof(TCHAR)] = 0;
	}
	else
	{
		name[0] = 0;
	}
}
//----------------------------------------------------------------------------
typedef struct {
	TCHAR name[user_info_max_length_bytes/sizeof(TCHAR) + 1];
	bool show_orig;
	TCHAR orig[user_info_max_length_bytes/sizeof(TCHAR) + 1];
	NODE nd;
	bool auth;
	int status;
} info_t;
//----------------------------------------------------------------------------
static const LPTSTR user_status[USER_STATUS_MAX] = {TEXT("OFFLINE"),TEXT("ONLINE"),TEXT("AWAY")};
static INT_PTR CALLBACK ShowInfo(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
	{
		info_t* info = (info_t*)lParam;
		TCHAR str[100];
		SetDlgItemText(hWnd,IDC_USER_NAME,info->name);
		if (info->show_orig) SetDlgItemText(hWnd,IDC_USER_ORIG_NAME,info->orig);
		else SetDlgItemText(hWnd,IDC_USER_ORIG_NAME,info->name);
		id2str(info->nd.id,str);
		SetDlgItemText(hWnd,IDC_USER_ID,str);
		ip2str(info->nd.ip,str);
		_stprintf_s(&str[_tcslen(str)],50,TEXT(" : %d"),info->nd.port);
		SetDlgItemText(hWnd,IDC_USER_IP,str);
		SetDlgItemText(hWnd,IDC_AUTH,info->auth?TEXT("YES"):TEXT("NO"));
		if (info->status <= USER_STATUS_MAX) SetDlgItemText(hWnd,IDC_STATUS,user_status[info->status]);
		else SetDlgItemText(hWnd,IDC_STATUS,TEXT("WRONG STATUS"));
		return TRUE;
	}
	case WM_DESTROY:
		inst->hInfo = NULL;
		return 0;
	case WM_CLOSE:
		DestroyWindow(hWnd);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			DestroyWindow(hWnd);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
static void ShowMyInfo(HWND hWnd)
{
	info_t info;
	info.show_orig = false;
	info.auth = true;
	info.status = USER_STATUS_ONLINE;
	if (inst->sdata.data_is_initialized)
	{
		GetMyName(info.name);
	}
	else
	{
		cp_strcchcpy(info.name,user_info_max_length_bytes/sizeof(TCHAR) + 1,_T("<unknown>"));
	}
	memcpy(info.nd.id,inst->sdata.certificate.fingerprint,fingerprint_length_x32*sizeof(u32));
	info.nd.port = ((settings_t*)GetProp(hWnd,TEXT("set")))->remotePort;
	info.nd.ip = getExternIp();
	if (inst->hInfo) DestroyWindow(inst->hInfo);
	inst->hInfo = CreateDialogParam(GetModuleHandle(NULL),TEXT("Info"),hWnd,ShowInfo,(LPARAM)&info);
}
//----------------------------------------------------------------------------
void ShowUserInfo(HWND hWnd, NODE* nd)
{
	info_t info = {0};
	like_t* ptr;
	info.nd = *nd;
	for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		if (isIdIdentical(nd->id, ptr->nd.id))
		{
			info.show_orig = true;
			_tcscpy_s(info.name,ptr->name);
			_tcscpy_s(info.orig,ptr->name_orig);
			info.auth = ptr->trust;
			break;
		}
	}
	if (!ptr)
	{
		getNet()->get_user_info(nd,info.name);
	}
	info.status = getNet()->get_user_status(nd->id);
	if (inst->hInfo) DestroyWindow(inst->hInfo);
	inst->hInfo = CreateDialogParam(GetModuleHandle(NULL),TEXT("Info"),hWnd,ShowInfo,(LPARAM)&info);
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK RequestConnection(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		RequestConnectionInit(hWnd, (NODE*)lParam);
		return TRUE;
	case WM_DESTROY:
		LocalFree(GetProp(hWnd, TEXT("node")));
		return 0;
	case WM_CLOSE:
		DestroyWindow(hWnd);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			appendLike((NODE*)GetProp(hWnd, TEXT("node")));
			DestroyWindow(hWnd);
			return TRUE;
		case IDCANCEL:
			DestroyWindow(hWnd);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
static void MainWndSize(HWND hWnd)
{
	RECT rc;
	GetClientRect(hWnd,&rc);

	MoveWindow(GetDlgItem(hWnd,IDC_USERS),0,0,rc.right,rc.bottom,TRUE);
}
//----------------------------------------------------------------------------
static void MainWndInit(HWND hWnd)
{
	LVITEM item = {0};
	LVCOLUMN col;
	HWND hSplash;
	HWND hUsers = CreateWindowEx(WS_EX_CLIENTEDGE, WC_LISTVIEW, TEXT(""), WS_CHILD | WS_VISIBLE | LVS_LIST | LVS_SINGLESEL | LVS_SORTASCENDING, 0, 0, 10, 20, hWnd, (HMENU)IDC_USERS, GetModuleHandle(NULL), NULL);
	ListView_SetExtendedListViewStyle(hUsers, LVS_EX_FULLROWSELECT | LVS_EX_AUTOSIZECOLUMNS | LVS_EX_GRIDLINES);
	// create columns
	col.mask = LVCF_WIDTH|LVCF_TEXT;
	col.pszText = TEXT("Name");
	col.cx = 100;
	ListView_InsertColumn(hUsers,0,&col);
	col.pszText = TEXT("IP");
	ListView_InsertColumn(hUsers,1,&col);
	col.pszText = TEXT("Port");
	ListView_InsertColumn(hUsers,2,&col);
	col.pszText = TEXT("Status");
	ListView_InsertColumn(hUsers,3,&col);
	col.pszText = TEXT("ID");
	col.cx = 500;
	ListView_InsertColumn(hUsers,4,&col);
	// add image list
	HIMAGELIST hLarge = ImageList_Create(GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), ILC_COLOR32, 4, 1);
	HIMAGELIST hSmall = ImageList_Create(GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), ILC_COLOR32, 4, 1);
	HICON hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_ONLINE));
    ImageList_AddIcon(hLarge, hiconItem);
    ImageList_AddIcon(hSmall, hiconItem);
    DestroyIcon(hiconItem);
	hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_OFFLINE));
	ImageList_AddIcon(hLarge, hiconItem);
    ImageList_AddIcon(hSmall, hiconItem);
    DestroyIcon(hiconItem);
	hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_UNAUTH));
	ImageList_AddIcon(hLarge, hiconItem);
	ImageList_AddIcon(hSmall, hiconItem);
	DestroyIcon(hiconItem);
	hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_AWAY));
	ImageList_AddIcon(hLarge, hiconItem);
    ImageList_AddIcon(hSmall, hiconItem);
    DestroyIcon(hiconItem);
	ListView_SetImageList(hUsers, hLarge, LVSIL_NORMAL);
	ListView_SetImageList(hUsers,hSmall,LVSIL_SMALL);
	
	//instance
	inst = (instance_t*)VirtualAlloc(NULL,sizeof(instance_t),MEM_RESERVE|MEM_COMMIT,PAGE_READWRITE);
	if (!VirtualLock(inst,sizeof(instance_t))) // For security reason this memory must not be written to the pagefile
	{
		MessageBox(hWnd,TEXT("Can't lock secure data memory"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
	}
	inst->hMainWnd = hWnd;
	InitializeCriticalSection(&inst->cs);
	InitializeCriticalSection(&inst->chat_cs);
	InitializeCriticalSection(&inst->sig_cs);
	// settings
	settings_t *set = (settings_t*)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(settings_t));
	SetProp(hWnd,TEXT("set"),set);
	inst->set = set;
	LoadSettings(hWnd,set);
	CreateDialogParam(GetModuleHandle(NULL),TEXT("Log"),hWnd,LogProc,(LPARAM)set);
	if (set->showSplash)
	{
		hSplash = CreateDialog(GetModuleHandle(NULL),TEXT("About"),hWnd,AboutDialog);
	}
	else
	{
		ShowWindow(hWnd,SW_NORMAL);
	}

	LoadKey(hWnd,set->keyFileName,&inst->sdata);

#ifndef _DEBUG
	TCHAR id_str[ID_LENGTH_CCH];
	id2str(inst->sdata.certificate.fingerprint,id_str);
	// check that another copy of ZAS is running
	gMutex = CreateMutex(NULL, TRUE, id_str);
	if (!gMutex)
	{
		MessageBox(NULL,TEXT("Can't create mutex"),TEXT("Fatal Error"),MB_OK|MB_ICONERROR);
		DestroyWindow(hWnd);
	}
	if (GetLastError() == ERROR_ALREADY_EXISTS)
	{
		MessageBox(NULL,TEXT("Another copy of ZAS telephone with the same ID is running. This applicaton will be closed and anoter shown"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
		HWND wnd = FindWindow(TEXT("MainWindowClass"),TEXT("ZAS Telephone"));
		if (wnd)
		{
			ShowWindow(wnd,SW_NORMAL);
			SetForegroundWindow(wnd);
		}
		DestroyWindow(hWnd);
	}
#endif
	if (set->enableUPnP)
	{
		if (!setupUPnP(set,0))
		{
			MessageBox(hWnd,TEXT("Can't setup UPnP"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
		}
	}

	NODE me;
	CopyMemory(me.id,inst->sdata.certificate.fingerprint,NETWORK_ID_LENGTH*sizeof(u32));
	me.ip = MAKEIPADDRESS(127,0,0,1);
	me.port = set->localPort;
	inst->kad = new KADEMLIA(&me,hWnd);
	inst->net = new NETWORK(&me,hWnd);
	inst->usr = new USERS();
	inst->ft = new FILE_TRANSFER(hWnd);
	SetTimer(hWnd, 1, 1000, NULL);
	inst->eStopThreadPassword = CreateEvent(NULL,TRUE,FALSE,NULL);
	inst->thPassword = CreateThread(NULL,0,PasswordThread,inst,0,NULL);
	SetWaveParam(set);
	inst->usr->load_nodes(inst->kad,set->userFileName);
	if (inst->usr->load_likes(set->likesFileName) && !verify_file_signature(set->likesFileName))
	{
		MessageBox(hWnd,TEXT("Wrong user list signature. All authorizations will be ignored"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
		inst->usr->unauthorize_likes();
	}
	inst->usr->load_downloads(set->downloadsFileName);
	initialize_seed(set->seedFileName);
	inst->state = STATE_ONLINE;
	CheckMenuRadioItem(GetMenu(hWnd),IDM_ONLINE,IDM_AWAY,IDM_ONLINE,MF_BYCOMMAND);
	CreateListsDialog(hWnd);
	if (set->showSplash)
	{
		DestroyWindow(hSplash);
	}
	//ShowWindow(g_logWnd,SW_NORMAL);
	//ShowWindow(g_NetList,SW_NORMAL);
	//ShowWindow(g_KadList,SW_NORMAL);
}
//----------------------------------------------------------------------------
static void MainWndDestroy(HWND hWnd)
{
	settings_t *set = (settings_t*)RemoveProp(hWnd,TEXT("set"));
	DestroyWindow(g_NetList);
	DestroyWindow(g_KadList);
	KillTimer(hWnd,1);
	SetEvent(inst->eStopThreadPassword);
	WaitForSingleObject(inst->thPassword,1000);
	CloseHandle(inst->eStopThreadPassword);
	CloseHandle(inst->thPassword);
	inst->usr->save_nodes(set->userFileName);
	if (!save_file_signature(set->likesFileName))
	{
		MessageBox(hWnd,TEXT("Can't create signature for user list"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
	}
	inst->usr->save_downloads(set->downloadsFileName);
	initialize_seed(set->seedFileName);
	if (inst->chats)
	{
		u32 i;
		for (i = 0; i < inst->chats_num; i++)
		{
			if (inst->chats[i].hWnd) DestroyWindow(inst->chats[i].hWnd);
		}
		HeapFree(GetProcessHeap(),0,inst->chats);
	}
	delete inst->ft; inst->ft = NULL;
	delete inst->kad; inst->kad = NULL;
	delete inst->net; inst->net = NULL;
	delete inst->usr; inst->usr = NULL;
	set->volume = inst->volume;
	set->mic = inst->mic;
	SaveSettings(hWnd,set);
	DeleteCriticalSection(&inst->cs);
	DeleteCriticalSection(&inst->chat_cs);
	DeleteCriticalSection(&inst->sig_cs);
	if (set->removeUPnPonExit && set->enableUPnP)
	{
		setupUPnP(set,1);
	}
	ZeroMemory(inst->password,password_max_length); // clear password
	inst->sdata.reset();
	DestroyWindow(g_logWnd);
	VirtualUnlock(inst,sizeof(instance_t));
	VirtualFree(inst,sizeof(instance_t),MEM_RELEASE);
	HeapFree(GetProcessHeap(),0,set);
}
//----------------------------------------------------------------------------
LRESULT CALLBACK MainWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_CREATE:
		MainWndInit(hWnd);
		return 0;
		//
    case WM_DESTROY:
		MainWndDestroy(hWnd);
		PostQuitMessage(0);
		return 0;
		//
	case WM_SIZE:
		MainWndSize(hWnd);
		return 0;
		//
	case WM_SIZING:
		MainWndSizing(hWnd, wParam, (LPRECT)lParam);
		return TRUE;
		//
	case WM_TIMER:
		switch (wParam)
		{
		case 1:
			updateTable(hWnd);
			return 0;
		}
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDM_EXIT:
			DestroyWindow(hWnd);
			return 0;
		case IDM_ABOUT:
			DialogBox(GetModuleHandle(NULL),TEXT("About"),hWnd,AboutDialog);
			return 0;
		case IDM_HELP:
			return 0;
		case IDM_SETTINGS:
			SettingDialog(hWnd,(settings_t*)GetProp(hWnd,TEXT("set")),inst->usr,inst->kad);
			return 0;
		case IDM_GENERATE:
			GenerateKey(hWnd,((settings_t*)GetProp(hWnd,TEXT("set")))->keyFileName);
			return 0;
		case IDM_CHANGE_PASS:
		{
			u8 password[password_max_length];
			u32 len = ChangePassword(hWnd,((settings_t*)GetProp(hWnd,TEXT("set")))->keyFileName,password);
			if (len > 0)
			{
				EnterCriticalSection(&inst->cs);
				inst->passwordLength = len;
				cheap_random_number(inst->passwordMix, (inst->passwordLength+3)/4);
				for (u32 i = 0; i < inst->passwordLength; i++)
				{
					inst->password[i] = password[i]^inst->passwordMix[i];
				}
				LeaveCriticalSection(&inst->cs);
				ZeroMemory(password,sizeof(password));
			}
			return 0;
		}
		case IDM_BOOTSTRAP:
			if (inst->net)
			{
				inst->net->reset_sessions();
			}
			if (inst->kad)
			{
				inst->kad->start();
				LogMessage(module,TEXT("Bootstrap started"));
			}
			return 0;
		case IDM_SHOW_LOG:
			ShowWindow(g_logWnd,SW_SHOWNORMAL);
			SetForegroundWindow(g_logWnd);
			return 0;
		case IDM_SHOW_NET:
			ShowWindow(g_NetList,SW_SHOWNORMAL);
			SetForegroundWindow(g_NetList);
			return 0;
		case IDM_SHOW_KAD:
			ShowWindow(g_KadList,SW_SHOWNORMAL);
			SetForegroundWindow(g_KadList);
			return 0;
		case IDM_SHOW_DWNLDS:
			ShowWindow(g_DownloadsList,SW_SHOWNORMAL);
			SetForegroundWindow(g_DownloadsList);
			PostMessage(g_DownloadsList,WM_UPDATE_PROGRESS,0,0);
			return 0;
		case IDM_FIND:
			findUser(hWnd);
			return 0;
		case IDM_MY_INFO:
			ShowMyInfo(hWnd);
			return 0;
		case IDM_ONLINE:
			inst->state = STATE_ONLINE;
			inst->net->set_away(false);
			CheckMenuRadioItem(GetMenu(hWnd),IDM_ONLINE,IDM_AWAY,IDM_ONLINE,MF_BYCOMMAND);
			SetWindowText(hWnd,TEXT("ZAS Telephone - online"));
			break;
		case IDM_AWAY:
			inst->state = STATE_AWAY_MANUAL;
			inst->net->set_away(true);
			SetWindowText(hWnd,TEXT("ZAS Telephone - away"));
			CheckMenuRadioItem(GetMenu(hWnd),IDM_ONLINE,IDM_AWAY,IDM_AWAY,MF_BYCOMMAND);
			break;
		case IDM_CONNECT_TO:
			connectto(hWnd);
			break;
		}
		return TRUE;
	case WM_NOTIFY:
	{
		LPNMITEMACTIVATE ntf = (LPNMITEMACTIVATE)lParam;
		if (ntf->hdr.idFrom == IDC_USERS)
		{
			switch (ntf->hdr.code)
			{
			case NM_DBLCLK:
				ActivateChat(hWnd,ntf->iItem);
				break;
			case NM_RCLICK:
				userMenu(hWnd,ntf);
				break;
			}
		}
		return FALSE;
	}
	case WM_SHOWTEXT:
		return 0;
	case WM_CREATECHAT:
		OpenChat((NODE*)lParam);
		return 0;
	case WM_DELETECHAT:
		DeleteChat(hWnd,(HWND)lParam);
		return 0;
	case WM_USER_FOUND:
	{
		NODE nd;
		inst->kad->findNode((u32*)lParam,&nd);
		selectUser(hWnd,(u32*)lParam);
		delete[] (u32*)lParam;
		return TRUE;
	}
	case WM_FILETRANSFER:
		CreateDialogParam(GetModuleHandle(NULL), TEXT("FileTransfer"), hWnd, TransferProc, lParam);
		return TRUE;
	case WM_REQUEST_CONNECTION:
	{
		settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
		NODE* nd = (NODE*)lParam;
		if (set->allowConnection)
		{
			CreateDialogParam(GetModuleHandle(NULL), TEXT("RequestConnection"), hWnd, RequestConnection, (LPARAM)nd);
		}
		else
		{
			SendReject(nd);
		}
		return TRUE;
	}
	case WM_SYSTEMTRAY:
		switch (LOWORD(lParam))
		{
		case WM_LBUTTONDOWN:
			ShowWindow(hWnd, SW_NORMAL);
			SetForegroundWindow(hWnd);
			break;
		case WM_RBUTTONDOWN:
			TrayMenu(hWnd);
			break;
		}
		return TRUE;
	case WM_CLOSE:
		ShowWindow(hWnd, SW_HIDE);
		return TRUE;
	case WM_POWERBROADCAST:
		if ((wParam == PBT_APMRESUMESUSPEND) || (wParam == PBT_APMRESUMECRITICAL) || (wParam == PBT_APMRESUMEAUTOMATIC))  // reset UPnP after any wake up
		{
			settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
			if (set->enableUPnP)
			{
				setupUPnP(set,0);
			}
		}
		break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}
//----------------------------------------------------------------------------
#include "../Audio/interpolator.h"
int APIENTRY WinMain(HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG AMsg;
	BOOL ret;
	WNDCLASSEX wc;
    INITCOMMONCONTROLSEX icex;
	NOTIFYICONDATA nid;

	WSADATA wsaData;
	WSAStartup(MAKEWORD(2, 2), &wsaData);

	// common controls
	icex.dwSize = sizeof(INITCOMMONCONTROLSEX);
   	icex.dwICC   = ICC_LISTVIEW_CLASSES|ICC_TAB_CLASSES|ICC_BAR_CLASSES|ICC_PROGRESS_CLASS|ICC_INTERNET_CLASSES;
   	InitCommonControlsEx(&icex);
	
	// window class
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = MainWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_MAIN));
	wc.hbrBackground = (HBRUSH)(COLOR_GRAYTEXT + 1);
	wc.hInstance = hInst;
	wc.lpszMenuName = MAKEINTRESOURCE(ID_MENU);
	wc.lpszClassName = TEXT("MainWindowClass");
	wc.hIconSm = LoadIcon(hInst, MAKEINTRESOURCE(IDI_MAIN));

	if (!RegisterClassEx(&wc)) 
		return 0;

	LoadLibrary(TEXT("Riched20.dll")); // for rich edit

	HWND hwndMain = CreateWindow(TEXT("MainWindowClass"),TEXT("ZAS Telephone - online"),WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN|WS_CLIPSIBLINGS,0,0,200,600,NULL,NULL,hInst,NULL);
	if (!hwndMain) return 0;

	ShowWindow(hwndMain,nCmdShow);

	PostMessage(hwndMain, WM_COMMAND, IDM_BOOTSTRAP, 0); // bootstrap immediately after start

	HACCEL hAccel = LoadAccelerators(hInst, MAKEINTRESOURCE(ID_ACCEL));

	ZeroMemory(&nid,sizeof(nid));
	nid.cbSize = sizeof(nid);
	nid.hWnd = hwndMain;
	nid.uFlags = NIF_ICON|NIF_TIP|NIF_MESSAGE;
	nid.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_MAIN));
	cp_strcpy_s(nid.szTip,TEXT("ZAS Telephone"));
	nid.uVersion = NOTIFYICON_VERSION;
	nid.uCallbackMessage = WM_SYSTEMTRAY;
	Shell_NotifyIcon(NIM_ADD,&nid);

	while ((ret = GetMessage(&AMsg, NULL, 0, 0)) != 0) {
		if (ret == -1)
			continue;
        if (!TranslateAccelerator(hwndMain, hAccel, &AMsg)) {
			BOOL trans = TRUE;
			u32 i;
			// check all modeless dialogs
			EnterCriticalSection(&inst->chat_cs);
			for (i = 0; (i < inst->chats_num) && trans; i++)
				trans = trans && !(inst->chats[i].hWnd && IsWindow(inst->chats[i].hWnd) && IsDialogMessage(inst->chats[i].hWnd,&AMsg));
			LeaveCriticalSection(&inst->chat_cs);
			if (trans && !IsDialogMessage(g_logWnd,&AMsg) && !IsDialogMessage(g_NetList,&AMsg) && !IsDialogMessage(g_KadList,&AMsg) && !(IsWindow(inst->hInfo) && IsDialogMessage(inst->hInfo,&AMsg)))
			{
				TranslateMessage(&AMsg);
				DispatchMessage(&AMsg);
			}
		}
	}

	Shell_NotifyIcon(NIM_DELETE,&nid);

#ifndef _DEBUG
	CloseHandle(gMutex);
#endif
	WSACleanup();

	return 0;
}
