#include <winsock2.h>
#include <Mswsock.h>
#include "network.h"
#include "password.h"
#include "resource.h"
#include "log.h"
#include "hash_func.h"
#include "random.h"
#include "diffie_helman.h"
#include "main.h"
#include "build.h"
#include <tchar.h>
#include <CommCtrl.h>
#include <Windows.h>
//-----------------------------------------
static LPTSTR module = TEXT("Network");
//-----------------------------------------
ESTABLISHMENT::ESTABLISHMENT()
{
	state = NET_STATE_IDLE;
	timeout = 0;
	send_requests = 0;
	rec_flags = 0;
	rec_size = 0;
	last_size = 0;
	sent_size = 0;
	conf_size = 0;
}
//-----------------------------------------
ESTABLISHMENT::~ESTABLISHMENT()
{
	tx_session.reset();
	rx_session.reset();
}
//-----------------------------------------

CONNECTION::CONNECTION(const NODE* node)
{
	CopyMemory(&nd,node,sizeof(NODE));
	established = 0;
	next_connection = NULL;
	est = NULL;
	pack_num_est = 0;
	pack_num_sec = 0;
	pack_num_drop = 0;
	pack_num_sent = 0;
	send_cnt = 0;
	away = false;
	cp_strcchcpy((tchar_t*)userInfo,sizeof(userInfo)/sizeof(tchar_t),_T("<unknown>"));
	userInfoSize = (u32)cp_strlen((tchar_t*)userInfo)*sizeof(tchar_t);
	InitializeCriticalSection(&est_cs);
}
//-----------------------------------------
CONNECTION::~CONNECTION()
{
	EnterCriticalSection(&est_cs);
	if (est) delete est;
	est = NULL;
	LeaveCriticalSection(&est_cs);
	DeleteCriticalSection(&est_cs);
	ses.reset();
}
//-----------------------------------------------------------------------
DWORD WINAPI receiverProc(NETWORK* n);
DWORD WINAPI timeoutProc(NETWORK* n);
//-----------------------------------------------------------------------
NETWORK::NETWORK(NODE* me, HWND mWnd)
{
	InitializeCriticalSection(&cs);
	con = NULL;
	away = false;
	session_is_initialized = false;

	CopyMemory(my_id,me->id,NETWORK_ID_LENGTH*sizeof(u32));

    mainWnd = mWnd;
	recThread = NULL;

	// open new socket
	soc = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
	if (soc == INVALID_SOCKET)
	{
		MessageBox(mWnd,TEXT("Can't create socket"),TEXT("Error"),MB_OK|MB_ICONERROR);
		return;
	}
	ZeroMemory(&my_addr,sizeof(my_addr));
	my_addr.sin_family = AF_INET;
	my_addr.sin_addr.s_addr = INADDR_ANY;
	my_addr.sin_port = htons(me->port);
	if (SOCKET_ERROR == bind(soc,(sockaddr*)&my_addr,sizeof(my_addr)))
	{
		MessageBox(mWnd,TEXT("Can't bind socket to local port"),TEXT("Error"),MB_OK|MB_ICONERROR);
		return;
	}
    // set receiver timeout in ms
	DWORD timeout = 100;
	setsockopt(soc,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeout,sizeof(timeout));
	// disable  new behavior using
	// IOCTL: SIO_UDP_CONNRESET
	DWORD dwBytesReturned = 0;
	BOOL bNewBehavior = FALSE;
	WSAIoctl(soc, SIO_UDP_CONNRESET,&bNewBehavior, sizeof(bNewBehavior),NULL,0,&dwBytesReturned,NULL,NULL);
	// Initialize sending packet
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	if (pas_len)
	{
		sending_size = cmn_tx_session.init_session_1((u8*)sending_buff);
		// create receiver thread and stop event
		eStopThreads = CreateEvent(NULL,TRUE,FALSE,NULL);
		recThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)receiverProc,this,0,NULL);
		tmoThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)timeoutProc, this, 0, NULL);
	}
	else
	{
		recThread = NULL;
	}
}
//-----------------------------------------------------------------------
NETWORK::~NETWORK()
{
	u32 buffer_coded[100];
	CopyMemory(buffer_coded,my_id,NETWORK_ID_LENGTH*sizeof(u32));
	buffer_coded[NETWORK_ID_LENGTH] = SESSION_TERMINATE|(away?SESSION_USER_AWAY:0);
	if (recThread)
	{
		SetEvent(eStopThreads);
		WaitForSingleObject(recThread,INFINITE);
		CloseHandle(recThread);
		WaitForSingleObject(tmoThread,INFINITE);
		CloseHandle(tmoThread);
	}
	if (con) // clear all connections
	{
		while(con)
		{
			CONNECTION* cn = con->next_connection;
			if (con->established)
			{
				sockaddr_in addr = {0};
				addr.sin_family = AF_INET;
				addr.sin_addr.s_addr = htonl(con->nd.ip);
				addr.sin_port = htons(con->nd.port);
				LogLowLevel(true,con->nd.ip,con->nd.port,con->nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32));
				sendto(soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			}
			delete con;
			con = cn;
		}
	}
	closesocket(soc);
    DeleteCriticalSection(&cs);
	cmn_tx_session.reset();
}
//-----------------------------------------------------------------------
CONNECTION* NETWORK::find_connection(const NODE* nd, bool establish)
{
	CONNECTION *s, *prev_s;
	if (!con) // there is no any session
	{
		if (establish)
		{
			con = new CONNECTION(nd);
			return con;
		}
		else
		{
			return NULL;
		}
	}
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(s->nd.id,nd->id))
		{
			return s;
		}
		prev_s = s;
	}
	if (!establish) return NULL;
	prev_s->next_connection = new CONNECTION(nd);
	return prev_s->next_connection;
}
//-----------------------------------------------------------------------
void NETWORK::send_big_buffer(u32 ip, u16 port, const u32* id, u32* buf, u32 size, u32 command, u32* pack_cnt)
{
	u32 tx_buffer[NETWORK_SIZE_UDP/sizeof(u32)];
	sockaddr_in addr = {0};
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(ip);
	addr.sin_port = htons(port);
	
	CopyMemory(tx_buffer,my_id,NETWORK_ID_LENGTH*sizeof(u32));

	u32 total_pack = (size+NETWORK_SIZE_SECURED-1)/NETWORK_SIZE_SECURED; // number of packets for sending
	for (u32 p = 0; p < total_pack; p++)
	{
		u32 chunk_size = (p==(total_pack-1)) ? (size-(total_pack-1)*NETWORK_SIZE_SECURED) : NETWORK_SIZE_SECURED; // size of data chunk
		tx_buffer[NETWORK_ID_LENGTH] = command|(away?SESSION_USER_AWAY:0)|(total_pack<<8)|(p<<16);
		CopyMemory(&tx_buffer[NETWORK_ID_LENGTH+1],&buf[p*NETWORK_SIZE_SECURED/sizeof(u32)],chunk_size);
		LogLowLevel(true,ip,port,id,command,chunk_size+(NETWORK_ID_LENGTH+1)*sizeof(u32));
		sendto(soc,(char*)tx_buffer,chunk_size+(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
	}
	*pack_cnt += total_pack;
}
//-----------------------------------------------------------------------
bool NETWORK::collect_big_buffer(u32* cur_buf, u32 size, ESTABLISHMENT* est)
{
	u32 total_pack = (cur_buf[0]>>8)&0xFF;
	u32 pack_num = (cur_buf[0]>>16)&0xFF;
	if ((pack_num >= total_pack) || (total_pack*NETWORK_SIZE_SECURED > NETWORK_SIZE_DH_MAX) ||
		((pack_num != (total_pack-1)) && (size != (NETWORK_SIZE_SECURED+4))) || 
		((pack_num == (total_pack-1)) && (size >= (NETWORK_SIZE_SECURED+4)))) // error
	{
		est->rec_flags = 0;
		return false;
	}
	CopyMemory(&est->buff[pack_num*NETWORK_SIZE_SECURED/sizeof(u32)],&cur_buf[1],size-4);
	est->rec_flags |= 1<<pack_num;
	if (pack_num == (total_pack-1)) // last packet
	{
		est->rec_size = (total_pack-1)*NETWORK_SIZE_SECURED + size-4;
	}
	for (u32 i = 0; i < total_pack; i++)
	{
		if (!(est->rec_flags & (1<<i))) return false;
	}
	return true;
}
//-----------------------------------------------------------------------
void NETWORK::send_small_buffer(u32 ip, u16 port, const u32* id, const u32* buf, u32 size, u32* pack_cnt)
{
	u32 tx_buffer[NETWORK_SIZE_UDP/sizeof(u32)];
	sockaddr_in addr = {0};
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(ip);
	addr.sin_port = htons(port);
	CopyMemory(tx_buffer,my_id,NETWORK_ID_LENGTH*sizeof(u32));
	if (size < NETWORK_SIZE_SECURED/sizeof(u32))
	{
		CopyMemory(&tx_buffer[NETWORK_ID_LENGTH],buf,size*sizeof(u32));
		LogLowLevel(true,ip,port,id,tx_buffer[NETWORK_ID_LENGTH]&0xFF,size*sizeof(u32));
		sendto(soc,(char*)tx_buffer,(NETWORK_ID_LENGTH+size)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
		(*pack_cnt)++;
	}
}
//-----------------------------------------------------------------------
u32 NETWORK::send_packet(PACKET_FOR_SEND* pack)
{
	u32 size = 0;
	u32 tx_buffer_plain[2048];
	u32 tx_buffer_coded[2048];
	CONNECTION* s;
	sockaddr_in addr = {0};
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(pack->nd->ip);
	addr.sin_port = htons(pack->nd->port);

	if (isIdIdentical(pack->nd->id,my_id) || !session_is_initialized) return 0; // don't send data to myself or if not initialized

	CopyMemory(tx_buffer_coded,my_id,NETWORK_ID_LENGTH*sizeof(u32));

	EnterCriticalSection(&cs);
    if (isZeroID(pack->nd->id)) // for zero ID try to find IP:port, or just send ping
    {
		for (s = con; s; s = s->next_connection)
		{
			if ((s->nd.ip == pack->nd->ip) && (s->nd.port == pack->nd->port))
			{
				memcpy((void*)pack->nd->id,s->nd.id,NETWORK_ID_LENGTH*sizeof(u32));
				break;
			}
		}
		if (!s) // can't find ID:port combination
		{
			LeaveCriticalSection(&cs);
			tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_PING_REQ|(away?SESSION_USER_AWAY:0);
			tx_buffer_coded[NETWORK_ID_LENGTH+1] = NETWORK_PROTOCOL_VERSION;
			tx_buffer_coded[NETWORK_ID_LENGTH+2] = atol(BUILD_NUM);
			LogLowLevel(true,pack->nd->ip,pack->nd->port,pack->nd->id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+3)*sizeof(u32));
			sendto(soc,(char*)tx_buffer_coded,(NETWORK_ID_LENGTH+3)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			return 0;
		}
    }
	else
	{
		s = find_connection(pack->nd,true);
	}

	if (!s) // not enough memory for new connection
	{
		LeaveCriticalSection(&cs);
		LogMessage(module, TEXT("Not enough memory for new connection"));
		return 0;
	}

	if(s->established)
	{
		if (s->send_cnt >= NETWORK_OFFLINE_SEND_LIMIT) // so long time no any answer. Reset session
		{
			s->established = false;
			s->send_cnt = 0;
			s->ses.reset();
		}
		else // send sequred packet
		{
			addr.sin_addr.s_addr = htonl(s->nd.ip);
			addr.sin_port = htons(s->nd.port);
			tx_buffer_plain[0] = pack->protocol;
			CopyMemory(&tx_buffer_plain[1], pack->buffer, pack->buf_len);
			LogLowLevelEnc(true,pack->buffer);
			tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_ENCODED|(away?SESSION_USER_AWAY:0);
			size = s->ses.encrypt((u8*)&tx_buffer_coded[NETWORK_ID_LENGTH+1],(u8*)tx_buffer_plain,pack->buf_len+sizeof(u32));
			LogLowLevel(true,s->nd.ip,s->nd.port,s->nd.id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,size+(NETWORK_ID_LENGTH+1)*sizeof(u32));
			size = (u32)sendto(soc,(char*)tx_buffer_coded,size+(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			s->send_cnt++;
			LeaveCriticalSection(&cs);
			return size;
		}
	}
	if (TryEnterCriticalSection(&s->est_cs))
	{
		if (!s->est) s->est = new ESTABLISHMENT;
		if ((s->est->state == NET_STATE_IDLE) || (s->est->state == NET_STATE_PINGING))
		{
			s->est->state = NET_STATE_PINGING;
			s->est->timeout = 50;
			s->est->ip = pack->nd->ip;
			s->est->port = pack->nd->port;
			tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_PING_REQ|(away?SESSION_USER_AWAY:0);
			tx_buffer_coded[NETWORK_ID_LENGTH+1] = NETWORK_PROTOCOL_VERSION;
			tx_buffer_coded[NETWORK_ID_LENGTH+2] = atol(BUILD_NUM);
			LogLowLevel(true,pack->nd->ip,pack->nd->port,pack->nd->id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+3)*sizeof(u32));
			sendto(soc,(char*)tx_buffer_coded,(NETWORK_ID_LENGTH+3)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			s->pack_num_sent++;
		}
		LeaveCriticalSection(&s->est_cs);
	}
	LeaveCriticalSection(&cs);
	return size;
}
//-----------------------------------------------------------------------
s8 name_compare(const u8* buf_ref,const u8* buf_test,u32 max_size)
{
	while (max_size--)
	{
		s8 diff = *buf_ref++ - *buf_test++;
		if (diff) return diff;
	}
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI timeoutProc(NETWORK* n)
{
	CONNECTION* s;

	while (WAIT_TIMEOUT == WaitForSingleObject(n->eStopThreads, 100))
	{
		EnterCriticalSection(&n->cs);
		if (!n->session_is_initialized)
		{
			LeaveCriticalSection(&n->cs);
			continue;
		}
		for (s = n->con; s; s = s->next_connection)
		{
			if (!s->established && TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est)
				{
					if (s->est->timeout) s->est->timeout--;
					else
					{
						u32 buf[3];
						TCHAR str[400];
						TCHAR ip_port[IP_LENGTH_CCH+5];
						ip_port2str(s->est->ip,s->est->port,ip_port);
						switch(s->est->state)
						{
						case NET_STATE_PINGING:
							buf[0] = SESSION_PING_REQ|(n->away?SESSION_USER_AWAY:0);
							buf[1] = NETWORK_PROTOCOL_VERSION;
							buf[2] = atol(BUILD_NUM);
							n->send_small_buffer(s->est->ip,s->est->port,s->nd.id,buf,3,&s->pack_num_sent);
							s->est->timeout = 100;
							break;
						case NET_STATE_S_REQUESTING:
							if (s->est->send_requests)
							{
								n->send_big_buffer(s->est->ip,s->est->port,s->nd.id, s->est->sent_buf, s->est->sent_size, SESSION_REQUEST, &s->pack_num_sent);
								s->est->timeout = 100;
								s->est->send_requests--;
								_stprintf_s(str, TEXT("Again send session request to %s"), ip_port);
							}
							else
							{
								delete s->est;
								s->est = NULL;
								_stprintf_s(str, TEXT("Timeout for sending requests to %s"), ip_port);
							}
							break;
						case NET_STATE_S_CONFIRMING:
							if (s->est->send_requests)
							{
								n->send_big_buffer(s->est->ip,s->est->port,s->nd.id, s->est->conf_buf, s->est->conf_size, SESSION_CONFIRMATION, &s->pack_num_sent);
								s->est->send_requests--;
								s->est->timeout = 100;
								_stprintf_s(str, TEXT("Again send session confirmation to %s"), ip_port);
							}
							else
							{
								delete s->est;
								s->est = NULL;
								_stprintf_s(str, TEXT("Timeout for sending requests to %s"), ip_port);
							}
							break;
						case NET_STATE_P_REQUESTED:
						case NET_STATE_S_REQUESTED:
						case NET_STATE_WAIT_CONFIRMATION:
							delete s->est;
							s->est = NULL;
							_stprintf_s(str, TEXT("Timeout for connection from %s"), ip_port);
							break;
						case NET_STATE_REJECTED:
							delete s->est;
							s->est = NULL;
							break;
						}
						//LogMessage(module, str); // !!! inside critical section
					}
				}
				LeaveCriticalSection(&s->est_cs);
			}
		}
		LeaveCriticalSection(&n->cs);
	}
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI rx_signature(CONNECTION* s)
{
	EnterCriticalSection(&s->est_cs);
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	NETWORK* n = getNet();
	TCHAR str[400];
	TCHAR ip_port[IP_LENGTH_CCH+5];
	if (s->est)
	{
		ip_port2str(s->est->ip,s->est->port,ip_port);
		if (s->est->protocol == NETWORK_PROTOCOL_VERSION)
		{
			s->est->sent_size = s->est->rx_session.response_to_init_session_1((u8*)s->est->sent_buf,(u8*)s->est->buff, getSecretData(), passw, pas_len);
			if (s->est->sent_size > 0)
			{
				n->send_big_buffer(s->est->ip,s->est->port,s->nd.id,s->est->sent_buf,s->est->sent_size,SESSION_RESPONSE,&s->pack_num_sent);
				s->est->state = NET_STATE_WAIT_CONFIRMATION;
				s->est->timeout = 300;
				_stprintf_s(str,TEXT("Wait confirmation because of successful session request from %s"),ip_port);
				s->est->last_size = s->est->rec_size;
				memcpy(s->est->last_buf,s->est->buff,s->est->rec_size);
				s->est->rec_size = 0;
				s->est->rec_flags = 0;
			}
			else
			{
				// WARNING: someone sent incorrect request, why?
				_stprintf_s(str,TEXT("Wrong Session request from %s"),ip_port);
				delete s->est;
				s->est = NULL;
			}
		}
		else
		{
			u32 cmd = SESSION_REJECT|(n->away?SESSION_USER_AWAY:0);
			n->send_small_buffer(s->est->ip,s->est->port,s->nd.id,&cmd,1,&s->pack_num_sent);
			_stprintf_s(str,TEXT("Another protocol version on %s, request is ignored"),ip_port);
			delete s->est;
			s->est = NULL;
		}
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
	}
	else LeaveCriticalSection(&s->est_cs);
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI message_thread(void* str)
{
	MessageBox(NULL,(LPCTSTR)str,TEXT("Warning"),MB_OK|MB_ICONSTOP);
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI tx_signature(CONNECTION* s)
{
	EnterCriticalSection(&s->est_cs);
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	u32 session_id[fingerprint_length_x32];
	NETWORK* n = getNet();
	TCHAR str[400];
	TCHAR ip_port[IP_LENGTH_CCH+5];
	if (s->est)
	{
		ip_port2str(s->est->ip,s->est->port,ip_port);
		if (s->est->protocol == NETWORK_PROTOCOL_VERSION)
		{
			s->est->conf_size = s->est->tx_session.process_response_confirm_session_1((u8*)s->est->conf_buf,(u8*)s->est->buff,getSecretData(),passw,pas_len);
			if ((s->est->conf_size > 0) && s->est->tx_session.session_is_active_1(session_id))
			{
				s->userInfoSize = s->est->tx_session.get_user_info_1(s->userInfo);
				s->userInfo[s->userInfoSize] = s->userInfo[s->userInfoSize+1] = 0;

				if (isIdIdentical(s->nd.id,session_id))
				{
					s->est->state = NET_STATE_S_CONFIRMING;
					s->est->timeout = 100;
					s->est->send_requests = 10;
					_stprintf_s(str,TEXT("Successful session response from %s, user name: %s, version %d, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),ip_port,
						s->userInfo,s->est->version,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
					n->send_big_buffer(s->est->ip, s->est->port,s->nd.id,s->est->conf_buf,s->est->conf_size,SESSION_CONFIRMATION,&s->pack_num_sent);
					s->est->last_size = s->est->rec_size;
					memcpy(s->est->last_buf,s->est->buff,s->est->rec_size);
					s->est->rec_size = 0;
					s->est->rec_flags = 0;
				}
				else // alarm!!!
				{
					s->established = false;
					_stprintf_s(str,TEXT("USER ID IS NOT THE SAME AS KEY FINGERPTING from %s, user name: %s, version %d, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),ip_port,
						s->userInfo,s->est->version,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
					delete s->est;
					s->est = NULL;
					CloseHandle(CreateThread(NULL,0,message_thread,str,0,NULL));
					Sleep(50);
				}
			}
			else
			{
				// WARNING: Someone sent incorrect response
				_stprintf_s(str,TEXT("Wrong Session response from %s"),ip_port);
				delete s->est;
				s->est = NULL;
			}
		}
		else
		{
			_stprintf_s(str,TEXT("Another protocol version on %s"),ip_port);
			delete s->est;
			s->est = NULL;
		}
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
	}
	else LeaveCriticalSection(&s->est_cs);
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI conf_signature(CONNECTION* s)
{
	EnterCriticalSection(&s->est_cs);
	TCHAR str[400];
	u32 session_id[fingerprint_length_x32];
	NETWORK* n = getNet();
	TCHAR ip_port[IP_LENGTH_CCH+5];
	if (s->est)
	{
		u32 size = s->est->rx_session.check_confirmation_1((u8*)s->est->buff);
		ip_port2str(s->est->ip,s->est->port,ip_port);
		if ((size > 0) && s->est->rx_session.session_is_active_1(session_id)) // normal response
		{
			s->userInfoSize = s->est->rx_session.get_user_info_1(s->userInfo);
			s->userInfo[s->userInfoSize] = s->userInfo[s->userInfoSize+1] = 0;
			if (isIdIdentical(s->nd.id,session_id))
			{
				s->est->rx_session.start_initialized_session(&s->ses);
				s->established = true; // here we start the new established session
				s->nd.ip = s->est->ip;
				s->nd.port = s->est->port;
				s->send_cnt = 0;
				u32 cmd = SESSION_CONFIRMED|(n->away?SESSION_USER_AWAY:0);
				n->send_small_buffer(s->nd.ip,s->nd.port,s->nd.id,&cmd,1,&s->pack_num_sent);

				_stprintf_s(str,TEXT("Successful session confirmation from %s, user name: %s, version %d, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),ip_port,
					s->userInfo,s->est->version,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
			}
			else // alarm!!!
			{
				s->established = false;
				_stprintf_s(str,TEXT("USER ID IS NOT THE SAME AS KEY FINGERPTING from %s, user name: %s, version %d, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),ip_port,
					s->userInfo,s->est->version,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
				CloseHandle(CreateThread(NULL,0,message_thread,str,0,NULL));
				Sleep(50);
			}
		}
		else
		{
			_stprintf_s(str,TEXT("Wrong Session confirmation from %s"),ip_port);
		}
		delete s->est;
		s->est = NULL;
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
	}
	else LeaveCriticalSection(&s->est_cs);
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI receiverProc(NETWORK* n)
{
	TCHAR str[400];
	TCHAR ip_port[IP_LENGTH_CCH+5];
	int pack_len;
	sockaddr_in addr = {0};
	CONNECTION* s;
	u32 size;
	u32 buffer_plain[2048];
	u32 buffer_coded[2048];
	u32 cmd;

    while (WAIT_TIMEOUT == WaitForSingleObject(n->eStopThreads,0))
	{
		int addr_size = sizeof(addr);
		memset(&addr,0,sizeof(addr));
		addr.sin_family = AF_INET;
		pack_len = recvfrom(n->soc,(char*)buffer_coded,sizeof(buffer_coded),0,(sockaddr*)&addr,&addr_size);
		KADEMLIA* kad = getKad();
		if (pack_len == SOCKET_ERROR) // error parsing
        {
			int err;
			if ((err = WSAGetLastError()) == WSAETIMEDOUT) continue;
			closesocket(n->soc);
			// open new socket
			n->soc = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
			if (n->soc == INVALID_SOCKET)
			{
				return 0;
			}
			if (SOCKET_ERROR == bind(n->soc,(sockaddr*)&n->my_addr,sizeof(n->my_addr)))
			{
				return 0;
			}
			// disable  new behavior using
			// IOCTL: SIO_UDP_CONNRESET
			DWORD dwBytesReturned = 0;
			BOOL bNewBehavior = FALSE;
			WSAIoctl(n->soc, SIO_UDP_CONNRESET,&bNewBehavior, sizeof(bNewBehavior),NULL,0,&dwBytesReturned,NULL,NULL);
			// set receiver timeout in ms
			DWORD timeout = 100;
			setsockopt(n->soc,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeout,sizeof(timeout));
			continue;
        }
        if (pack_len < (int)(NETWORK_ID_LENGTH+1)*sizeof(u32)) continue; // too small packet
        NODE nd;
        CopyMemory(nd.id,buffer_coded,NETWORK_ID_LENGTH*sizeof(u32));
		nd.ip = ntohl(addr.sin_addr.s_addr);
		nd.port = ntohs(addr.sin_port);
		if (isIdIdentical(nd.id,n->my_id)) continue; // don't process data from myself (something wrong with received packet)
        EnterCriticalSection(&n->cs);
		if (!n->session_is_initialized)
		{
			LeaveCriticalSection(&n->cs);
			continue;
		}
        s = n->find_connection(&nd,true);
		u8 pack_type = buffer_coded[NETWORK_ID_LENGTH]&0xFF;
		LogLowLevel(false,nd.ip,nd.port,nd.id,pack_type,pack_len);
		s->away = (buffer_coded[NETWORK_ID_LENGTH]&SESSION_USER_AWAY)!=0;
		ip_port2str(nd.ip,nd.port,ip_port);
		str[0] = 0;
		if (kad) kad->updateNode(&nd,true,getUserList());
        switch(pack_type)
        {
        case SESSION_PING_REQ: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(TryEnterCriticalSection(&s->est_cs))
			{
				if (!s->est) s->est = new ESTABLISHMENT;
				s->est->protocol = (pack_len < (int)(NETWORK_ID_LENGTH+2)*sizeof(u32)) ? 1 : buffer_coded[NETWORK_ID_LENGTH+1];
				s->est->version = (pack_len < (int)(NETWORK_ID_LENGTH+3)*sizeof(u32)) ? 0 : buffer_coded[NETWORK_ID_LENGTH+2];
				if (s->est->protocol != NETWORK_PROTOCOL_VERSION)
				{
					u32 cmd = SESSION_REJECT|(n->away?SESSION_USER_AWAY:0);
					n->send_small_buffer(nd.ip,nd.port,nd.id,&cmd,1,&s->pack_num_sent);
					_stprintf_s(str,TEXT("Another protocol version %d on %s"),s->est->protocol,ip_port);
					delete s->est;
					s->est = NULL;
				}
				else
				{
					switch(s->est->state)
					{
					case NET_STATE_IDLE:
					case NET_STATE_P_REQUESTED:
					case NET_STATE_REJECTED:
						s->est->state = NET_STATE_P_REQUESTED;
						s->est->timeout = 50;
						s->est->ip = nd.ip;
						s->est->port = nd.port;
						u32 buf[3];
						buf[0] = SESSION_PING_RESP|(n->away?SESSION_USER_AWAY:0);
						buf[1] = NETWORK_PROTOCOL_VERSION;
						buf[2] = atol(BUILD_NUM);
						n->send_small_buffer(nd.ip,nd.port,nd.id,buf,3,&s->pack_num_sent);
						break;
					case NET_STATE_PINGING:
					case NET_STATE_S_REQUESTING:
					case NET_STATE_S_CONFIRMING:
						if (nd.id[0] > n->my_id[0]) // both are trying to connect.
						{
							s->est->state = NET_STATE_P_REQUESTED;
							s->est->timeout = 50;
							s->est->ip = nd.ip;
							s->est->port = nd.port;
							u32 buf[3];
							buf[0] = SESSION_PING_RESP|(n->away?SESSION_USER_AWAY:0);
							buf[1] = NETWORK_PROTOCOL_VERSION;
							buf[2] = atol(BUILD_NUM);
							n->send_small_buffer(nd.ip,nd.port,nd.id,buf,3,&s->pack_num_sent);
							_stprintf_s(str,TEXT("Conflicted ping requrest from %s, switched to passive mode"),ip_port);
						}
						else
						{
							u32 cmd = SESSION_REJECT|(n->away?SESSION_USER_AWAY:0);
							n->send_small_buffer(nd.ip,nd.port,nd.id,&cmd,1,&s->pack_num_sent);
							_stprintf_s(str,TEXT("Conflicted ping requrest from %s, request rejected"),ip_port);
						}
						break;
					}
				}
				LeaveCriticalSection(&s->est_cs);
			}
            break;
        case SESSION_PING_RESP: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(TryEnterCriticalSection(&s->est_cs))
			{
				if (s->established) // check that noone else has the same identity
				{
					CONNECTION* ptr,*prev_s;
					for (prev_s = ptr = n->con; ptr;)
					{
						if (ptr != s && (((ptr->nd.ip == nd.ip) && (ptr->nd.port == nd.port)) || 
							(ptr->est && (ptr->est->ip == nd.ip) && (ptr->est->port == nd.port))))
						{
							// delete connection
							if (ptr == n->con)
							{
								n->con = n->con->next_connection;
							}
							else
							{
								prev_s->next_connection = ptr->next_connection;
							}
							getUserList()->delete_node(&ptr->nd);
							if (kad) kad->deleteNode(&ptr->nd);
							delete ptr;
							ptr = prev_s->next_connection;
							break;
						}
						else
						{
							prev_s = ptr;
							ptr = ptr->next_connection;
						}
					}
				}
				if (!s->est) s->est = new ESTABLISHMENT;
				switch(s->est->state)
				{
				case NET_STATE_IDLE:
				case NET_STATE_PINGING:
					s->est->protocol = (pack_len < (int)(NETWORK_ID_LENGTH+2)*sizeof(u32)) ? 1 : buffer_coded[NETWORK_ID_LENGTH+1];
					s->est->version = (pack_len < (int)(NETWORK_ID_LENGTH+3)*sizeof(u32)) ? 0 : buffer_coded[NETWORK_ID_LENGTH+2];
					if (s->est->protocol == NETWORK_PROTOCOL_VERSION)
					{
						s->est->state = NET_STATE_S_REQUESTING;
						s->est->timeout = 100;
						s->est->send_requests = 10;
						s->est->ip = nd.ip;
						s->est->port = nd.port;
						s->est->tx_session = n->cmn_tx_session;
						s->est->sent_size = n->sending_size;
						CopyMemory(s->est->sent_buf,n->sending_buff,n->sending_size);
						n->send_big_buffer(nd.ip,nd.port,nd.id,s->est->sent_buf,s->est->sent_size,SESSION_REQUEST,&s->pack_num_sent);
						_stprintf_s(str,TEXT("Send session request to %s"),ip_port);
					}
					else
					{
						_stprintf_s(str,TEXT("Another protocol version %d on %s"),s->est->protocol,ip_port);
						delete s->est;
						s->est = NULL;
					}
					break;
				}
				LeaveCriticalSection(&s->est_cs);
			}
            break;
        case SESSION_REQUEST: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est)
				{
					switch(s->est->state)
					{
					case NET_STATE_PINGING:
					case NET_STATE_S_REQUESTING:
					case NET_STATE_S_CONFIRMING:
						if (nd.id[0] > n->my_id[0]) // both are trying to connect.
						{
							s->est->state = NET_STATE_IDLE;
							_stprintf_s(str,TEXT("Conflicted session requrest from %s, switched to passive mode"),ip_port);
						}
						else
						{
							cmd = SESSION_REJECT|(n->away?SESSION_USER_AWAY:0);
							n->send_small_buffer(nd.ip,nd.port,nd.id,&cmd,1,&s->pack_num_sent);
							_stprintf_s(str,TEXT("Conflicted session requrest from %s, request rejected"),ip_port);
						}
						break;
					case NET_STATE_P_REQUESTED:
						s->est->state = NET_STATE_S_REQUESTED;
						s->est->timeout = 300;
					case NET_STATE_S_REQUESTED:
						if (n->collect_big_buffer(&buffer_coded[NETWORK_ID_LENGTH],pack_len-NETWORK_ID_LENGTH*sizeof(u32),s->est))
						{
							bool same_request = false;
							if (s->est->last_size && s->est->sent_size && (s->est->last_size == s->est->rec_size))
							{
								same_request = true;
								for (u32 i = 0; i < s->est->rec_size/4; i++)
									if (s->est->buff[i] != s->est->last_buf[i])
									{
										same_request = false;
										break;
									}
							}
							if (same_request)
							{
								n->send_big_buffer(nd.ip,nd.port,nd.id,s->est->sent_buf,s->est->sent_size,SESSION_RESPONSE,&s->pack_num_sent);
								LeaveCriticalSection(&s->est_cs);
								LeaveCriticalSection(&n->cs);
								_stprintf_s(str,TEXT("Same session request from %s"),ip_port);
								s->est->rec_size = 0;
								s->est->rec_flags = 0;
								break;
							}
							else
							{
								CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)rx_signature,s,0,NULL));
							}
							s->est->timeout=200;
						}
						break;
					}
				}
				LeaveCriticalSection(&s->est_cs);
			}
			else
			{
				u32 cmd = SESSION_BUSY|(n->away?SESSION_USER_AWAY:0);
				n->send_small_buffer(nd.ip,nd.port,nd.id,&cmd,1,&s->pack_num_sent);
			}
            break;
        case SESSION_RESPONSE: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est)
				{
					switch(s->est->state)
					{
					case NET_STATE_S_REQUESTING:
						if (n->collect_big_buffer(&buffer_coded[NETWORK_ID_LENGTH],pack_len-NETWORK_ID_LENGTH*sizeof(u32),s->est))
						{
							CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)tx_signature,s,0,NULL));
						}
						break;
					case NET_STATE_S_CONFIRMING:
						n->send_big_buffer(nd.ip,nd.port,nd.id,s->est->conf_buf,s->est->conf_size,SESSION_CONFIRMATION,&s->pack_num_sent);
						break;
					}
				}
				LeaveCriticalSection(&s->est_cs);
			}
            break;
        case SESSION_CONFIRMATION: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est)
				{
					switch(s->est->state)
					{
					case NET_STATE_WAIT_CONFIRMATION:
						if (n->collect_big_buffer(&buffer_coded[NETWORK_ID_LENGTH],pack_len-NETWORK_ID_LENGTH*sizeof(u32),s->est))
						{
							CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)conf_signature,s,0,NULL));
						}
						break;
					}
				}
				LeaveCriticalSection(&s->est_cs);
			}
			else
			{
				u32 cmd = SESSION_BUSY|(n->away?SESSION_USER_AWAY:0);
				n->send_small_buffer(nd.ip,nd.port,nd.id,&cmd,1,&s->pack_num_sent);
			}
            break;
		case SESSION_BUSY: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est)
				{
					switch(s->est->state)
					{
					case NET_STATE_S_REQUESTING:
					case NET_STATE_S_CONFIRMING:
						s->est->timeout = 300;
						break;
					}
				}
				LeaveCriticalSection(&s->est_cs);
			}
			break;
        case SESSION_REJECT: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est)
				{
					s->est->state = NET_STATE_REJECTED;
					s->est->timeout = 300;
					_stprintf_s(str,TEXT("Rejected connection to %s"),ip_port);
				}
				LeaveCriticalSection(&s->est_cs);
			}
			break;
        case SESSION_CONFIRMED: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est)
				{
					switch(s->est->state)
					{
					case NET_STATE_S_CONFIRMING:
						s->est->tx_session.start_initialized_session(&s->ses);
						s->established = true; // here we start the new established session
						s->nd.ip = s->est->ip;
						s->nd.port = s->est->port;
						s->send_cnt = 0;
						_stprintf_s(str,TEXT("Session confirmed by %s"),ip_port);
						break;
					}
				}
				LeaveCriticalSection(&s->est_cs);
			}
			break;
        case SESSION_TERMINATE: //---------------------------------------------------------------------------------------------------------------------------------------------
			s->pack_num_est++;
			if(s->established && TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est) delete s->est;
				s->est = NULL;
				s->established = false;
				s->ses.reset();
				LeaveCriticalSection(&s->est_cs);
				_stprintf_s(str,TEXT("Session rerminated by %s"),ip_port);
			}
			break;
        case SESSION_ENCODED: //---------------------------------------------------------------------------------------------------------------------------------------------
		case SESSION_NAME_REQ:
		case SESSION_NAME_RESP:
			if (!s->established && TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est && s->est->state == NET_STATE_S_CONFIRMING) // session established normaly, don't need data anymore
				{
					s->est->tx_session.start_initialized_session(&s->ses);
					s->established = true; // here we start the new established session
					s->nd.ip = s->est->ip;
					s->nd.port = s->est->port;
					s->send_cnt = 0;
					_stprintf_s(str,TEXT("Session confirmed by %s"),ip_port);
				}
				LeaveCriticalSection(&s->est_cs);
			}
            if (s->established)
            {
                u32 decr_size = s->ses.decrypt((u8*)buffer_plain,(u8*)&buffer_coded[NETWORK_ID_LENGTH+1],pack_len-(NETWORK_ID_LENGTH+1)*sizeof(u32));
                if (decr_size > sizeof(u32))
                {
					s->send_cnt = 0;
					LogLowLevelEnc(false,buffer_plain);
					if (TryEnterCriticalSection(&s->est_cs))
					{
						if (s->est) // session established normaly, don't need data anymore
						{
							delete s->est;
							s->est = NULL;
						}
						LeaveCriticalSection(&s->est_cs);
					}
					s->pack_num_sec++;
					if (pack_type == SESSION_ENCODED)
					{
						NET_PACKET np;
						np.buffer = buffer_plain;
						np.buf_len = decr_size;
						np.nd = &nd;
						LeaveCriticalSection(&n->cs);
						if (WAIT_TIMEOUT == WaitForSingleObject(n->eStopThreads,0))
						{
							parsePacket(&np);
						}
						continue;
					}
					else if (pack_type == SESSION_NAME_REQ)
					{
						CONNECTION* co;
						if (decr_size >= buffer_plain[0])
						{
							for (co = n->con; co; co = co->next_connection)
							{
								if (!name_compare((u8*)&buffer_plain[1],co->userInfo,min(co->userInfoSize,buffer_plain[0])))
								{
									CopyMemory(buffer_coded,n->my_id,NETWORK_ID_LENGTH*sizeof(u32));
									buffer_coded[NETWORK_ID_LENGTH] = SESSION_NAME_RESP|(n->away?SESSION_USER_AWAY:0);
									CopyMemory(buffer_plain,co->nd.id,NETWORK_ID_LENGTH*sizeof(u32));
									size = s->ses.encrypt((u8*)&buffer_coded[NETWORK_ID_LENGTH+1],(u8*)buffer_plain,NETWORK_ID_LENGTH*sizeof(u32));
									LogLowLevel(true,nd.ip,nd.port,nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32)+size);
									sendto(n->soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32)+size,0,(sockaddr*)&addr,sizeof(sockaddr_in));
								}
							}
						}
					}
					else
					{
						LeaveCriticalSection(&n->cs);
						if (decr_size >= NETWORK_ID_LENGTH*sizeof(u32))
						{
							u32* id = new u32[NETWORK_ID_LENGTH];
							CopyMemory(id,buffer_plain,NETWORK_ID_LENGTH*sizeof(u32));
							PostMessage(n->mainWnd,WM_USER_FOUND,0,(LPARAM)id);
						}
						continue;
					}
                }
                else
				{
					s->pack_num_drop++;
				}
            }
            break;
        }
		LeaveCriticalSection(&n->cs);
		if (str[0]) LogMessage(module,str);
	}
	return 0;
}
//---------------------------------------------------
CONNECTION* NETWORK::get_next_node(CONNECTION* cur)
{
    CONNECTION* curc = cur;
    if (cur)
    {
        return cur->next_connection;
    }
    return cur;
}
//---------------------------------------------------
void NETWORK::get_counters(const NODE* nd, u32* est, u32* sec, u32* drop, u32* sent, u32* missed)
{
	*est = 0;
	*sec = 0;
	*drop = 0;
	*sent = 0;
	*missed = 0;
    EnterCriticalSection(&cs);
	CONNECTION* s = find_connection(nd,false);
	if (s)
	{
		*est = s->pack_num_est;
		*sec = s->pack_num_sec;
		*drop = s->pack_num_drop;
		*sent = s->pack_num_sent;
		*missed = (u32)s->ses.dropped_rx_bytes();
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
void NETWORK::get_user_info(NODE* nd, LPTSTR info)
{
    EnterCriticalSection(&cs);
	CONNECTION* s = find_connection(nd,false);
	if (s)
	{
		memcpy(info,s->userInfo,s->userInfoSize);
		info[s->userInfoSize/sizeof(TCHAR)]=0;
	}
	else
	{
		cp_strcchcpy(info,user_info_max_length_bytes/sizeof(tchar_t),_T("<unknown>"));
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
bool NETWORK::findUserByName(u8* userInfo, u16 size, u32* id)
{
	CONNECTION *s;
	u32 tx_buffer_plain[200];
	u32 tx_buffer_coded[220];
	sockaddr_in addr = {0};
	//
	if (size > user_info_max_length_bytes) return false;
    EnterCriticalSection(&cs);
	// try to find locally
	for (s = con; s; s = s->next_connection)
	{
		if (!name_compare(userInfo,s->userInfo,min(s->userInfoSize,size)))
		{
			u32* id_msg = new u32[NETWORK_ID_LENGTH];
			memcpy(id,s->nd.id,sizeof(u32)*NETWORK_ID_LENGTH);
			memcpy(id_msg,s->nd.id,sizeof(u32)*NETWORK_ID_LENGTH);
			LeaveCriticalSection(&cs);
			PostMessage(mainWnd,WM_USER_FOUND,0,(LPARAM)id_msg);
			return true;
		}
	}
	// try to find externally
	CopyMemory(tx_buffer_coded,my_id,NETWORK_ID_LENGTH*sizeof(u32));
	for (s = con; s; s = s->next_connection)
	{
		if(s->established)
		{
			addr.sin_family = AF_INET;
			addr.sin_addr.s_addr = htonl(s->nd.ip);
			addr.sin_port = htons(s->nd.port);
			tx_buffer_plain[0] = size;
			CopyMemory(&tx_buffer_plain[1], userInfo, size);
			tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_NAME_REQ|(away?SESSION_USER_AWAY:0);
			size = s->ses.encrypt((u8*)&tx_buffer_coded[NETWORK_ID_LENGTH+1],(u8*)tx_buffer_plain,((size+3)/4)*4+sizeof(u32));
			LogLowLevel(true,s->nd.ip,s->nd.port,s->nd.id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,size+(NETWORK_ID_LENGTH+1)*sizeof(u32));
			size = (u32)sendto(soc,(char*)tx_buffer_coded,size+(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			s->pack_num_sent++;
			s->send_cnt++;
			if (s->send_cnt >= NETWORK_OFFLINE_SEND_LIMIT) // so long time no any answer. Reset session
			{
				s->established = false;
				s->ses.reset();
			}
		}	
	}
	LeaveCriticalSection(&cs);
	return false;
}
//---------------------------------------------------
int NETWORK::get_user_status(const u32* id)
{
	CONNECTION *s;
	int res = USER_STATUS_OFFLINE;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			if (s->established) res = s->away?USER_STATUS_AWAY:USER_STATUS_ONLINE;
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return res;
}
//---------------------------------------------------
u16* NETWORK::get_session_hash(const u32* id)
{
	CONNECTION *s;
	u16* ptr = NULL;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			ptr = (u16*)s->ses.session_fingerprint(); // 256-bit value
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return ptr;
}
//---------------------------------------------------
bool NETWORK::findNode(const u32 *id, NODE *nd)
{
	CONNECTION *s;
	bool res = false;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			*nd = s->nd;
			res = true;
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return res;
}
//---------------------------------------------------
void NETWORK::get_ip(NODE* nd)
{
	CONNECTION *s;
	bool res = false;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(nd->id,s->nd.id))
		{
			nd->ip = s->nd.ip;
			nd->port = s->nd.port;
			break;
		}
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
void NETWORK::set_ip(NODE* nd)
{
	CONNECTION *s;
	bool new_node = true;
	u32 old_ip = 0;
	u16 old_port = 0;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(nd->id,s->nd.id))
		{
			new_node = false;
			if (!s->established && nd->ip && nd->port)
			{
				old_ip = s->nd.ip;
				old_port = s->nd.port;
				s->nd.ip = nd->ip;
				s->nd.port = nd->port;
			}
			break;
		}
	}
	LeaveCriticalSection(&cs);
	if (new_node) // new node, nevere seen before
	{
		u32 msg = 1;
		PACKET_FOR_SEND pack;
		pack.buffer = &msg;
		pack.buf_len = sizeof(u32);
		pack.nd = nd;
		pack.protocol = NETWORK_PROT_KADEMLIA;
		send_packet(&pack);
	}
	else if ((old_ip!=nd->ip)||(old_port!=nd->port)) // IP is updated
	{
		TCHAR log_msg[1000];
		getUserList()->update_node(nd);
		_stprintf_s(log_msg,TEXT("Updated IP and port for %08X %08X %08X %08X %08X %08X %08X %08X from %3d.%3d.%3d.%3d:%5d to %3d.%3d.%3d.%3d:%5d"),nd->id[0],nd->id[1],nd->id[2],nd->id[3],nd->id[4],nd->id[5],nd->id[6],nd->id[7],
			FIRST_IPADDRESS(old_ip),SECOND_IPADDRESS(old_ip),THIRD_IPADDRESS(old_ip),FOURTH_IPADDRESS(old_ip),old_port,
			FIRST_IPADDRESS(nd->ip),SECOND_IPADDRESS(nd->ip),THIRD_IPADDRESS(nd->ip),FOURTH_IPADDRESS(nd->ip),nd->port);
		LogString(log_msg);
	}
}
//---------------------------------------------------
void NETWORK::reset_sessions(void)
{
	CONNECTION *s;
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
    EnterCriticalSection(&cs);
	session_is_initialized = false;
    LeaveCriticalSection(&cs);
	if (pas_len)
	{
		sending_size = cmn_tx_session.init_session_1((u8*)sending_buff);
	}
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		s->established = 0;
		s->ses.reset();
		EnterCriticalSection(&s->est_cs);
		if (s->est) delete s->est;
		s->est = NULL;
		LeaveCriticalSection(&s->est_cs);
	}
	session_is_initialized = true;
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
