#include <windows.h>
#include <Windowsx.h>
#include <commctrl.h>
#include <tchar.h>
#include <Shlwapi.h>

#include "resource.h"
#include "settings.h"
#include "password.h"
#include "main.h"

#include "signature.h"
#include "random.h"
#include "cp_fileop.h"
//----------------------------------------------------------------------------
#define MAX_STEPS	800
VOID AddEntropy(HWND hWnd, WPARAM wParam, LPARAM lParam, UINT type)
{
	int i;
	SYSTEMTIME time;
	POINT* last = (POINT*)GetProp(hWnd,TEXT("LastMouse"));
	u32 data[2] = {(u32)wParam,(u32)lParam};
	u32* buf = (u32*)&time;
	POINT pt;
	GetLocalTime(&time);
	for (i = 0; i < sizeof(time)/sizeof(u32); i += 2)
	{
		collect_entropy(&buf[i]);
	}
	switch (type)
	{
	case  WM_TIMER:
		if (GetCursorPos(&pt)/* && pt.x != last->x && pt.y != last->y*/)
		{
			collect_entropy((u32*)&pt);
			last->x = pt.x;
			last->y = pt.y;
			SendDlgItemMessage(hWnd,IDC_PROGRESS,PBM_STEPIT,0,0);
		}
		break;
	case WM_MOUSEMOVE:
		pt.x = GET_X_LPARAM(lParam);
		pt.y = GET_Y_LPARAM(lParam);
		ClientToScreen(hWnd,&pt);
		last->x = pt.x;
		last->y = pt.y;
	default:
		collect_entropy(data);
		SendDlgItemMessage(hWnd,IDC_PROGRESS,PBM_STEPIT,0,0);
		break;
	}
	if (MAX_STEPS == SendDlgItemMessage(hWnd,IDC_PROGRESS,PBM_GETPOS,0,0))
	{
		EndDialog(hWnd,0);
	}
}
//----------------------------------------------------------------------------
static void EntropyInit(HWND hWnd)
{
	SetCapture(hWnd);
	SetFocus(hWnd);
	SendDlgItemMessage(hWnd,IDC_PROGRESS,PBM_SETRANGE32,0,MAX_STEPS);
	SendDlgItemMessage(hWnd,IDC_PROGRESS,PBM_SETSTEP,1,0);
	SetProp(hWnd,TEXT("LastMouse"),LocalAlloc(LPTR,sizeof(POINT)));
	SetTimer(hWnd,1,100,NULL);
}
//----------------------------------------------------------------------------
static void EntropyDestroy(HWND hWnd)
{
	ReleaseCapture();
	LocalFree(RemoveProp(hWnd,TEXT("LastMouse")));
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK EntropyProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		EntropyInit(hWnd);
		return FALSE;
    case WM_DESTROY:
		EntropyDestroy(hWnd);
		return 0;
    case WM_CLOSE:
		EndDialog(hWnd,1);
		return 0;
	case WM_SETFOCUS:
		return TRUE;
	case WM_KILLFOCUS:
		SetFocus(hWnd);
		SetForegroundWindow(hWnd);
		return FALSE;
    case WM_TIMER:
	case WM_SYSKEYDOWN:
	case WM_MOUSEMOVE:
	case WM_KEYDOWN:
		AddEntropy(hWnd,wParam,lParam,uMsg);
		return TRUE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
static LPTSTR module = TEXT("Key & Password");
//----------------------------------------------------------------------------
enum {
	PASS_REQUEST,
	PASS_CHANGE,
	PASS_GENERATE
};
//----------------------------------------------------------------------------
typedef struct {
	u8 type;
	u8 password[password_max_length];
	u32 password_len;
	int remember; // remember password
	LPTSTR keyFileName;
} params_t;
//----------------------------------------------------------------------------
static void PasswordInit(HWND hWnd, params_t* par)
{
	SetProp(hWnd,TEXT("par"),par);
	Edit_LimitText(GetDlgItem(hWnd,IDC_PASSWORD),password_max_length);
	Edit_LimitText(GetDlgItem(hWnd,IDC_CONFIRMATION),password_max_length);
	Edit_LimitText(GetDlgItem(hWnd,IDC_USERINFO),user_info_max_length_bytes);
	SetDlgItemText(hWnd,IDC_PG_FILE,par->keyFileName);
	SECRET_DATA* secret = getSecretData();
	SetDlgItemText(hWnd,IDC_USERINFO,(LPTSTR)secret->certificate.user_info);
	CheckDlgButton(hWnd,IDC_REMEMBER,par->remember?BST_CHECKED:BST_UNCHECKED);
}
//----------------------------------------------------------------------------
static void PasswordDestroy(HWND hWnd)
{
	RemoveProp(hWnd,TEXT("par"));
}
//----------------------------------------------------------------------------
static void SwitchPasswordView(HWND hWnd)
{
	RECT rc;
	TCHAR password[password_max_length];
	GetDlgItemText(hWnd,IDC_PASSWORD,password,password_max_length);
	GetWindowRect(GetDlgItem(hWnd,IDC_PASSWORD),&rc);
	MapWindowPoints(NULL,hWnd,(LPPOINT)&rc,2);
	DestroyWindow(GetDlgItem(hWnd,IDC_PASSWORD));
	if (IsDlgButtonChecked(hWnd,IDC_SHOWSYMB))
	{
		CreateWindow(WC_EDIT,TEXT(""),WS_CHILD|WS_BORDER|WS_VISIBLE,rc.left,rc.top,rc.right-rc.left,rc.bottom-rc.top,hWnd,(HMENU)IDC_PASSWORD,GetModuleHandle(NULL),NULL);
		ShowWindow(GetDlgItem(hWnd,IDC_CONFIRMATION),SW_HIDE);
		ShowWindow(GetDlgItem(hWnd,IDC_CONFIRMATIONSTR),SW_HIDE);
	}
	else
	{
		CreateWindow(WC_EDIT,TEXT(""),WS_CHILD|WS_BORDER|WS_VISIBLE|ES_PASSWORD,rc.left,rc.top,rc.right-rc.left,rc.bottom-rc.top,hWnd,(HMENU)IDC_PASSWORD,GetModuleHandle(NULL),NULL);
		ShowWindow(GetDlgItem(hWnd,IDC_CONFIRMATION),SW_SHOW);
		ShowWindow(GetDlgItem(hWnd,IDC_CONFIRMATIONSTR),SW_SHOW);
	}
	SendDlgItemMessage(hWnd,IDC_PASSWORD,WM_SETFONT,SendDlgItemMessage(hWnd,IDC_SHOWSYMB,WM_GETFONT,0,0),0);
	Edit_LimitText(GetDlgItem(hWnd,IDC_PASSWORD),password_max_length);
	SetDlgItemText(hWnd,IDC_PASSWORD,password);
}
//----------------------------------------------------------------------------
static BOOL GetPassords(HWND hWnd)
{
	params_t* par = (params_t*)GetProp(hWnd,TEXT("par"));
	u8 confirmation[password_max_length];
	u32 conf_len;
	par->password_len = GetDlgItemTextA(hWnd,IDC_PASSWORD,(char*)par->password,password_max_length);
	conf_len = GetDlgItemTextA(hWnd,IDC_CONFIRMATION,(char*)confirmation,password_max_length);
	if ((par->type != PASS_REQUEST) && !IsDlgButtonChecked(hWnd,IDC_SHOWSYMB) && ((par->password_len != conf_len) || strncmp((char*)par->password,(char*)confirmation,conf_len)))
	{
		MessageBox(hWnd,TEXT("Passwords are different"),TEXT("Error"),MB_OK|MB_ICONERROR);
		Edit_SetText(GetDlgItem(hWnd,IDC_PASSWORD),TEXT(""));
		Edit_SetText(GetDlgItem(hWnd,IDC_CONFIRMATION),TEXT(""));
		return FALSE;
	}
	if (par->type == PASS_REQUEST)
	{
		par->remember = IsDlgButtonChecked(hWnd,IDC_REMEMBER) == BST_CHECKED;
	}
	if (par->type == PASS_GENERATE)
	{
        DIGITAL_SIGNATURE ds;
		SECRET_DATA secret;
		FILE* F;
		u8 userInfo[user_info_max_length_bytes];
		u32 user_info_len;
        u32 certificate_length,key_length,salt_length;
        u8 certificate[certificate_length_bytes];

		DialogBox(GetModuleHandle(NULL),TEXT("Entropy"),hWnd,EntropyProc); // collect entropy

		GetDlgItemText(hWnd,IDC_PG_FILE,par->keyFileName,MAX_PATH);
		if (PathFileExists(par->keyFileName) && (IDYES != MessageBox(hWnd,TEXT("Key file will be rewriten with new key\nDo you want to continue?"),TEXT("Key generation"),MB_YESNO|MB_ICONQUESTION)))
			return FALSE;
		if (_tfopen_s(&F,par->keyFileName,TEXT("wb")))
		{
			TCHAR str[200];
			_stprintf_s(str,200,TEXT("Can't open key file %s"),par->keyFileName);
			MessageBox(hWnd,str,TEXT("Error"),MB_OK|MB_ICONERROR);
			return FALSE;
		}
		user_info_len = GetDlgItemText(hWnd,IDC_USERINFO,(TCHAR*)userInfo,user_info_max_length_bytes/sizeof(TCHAR));
		ds.generate_keys(&secret,par->password,par->password_len,userInfo,(u16)(user_info_len*sizeof(TCHAR)));
        certificate_length = ds.serialize_certificate(certificate,&secret.certificate);
        key_length = signature_length_x32*sizeof(u32);
		salt_length = salt_length_x32*sizeof(u32);

        if ((1 != fwrite((char*)&certificate_length,sizeof(u32),1,F)) ||
            (certificate_length != fwrite((char*)certificate,sizeof(u8),certificate_length,F)) ||
            (1 != fwrite((char*)&key_length,sizeof(u32),1,F)) ||
            (key_length != fwrite((char*)secret.encoded_secret_key.key,sizeof(u8),key_length,F)) ||
            (1 != fwrite((char*)&secret.encoded_secret_key.hash,sizeof(u32),1,F)) ||
            (1 != fwrite((char*)&salt_length,sizeof(u32),1,F)) ||
            (salt_length != fwrite((char*)secret.encoded_secret_key.salt,sizeof(u8),salt_length,F)))
		{
				TCHAR str[200];
				_stprintf_s(str,200,TEXT("Can't write key to file %s"),par->keyFileName);
				MessageBox(hWnd,str,TEXT("Error"),MB_OK|MB_ICONERROR);
				fclose(F);
                return FALSE;
		}
		fclose(F);
		ds.reset();
		TCHAR str[user_info_max_length_bytes+200];
		_stprintf_s(str,200,TEXT("Key file generated successfully. User info: %s. Fingerprint: %08X %08X %08X %08X %08X %08X %08X %08X"),userInfo,secret.encoded_secret_key.key[0],secret.encoded_secret_key.key[1],secret.encoded_secret_key.key[2],secret.encoded_secret_key.key[3],secret.encoded_secret_key.key[4],secret.encoded_secret_key.key[5],secret.encoded_secret_key.key[6],secret.encoded_secret_key.key[7]);
		LogMessage(module,str);
	}
	if (par->type == PASS_CHANGE)
	{
        DIGITAL_SIGNATURE ds;
		SECRET_DATA* secret = getSecretData();
		FILE* F;
		u8 userInfo[user_info_max_length_bytes];
		u32 user_info_len;
        u32 certificate_length,key_length,salt_length;
        u8 certificate[certificate_length_bytes];
		u8 old_pass[password_max_length];
		u32 old_pass_len = GetDlgItemTextA(hWnd,IDC_OLDPASSWORD,(char*)old_pass,password_max_length);

		if (!change_password(&secret->encoded_secret_key,old_pass,old_pass_len,par->password,par->password_len))
		{
			MessageBox(hWnd,TEXT("Can't change password. Try again"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
			return FALSE;
		}
		if (_tfopen_s(&F,par->keyFileName,TEXT("wb")))
		{
			TCHAR str[200];
			_stprintf_s(str,200,TEXT("Can't open key file %s"),par->keyFileName);
			MessageBox(hWnd,str,TEXT("Error"),MB_OK|MB_ICONERROR);
			return FALSE;
		}
		user_info_len = GetDlgItemText(hWnd,IDC_USERINFO,(TCHAR*)userInfo,user_info_max_length_bytes/sizeof(TCHAR));
		ds.include_user_info(&secret->certificate,userInfo,(u16)(user_info_len*sizeof(TCHAR)));
        certificate_length = ds.serialize_certificate(certificate,&secret->certificate);
        key_length = signature_length_x32*sizeof(u32);
		salt_length = salt_length_x32*sizeof(u32);

        if ((1 != fwrite((char*)&certificate_length,sizeof(u32),1,F)) ||
            (certificate_length != fwrite((char*)certificate,sizeof(u8),certificate_length,F)) ||
            (1 != fwrite((char*)&key_length,sizeof(u32),1,F)) ||
            (key_length != fwrite((char*)secret->encoded_secret_key.key,sizeof(u8),key_length,F)) ||
            (1 != fwrite((char*)&secret->encoded_secret_key.hash,sizeof(u32),1,F)) ||
            (1 != fwrite((char*)&salt_length,sizeof(u32),1,F)) ||
            (salt_length != fwrite((char*)secret->encoded_secret_key.salt,sizeof(u8),salt_length,F)))
		{
				TCHAR str[200];
				_stprintf_s(str,200,TEXT("Can't write key to file %s"),par->keyFileName);
				MessageBox(hWnd,str,TEXT("Error"),MB_OK|MB_ICONERROR);
				fclose(F);
                return FALSE;
		}
		fclose(F);
		ds.reset();
		TCHAR str[user_info_max_length_bytes+200];
		_stprintf_s(str,200,TEXT("Key file generated successfully. User info: %s. Fingerprint: %08X %08X %08X %08X %08X %08X %08X %08X"),secret->certificate.user_info,secret->encoded_secret_key.key[0],secret->encoded_secret_key.key[1],secret->encoded_secret_key.key[2],secret->encoded_secret_key.key[3],secret->encoded_secret_key.key[4],secret->encoded_secret_key.key[5],secret->encoded_secret_key.key[6],secret->encoded_secret_key.key[7]);
		LogMessage(module,str);
	}
	return TRUE;
}
//-----------------------------------------------------------------------
bool LoadKey(HWND hWnd, LPTSTR fileName, SECRET_DATA *data)
{
	u32 certificate_length,key_length,hash,salt_length;
	u8 certificate[certificate_length_bytes];
	u32 key_array[signature_length_x32];
	u32 salt[salt_length_x32];
	DIGITAL_SIGNATURE ds;
	FILE* F;
	if (!_tcslen(fileName) || _tfopen_s(&F,fileName,TEXT("rb")))
	{
		TCHAR msg[MAX_PATH+100];
		_stprintf_s(msg,MAX_PATH+100,TEXT("Can't open key file:\n%s\nDefine correct key file in settings menu"),fileName);
		MessageBox(hWnd,msg,TEXT("Warning"),MB_OK|MB_ICONWARNING);
	    return false;
	}
	if ((1 != fread(&certificate_length,sizeof(u32),1,F)) ||
	    (certificate_length != certificate_length_bytes) ||
	    (certificate_length != fread(certificate,sizeof(u8),certificate_length,F)) ||
		(1 != fread(&key_length,sizeof(u32),1,F)) ||
	    (key_length != signature_length_x32*sizeof(u32)) ||
	    (key_length != fread(key_array,sizeof(u8),key_length,F)) ||
	    (1 != fread(&hash,sizeof(u32),1,F)) ||
		(1 != fread(&salt_length,sizeof(u32),1,F)) ||
	    (salt_length != salt_length_x32*sizeof(u32)) ||
	    (salt_length != fread(salt,sizeof(u8),salt_length,F)))
	{
		MessageBox(hWnd,TEXT("Wrong key file format\nDefine correct key file in settings menu"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
		fclose(F);
		return false;
	}
	fclose(F);
	// copy to secret data
	ds.deserialize_certificate(&data->certificate,certificate);
	CopyMemory(data->encoded_secret_key.key,key_array,key_length);
	data->encoded_secret_key.hash = hash;
	CopyMemory(data->encoded_secret_key.salt,salt,salt_length);
	if (!ds.certificate_looks_ok(&data->certificate))
	{
		MessageBox(hWnd,TEXT("Certificate doesn't look OK"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
	}
	ds.reset();
	data->data_is_initialized = 1;
	ZeroMemory(certificate,certificate_length_bytes);
	ZeroMemory(key_array,key_length);
	ZeroMemory(salt,salt_length);
	return true;
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK PasswordProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		PasswordInit(hWnd,(params_t*)lParam);
		return TRUE;
		//
    case WM_DESTROY:
		PasswordDestroy(hWnd);
		return 0;
		//
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			if (GetPassords(hWnd))
			{
				EndDialog(hWnd,1);
			}
			return TRUE;
		case IDCANCEL:
			EndDialog(hWnd,0);
			return TRUE;
		case IDC_SHOWSYMB:
			SwitchPasswordView(hWnd);
			return TRUE;
		case IDC_PG_FILESET:
			SelectFile(hWnd,IDC_PG_FILE);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//-----------------------------------------------------------------------
u32 PasswordRequest(HWND hWnd, u8* password, int* remember)
{
	params_t par;
	par.type = PASS_REQUEST;
	par.remember = *remember;
	if (DialogBoxParam(GetModuleHandle(NULL),TEXT("Password"),hWnd,PasswordProc,(LPARAM)&par))
	{
		CopyMemory(password,par.password,par.password_len);
		*remember = par.remember;
		return par.password_len;
	}
	else
	{
		return 0;
	}
}
//-----------------------------------------------------------------------
BOOL GenerateKey(HWND hWnd, LPTSTR keyFileName)
{
	params_t par;
	par.type = PASS_GENERATE;
	par.keyFileName = keyFileName;
	if (DialogBoxParam(GetModuleHandle(NULL),TEXT("PasswordGen"),hWnd,PasswordProc,(LPARAM)&par))
	{
		return TRUE;
	}
	LogMessage(module,TEXT("Key was not generated"));
	return FALSE;
}
//-----------------------------------------------------------------------
u32 ChangePassword(HWND hWnd, LPTSTR keyFileName, u8* new_password)
{
	params_t par;
	par.type = PASS_CHANGE;
	par.keyFileName = keyFileName;
	if (DialogBoxParam(GetModuleHandle(NULL),TEXT("PasswordChange"),hWnd,PasswordProc,(LPARAM)&par))
	{
		CopyMemory(new_password,par.password,par.password_len);
		return par.password_len;
	}
	LogMessage(module,TEXT("Password was not changed"));
	return 0;
}
//-----------------------------------------------------------------------
bool save_file_signature(pctstr_t filename)
{
	SECRET_DATA* sd = getSecretData();
	u8 pass[password_max_length];
	u32 pass_len = getPassword(pass);
	bool success = false;
	if (pass_len && sd)
	{
		fd_t f = cp_open(filename,FO_RDONLY);
		if (f!=INVALID_FD)
		{
			int len = (int)cp_seek(f,0,FS_END);
			u8* data = new u8[len];
			if (data)
			{
				cp_seek(f,0,FS_BEGIN);
				if (len == cp_read(f,data,len))
				{
					SIGNATURE sig;
					if (calculate_signature(&sig,sd,pass,pass_len,data,len))
					{
						tchar_t* signature_file_name = new tchar_t[cp_strlen(filename)+5];
						if (signature_file_name)
						{
							cp_snprintf(signature_file_name,cp_strlen(filename)+5,_T("%s.sig"),filename);
							fd_t sf = cp_open(signature_file_name,FO_WRONLY);
							if (sf!=INVALID_FD)
							{
								cp_write(sf,sig.fingerprint,sizeof(sig.fingerprint));
								cp_write(sf,sig.r,sizeof(sig.r));
								cp_write(sf,sig.s,sizeof(sig.s));
								cp_close(sf);
								success = true;
							}
							delete[] signature_file_name;
						}
					}
				}
				delete[] data;
			}
			cp_close(f);
		}
	}
	return success;
}
//-----------------------------------------------------------------------
bool verify_file_signature(pctstr_t filename)
{
	SECRET_DATA* sd = getSecretData();
	bool success = false;
	if (sd)
	{
		fd_t f = cp_open(filename,FO_RDONLY);
		if (f!=INVALID_FD)
		{
			int len = (int)cp_seek(f,0,FS_END);
			u8* data = new u8[len];
			if (data)
			{
				cp_seek(f,0,FS_BEGIN);
				if (len == cp_read(f,data,len))
				{
					tchar_t* signature_file_name = new tchar_t[cp_strlen(filename)+5];
					if (signature_file_name)
					{
						cp_snprintf(signature_file_name,cp_strlen(filename)+5,_T("%s.sig"),filename);
						fd_t sf = cp_open(signature_file_name,FO_RDONLY);
						if (sf!=INVALID_FD)
						{
							SIGNATURE sig;
							if ((sizeof(sig.fingerprint) == cp_read(sf,sig.fingerprint,sizeof(sig.fingerprint))) &&
								(sizeof(sig.r) == cp_read(sf,sig.r,sizeof(sig.r))) &&
								(sizeof(sig.s) == cp_read(sf,sig.s,sizeof(sig.s))))
							{
								success = (VERIFY_SIGNATURE_OK == verify_signature(&sig,&sd->certificate,data,len));
							}
							cp_close(sf);
						}
						delete[] signature_file_name;
					}
				}
				delete[] data;
			}
			cp_close(f);
		}
	}
	return success;
}
//-----------------------------------------------------------------------
void write_data_to_file(FILE* F, const void* str, size_t size, XEX_ENCRYPT_DECRYPT* xex, u64* xex_num)
{
	if (!xex->initialized() || !F) return;
	size_t total_size = size+sizeof(u32);
	if ((total_size&7) != 0) // fill up to 8-bytes boundary
	{
		total_size = (total_size+8)&~7;
	}
	u32* data = (u32*)LocalAlloc(LPTR,total_size);
	if (data)
	{
		data[0] = size;
		memcpy(&data[1],str,size);
		xex->Encrypt(data,*xex_num,total_size/8);
		*xex_num += total_size/8;
		fwrite(data,1,total_size,F);
		LocalFree(data);
	}
}
//-----------------------------------------------------------------------
void write_string_to_file(FILE* F, pctstr_t str, XEX_ENCRYPT_DECRYPT* xex, u64* xex_num)
{
	write_data_to_file(F, str, cp_strlen(str)*sizeof(tchar_t), xex, xex_num);
}
//-----------------------------------------------------------------------
int read_string_from_file(FILE* F, ptstr_t str, u32 str_size, XEX_ENCRYPT_DECRYPT* xex, u64* xex_num)
{
	if (!xex->initialized() || !F) return -1;
	u32 hdr[2];
	if (2 != fread(hdr,sizeof(u32),2,F)) return 0;
	xex->Decrypt(hdr,*xex_num);
	if ((hdr[0] > 0x7FFFFFFF) || (hdr[0]&(sizeof(tchar_t)-1))) return -1;
	if (hdr[0] > str_size-sizeof(tchar_t))
	{
		fseek(F,-8,FS_CURRENT);
		return -(int)hdr[0];
	}
	(*xex_num)++;
	
	u32 total_size = hdr[0]+sizeof(u32);
	if ((total_size&7) != 0) // fill up to 8-bytes boundary
	{
		total_size = (total_size+8)&~7;
	}
	u32* data = (u32*)LocalAlloc(LPTR,total_size);
	if (data)
	{
		data[1] = hdr[1];
		total_size -= 8;
		if (total_size != fread(&data[2],1,total_size,F))
		{
			LocalFree(data);
			return 0;
		}
		xex->Decrypt(&data[2],*xex_num,total_size/8);
		*xex_num += total_size/8;
		memcpy(str,&data[1],hdr[0]);
		str[hdr[0]/sizeof(tchar_t)] = 0;
		LocalFree(data);
	}
	else
	{
		return -1;
	}
	return hdr[0];
}
//-----------------------------------------------------------------------
xml_writer_secret_file::xml_writer_secret_file(void* file, XEX_ENCRYPT_DECRYPT* xex, u64* xex_num)
{
	this->file = file;
	this->xex = xex;
	this->xex_num = xex_num;
}
//-----------------------------------------------------------------------
void xml_writer_secret_file::write(const void* data, size_t size)
{
	if (!xex->initialized() || !file) return;
	size_t total_size = size;
	if ((total_size&7) != 0) // fill up to 8-bytes boundary
	{
		total_size = (total_size+8)&~7;
	}
	u32* sdata = (u32*)LocalAlloc(LPTR,total_size);
	if (sdata)
	{
		memcpy(sdata,data,size);
		xex->Encrypt(sdata,*xex_num,total_size/8);
		*xex_num += total_size/8;
		fwrite(sdata,1,total_size,(FILE*)file);
		LocalFree(sdata);
	}
}
//-----------------------------------------------------------------------
