#include "users.h"
#include "settings.h"
#include "main.h"
#include "lists.h"
#include "password.h"
#include <Share.h>
//-----------------------------------------------------------------------
USERS::USERS()
{
	users = NULL;
	likes = NULL;
	downloads = NULL;
	cp_initlock(&lock);
}
//-----------------------------------------------------------------------
USERS::~USERS()
{
	user_t *d1, *d2;
	for (d1 = d2 = users; d1; d1 = d2)
	{
		d2 = d1->next;
		if (d1->url) delete[] d1->url;
		delete d1;
	}
	like_t *l1, *l2;
	for (l1 = l2 = likes; l1; l1 = l2)
	{
		l2 = l1->next;
		if (l1->name) delete[] l1->name;
		delete l1;
	}
	download_t *w1, *w2;
	for (w1 = w2 = downloads; w1; w1 = w2)
	{
		w2 = w1->next;
		if (w1->file) delete[] w1->file;
		if (w1->user) delete[] w1->user;
		delete w1;
	}
	cp_deletelock(&lock);
}
//-----------------------------------------------------------------------
void USERS::update_node(const NODE* nd)
{
	user_t *new_user = new user_t;
	new_user->nd = *nd;
	cp_get_time(&new_user->last_seen);
	new_user->next = NULL;
	new_user->url = NULL;
	cp_lock(&lock);
	_append_user(new_user);
	// update like table
	if (!likes)
	{
		cp_unlock(&lock);
		return;
	}
	for (like_t* cur = likes; cur; cur = cur->next)
	{
		if (isIdIdentical(cur->nd.id,nd->id))
		{
			cur->nd.ip = nd->ip;
			cur->nd.port = nd->port;
			break;
		}
	}
	cp_unlock(&lock);
}
//-----------------------------------------------------------------------
void USERS::delete_node(const NODE* nd)
{
	user_t* cur, *prev;
	cp_lock(&lock);
	for (prev = cur = users; cur;)
	{
		if (isIdIdentical(cur->nd.id,nd->id))
		{
			if (cur == users)
			{
				users = users->next;
			}
			else
			{
				prev->next = cur->next;
			}
			if (cur->url) delete[] cur->url;
			delete cur;
			prev = prev->next;
			break;
		}
		else
		{
			prev = cur;
			cur = cur->next;
		}
	}
	cp_unlock(&lock);
}
//-----------------------------------------------------------------------
void id2str(const u32 *id, ptstr_t str)
{
	cp_snprintf(str,ID_LENGTH_CCH,_T("%08X %08X %08X %08X %08X %08X %08X %08X"),id[0],id[1],id[2],id[3],id[4],id[5],id[6],id[7]);
}
//-----------------------------------------------------------------------
bool str2id(pctstr_t str, u32 *id)
{
	return (NETWORK_ID_LENGTH == cp_sscanf(str,_T("%08X %08X %08X %08X %08X %08X %08X %08X"),&id[0],&id[1],&id[2],&id[3],&id[4],&id[5],&id[6],&id[7]));
}
//-----------------------------------------------------------------------
void ip2str(u32 ip, ptstr_t str)
{
	cp_snprintf(str,IP_LENGTH_CCH,_T("%d.%d.%d.%d"),(ip>>24)&0xFF,(ip>>16)&0xFF,(ip>>8)&0xFF,ip&0xFF);
}
//-----------------------------------------------------------------------
void ip_port2str(u32 ip, u16 port, ptstr_t str)
{
	cp_snprintf(str,IP_LENGTH_CCH+5,_T("%d.%d.%d.%d:%d"),(ip>>24)&0xFF,(ip>>16)&0xFF,(ip>>8)&0xFF,ip&0xFF,port);
}
//-----------------------------------------------------------------------
bool str2ip(pctstr_t str, u32 *ip)
{
	u32 part[4];
	*ip = 0;
	if (4 == cp_sscanf(str,_T("%d.%d.%d.%d"),&part[0],&part[1],&part[2],&part[3]))
	{
		*ip = (part[0]<<24)+(part[1]<<16)+(part[2]<<8)+part[3];
		return true;
	}
	return false;
}
//-----------------------------------------------------------------------
void time2str(const cp_time_t t, ptstr_t str)
{
	cp_utc_time_t utc;
	cp_convert_time_to_utc(&t,&utc);

	cp_snprintf(str,TIME_LENGTH_CCH,_T("%02d/%02d/%02d %02d:%02d:%02d"),utc.wYear%100,utc.wMonth,utc.wDay,utc.wHour,utc.wMinute,utc.wSecond);
}
//-----------------------------------------------------------------------
static bool _str2time(pctstr_t str, cp_time_t *t)
{
	u32 part[6];
	if (6 == cp_sscanf(str,_T("%02d/%02d/%02d %02d:%02d:%02d"),&part[0],&part[1],&part[2],&part[3],&part[4],&part[5]) &&
		((part[0] < 100) || ((part[0] >= 2000) && (part[0] < 2100))) &&
		((part[1] >= 1) && (part[1] <= 12)) &&
		((part[2] >= 1) && (part[2] <= 31)) &&
		(part[3] < 60) && (part[4] < 60) && (part[5] < 60))
	{
		cp_utc_time_t utc = {0};
		utc.wYear = (part[0] < 100) ? 2000+part[0] : part[0]; utc.wMonth = part[1]; utc.wDay = part[2];
		utc.wHour = part[3]; utc.wMinute = part[4]; utc.wSecond = part[5];
		cp_convert_utc_to_time(&utc,t);
		return true;
	}
	return false;
}
//-----------------------------------------------------------------------
void USERS::load_nodes(KADEMLIA* kad, pctstr_t filename)
{
	xml_document doc;
	xml_node xml_users;
	if (!doc.load_file(filename) || !(xml_users = doc.child(_T("users"))))// try to get users declaration
	{
		doc.reset();
		xml_users = doc.append_child(_T("users"));
		xml_node node = xml_users.append_child(_T("node"));
		node.append_attribute(_T("url")).set_value(_T("boot.zas-comm.ru"));
		node.append_attribute(_T("port")).set_value(_T("31265"));
		node = xml_users.append_child(_T("node"));
		node.append_attribute(_T("url")).set_value(_T("abvolt.changeip.org"));
		node.append_attribute(_T("port")).set_value(_T("31265"));
	}
	// parse users
	cp_lock(&lock);
    for (xml_node_iterator it = xml_users.begin(); it != xml_users.end(); ++it)
    {
		if (!cp_strcmp(it->name(),_T("node")))
		{
			user_t* usr = new user_t;
			memset(usr,0,sizeof(user_t));
			for (xml_attribute_iterator ait = it->attributes_begin(); ait != it->attributes_end(); ++ait)
			{
				if (!cp_strcmp(ait->name(),_T("ip")))
				{
					str2ip(ait->value(),&usr->nd.ip);
				}
				else if (!cp_strcmp(ait->name(),_T("port")))
				{
					usr->nd.port = (u16)(ait->as_uint()&0xFFFF);
				}
				else if (!cp_strcmp(ait->name(),_T("id")))
				{
					str2id(ait->value(),usr->nd.id);
				}
				else if (!cp_strcmp(ait->name(),_T("url")))
				{
					size_t len = cp_strlen(ait->value())+1;
					usr->url = new tchar_t[len];
					cp_strncpy(usr->url,ait->value(),len);
				}
				else if (!cp_strcmp(ait->name(),_T("time")))
				{
					_str2time(ait->value(),&usr->last_seen);
				}
			}
			if (usr->url) // transfer url to IP
			{
				u32 ip;
				URL2IP(usr->url,&ip);
				if (ip) usr->nd.ip = ip;
			}
			if (usr->nd.ip && usr->nd.port) // key parameters
			{
				if (!usr->last_seen) cp_get_time(&usr->last_seen);
				_append_user(usr);
				kad->updateNode(&usr->nd,false,NULL);
			}
			else
			{
				delete usr;
			}
		}
    }
	cp_unlock(&lock);
}
void USERS::_append_user(user_t *new_user)
{
	user_t* cur;
	bool normal_id = isNonZeroID(new_user->nd.id);
	if (!users)
	{
		users = new_user;
		return;
	}
	for (cur = users; cur; cur = cur->next)
	{
		if (normal_id && (isIdIdentical(cur->nd.id,new_user->nd.id) || ((cur->nd.ip == new_user->nd.ip) && (cur->nd.port == new_user->nd.port))))
		{
			if (!isNonZeroID(cur->nd.id))
			{
				memcpy(cur->nd.id,new_user->nd.id,NETWORK_ID_LENGTH*sizeof(u32));
			}
			if (cur->last_seen < new_user->last_seen) // new user is freshier
			{
				cur->nd.ip = new_user->nd.ip;
				cur->nd.port = new_user->nd.port;
				cur->last_seen = new_user->last_seen;
				if (!cur->url && new_user->url)
				{
					cur->url = new_user->url;
					new_user->url = NULL;
				}
			}
			if (new_user->url) delete[] new_user->url;
			delete new_user; // merged with current user, so have to free instance
			return;
		}
		if (!cur->next) // nobody next to current one
		{
			cur->next = new_user;
			return;
		}
	}
}
//-----------------------------------------------------------------------
void USERS::save_nodes(pctstr_t filename)
{
	cp_time_t month_time;
	u32 usr_cnt = 0;
	user_t* cur;
	tchar_t str[ID_LENGTH_CCH];
	xml_document doc;
	xml_node xml_users = doc.append_child(_T("users"));
	cp_get_time(&month_time);
	month_time -= (cp_time_t)30*24*3600*10000000;

	cp_lock(&lock);
	for (cur = users; cur; cur = cur->next)
	{
		if ((usr_cnt < 1000) || (cur->last_seen >= month_time)) // save only fresh users, but not less than 1000
		{
			xml_node node = xml_users.append_child(_T("node"));
			ip2str(cur->nd.ip,str);
			node.append_attribute(_T("ip")).set_value(str);
			node.append_attribute(_T("port")).set_value(cur->nd.port);
			id2str(cur->nd.id,str);
			node.append_attribute(_T("id")).set_value(str);
			time2str(cur->last_seen,str);
			node.append_attribute(_T("time")).set_value(str);
			if (cur->url) node.append_attribute(_T("url")).set_value(cur->url);
			usr_cnt++;
		}
	}
	cp_unlock(&lock);
	doc.save_file(filename);
}
//-----------------------------------------------------------------------
bool USERS::load_likes(pctstr_t filename)
{
	xml_document doc;
	xml_node xml_users;
	if (!doc.load_file(filename) || !(xml_users = doc.child(_T("likes"))))// try to get users declaration
	{
		return false;
	}
	// parse users
    for (xml_node_iterator it = xml_users.begin(); it != xml_users.end(); ++it)
    {
		if (!cp_strcmp(it->name(),_T("node")))
		{
			NODE nd;
			pctstr_t name = NULL;
			pctstr_t orig_name = NULL;
			bool trust = false;
			memset(&nd, 0, sizeof(NODE));
			for (xml_attribute_iterator ait = it->attributes_begin(); ait != it->attributes_end(); ++ait)
			{
				if (!cp_strcmp(ait->name(),_T("ip")))
				{
					str2ip(ait->value(),&nd.ip);
				}
				else if (!cp_strcmp(ait->name(),_T("port")))
				{
					nd.port = (u16)(ait->as_uint()&0xFFFF);
				}
				else if (!cp_strcmp(ait->name(),_T("id")))
				{
					str2id(ait->value(),nd.id);
				}
				else if (!cp_strcmp(ait->name(),_T("name")))
				{
					name = ait->value();
				}
				else if (!cp_strcmp(ait->name(), _T("name_original")))
				{
					orig_name = ait->value();
				}
				else if (!cp_strcmp(ait->name(), _T("trust")))
				{
					trust = ait->as_bool();
				}
			}
			append_like(&nd, name, orig_name, trust);
		}
    }
	return true;
}
static void _fill_like(like_t* like, NODE* nd, pctstr_t name, pctstr_t orig_name, bool trust)
{
	like->nd = *nd;
	if (name)
	{
		size_t len = cp_strlen(name)+1;
		like->name = new tchar_t[len];
		cp_strncpy(like->name,name,len);
	}
	else
	{
		like->name = NULL;
	}
	if (orig_name)
	{
		size_t len = cp_strlen(orig_name) + 1;
		like->name_orig = new tchar_t[len];
		cp_strncpy(like->name_orig, orig_name, len);
	}
	else
	{
		like->name_orig = NULL;
	}
	like->trust = trust;
	like->next = NULL;
}
void USERS::append_like(NODE* nd, pctstr_t name, pctstr_t orig_name, bool trust)
{
	like_t* cur;
	cp_lock(&lock);
	if (!likes)
	{
		likes = new like_t;
		_fill_like(likes, nd, name, orig_name, trust);
		cp_unlock(&lock);
		return;
	}
	for (cur = likes; cur; cur = cur->next)
	{
		if (isIdIdentical(cur->nd.id,nd->id))
		{
			break; // it's already exists
		}
		if (!cur->next)
		{
			cur->next = new like_t;
			_fill_like(cur->next, nd, name, orig_name, trust);
			break;
		}
	}
	cp_unlock(&lock);
}
//-----------------------------------------------------------------------
void USERS::remove_like(u32* id)
{
	like_t *cur, *prev;
	cp_lock(&lock);
	for (prev = cur = likes; cur; cur = cur->next)
	{
		if (isIdIdentical(cur->nd.id,id))
		{
			if (prev == cur)
			{
				likes = cur->next;
			}
			else
			{
				prev->next = cur->next;
			}
			if (cur->name) delete[] cur->name;
			delete cur;
			break;
		}
		prev = cur;
	}
	cp_unlock(&lock);
}
//-----------------------------------------------------------------------
void USERS::unauthorize_likes(void)
{
	cp_lock(&lock);
	for (like_t *cur = likes; cur; cur = cur->next)
	{
		cur->trust = false;
	}
	cp_unlock(&lock);
}
//-----------------------------------------------------------------------
void USERS::save_likes(pctstr_t filename)
{
	like_t* cur;
	tchar_t str[ID_LENGTH_CCH];
	xml_document doc;
	xml_node xml_users = doc.append_child(_T("likes"));

	cp_lock(&lock);
	for (cur = likes; cur; cur = cur->next)
	{
		xml_node node = xml_users.append_child(_T("node"));
		ip2str(cur->nd.ip,str);
		node.append_attribute(_T("ip")).set_value(str);
		node.append_attribute(_T("port")).set_value(cur->nd.port);
		id2str(cur->nd.id,str);
		node.append_attribute(_T("id")).set_value(str);
		if (cur->name) node.append_attribute(_T("name")).set_value(cur->name);
		if (cur->name_orig) node.append_attribute(_T("name_original")).set_value(cur->name_orig);
		node.append_attribute(_T("trust")).set_value(cur->trust);
	}
	cp_unlock(&lock);
	doc.save_file(filename);
}
//-----------------------------------------------------------------------
void USERS::load_downloads(pctstr_t filename)
{
	xml_document doc;
	xml_node xml_users;
	XEX_ENCRYPT_DECRYPT xex;
	u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	if (!pas_len) return;
	FILE* F = _tfsopen(filename,TEXT("rb+"), _SH_DENYWR);
	if (!F) return;

	u32 salt[8];
	if (8!=fread(salt,sizeof(u32),8,F))
	{
		fclose(F);
		return;
	}
	xex.Init(salt,getSecretData(),passw,pas_len);

	fseek(F, 0, SEEK_END);
	long data_size = ftell(F)-sizeof(salt);
	fseek(F, sizeof(salt), SEEK_SET);

	if (data_size<=0)
	{
		fclose(F);
		return;
	}
	u32* data = (u32*)LocalAlloc(LPTR,data_size+sizeof(u32));
	if (!data)
	{
		fclose(F);
		return;
	}
	fread(data,1,data_size,F);
	fclose(F);
	xex.Decrypt(data,0,data_size/8);
	data[data_size/sizeof(u32)]=0;
	if (!doc.load((pctstr_t)data) || !(xml_users = doc.child(_T("downloads"))))// try to get users declaration
	{
		LocalFree(data);
		return;
	}
	LocalFree(data);
	// parse users
    for (xml_node_iterator it = xml_users.begin(); it != xml_users.end(); ++it)
    {
		if (!cp_strcmp(it->name(),_T("file")))
		{
			NODE nd;
			pctstr_t file = NULL;
			pctstr_t user = NULL;
			cp_time_t time = 0;
			memset(&nd, 0, sizeof(NODE));
			for (xml_attribute_iterator ait = it->attributes_begin(); ait != it->attributes_end(); ++ait)
			{
				if (!cp_strcmp(ait->name(),_T("ip")))
				{
					str2ip(ait->value(),&nd.ip);
				}
				else if (!cp_strcmp(ait->name(),_T("port")))
				{
					nd.port = (u16)(ait->as_uint()&0xFFFF);
				}
				else if (!cp_strcmp(ait->name(),_T("id")))
				{
					str2id(ait->value(),nd.id);
				}
				else if (!cp_strcmp(ait->name(),_T("name")))
				{
					file = ait->value();
				}
				else if (!cp_strcmp(ait->name(), _T("user")))
				{
					user = ait->value();
				}
				else if (!cp_strcmp(ait->name(),_T("time")))
				{
					_str2time(ait->value(),&time);
				}
			}
			append_download(&nd, file, user, time);
		}
    }
}
//-----------------------------------------------------------------------
void USERS::save_downloads(pctstr_t filename)
{
	download_t* cur;
	tchar_t str[ID_LENGTH_CCH];
	xml_document doc;
	xml_node xml_users = doc.append_child(_T("downloads"));

	cp_lock(&lock);
	for (cur = downloads; cur; cur = cur->next)
	{
		xml_node node = xml_users.append_child(_T("file"));
		ip2str(cur->nd.ip,str);
		node.append_attribute(_T("ip")).set_value(str);
		node.append_attribute(_T("port")).set_value(cur->nd.port);
		id2str(cur->nd.id,str);
		node.append_attribute(_T("id")).set_value(str);
		if (cur->file) node.append_attribute(_T("name")).set_value(cur->file);
		if (cur->user) node.append_attribute(_T("user")).set_value(cur->user);
		time2str(cur->time,str);
		node.append_attribute(_T("time")).set_value(str);
	}
	cp_unlock(&lock);

	XEX_ENCRYPT_DECRYPT xex;
	u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	if (!pas_len) return;
	u32 salt[8];
	cheap_random_number(salt,8);
	xex.Init(salt,getSecretData(),passw,pas_len);
	FILE* F = _tfsopen(filename,TEXT("wb+"), _SH_DENYWR);
	if (!F) return;
	fwrite(salt,sizeof(u32),8,F);
	u64 xex_num = 0;
	xml_writer_secret_file swriter(F,&xex,&xex_num);
	doc.save(swriter,_T(""),1,pugi::encoding_utf16);
	fclose(F);
//	doc.save_file(filename);
}
//-----------------------------------------------------------------------
static void _fill_download(download_t* download, NODE* nd, pctstr_t file, pctstr_t user, cp_time_t time)
{
	if (!download) return;
	download->nd = *nd;
	if (file)
	{
		size_t len = cp_strlen(file)+1;
		download->file = new tchar_t[len];
		cp_strncpy(download->file,file,len);
	}
	else
	{
		download->file = NULL;
	}
	if (user)
	{
		size_t len = cp_strlen(user) + 1;
		download->user = new tchar_t[len];
		cp_strncpy(download->user, user, len);
	}
	else
	{
		download->user = NULL;
	}
	download->time = time;
	download->next = NULL;
}
void USERS::append_download(NODE* nd, pctstr_t file, pctstr_t user, cp_time_t time)
{
	download_t* cur;
	cp_lock(&lock);
	if (time)
	{
		if (!downloads)
		{
			downloads = new download_t;
			cur = downloads;
		}
		else
		{
			for (cur = downloads; cur->next; cur = cur->next);   // append to bottom
			cur->next = new download_t;
			cur = cur->next;
		}
		_fill_download(cur, nd, file, user, time);
	}
	else
	{
		cp_get_time(&time);
		cur = downloads;  // insert in top
		downloads = new download_t;
		_fill_download(downloads, nd, file, user, time);
		downloads->next = cur;
	}
	cp_unlock(&lock);
	if (acceptFileTransfer())
	{
		ShowWindow(g_DownloadsList,SW_NORMAL);
	}
}
//-----------------------------------------------------------------------
void USERS::remove_download(u32 num)
{
	download_t* cur = NULL;
	download_t* prev;
	u32 cnt = 0;
	cp_lock(&lock);
	if (num == 0)
	{
		if (downloads)
		{
			cur = downloads;
			downloads = cur->next;
		}
	}
	else
	{
		for (cur = prev = downloads; cur; cur = cur->next)
		{
			if (cnt == num)
			{
				prev->next = cur->next;
				break;
			}
			cnt++;
			prev = cur;
		}
	}
	if (cur)
	{
		if (cur->file) delete[] cur->file;
		if (cur->user) delete[] cur->user;
		delete cur;
	}
	cp_unlock(&lock);
}
//-----------------------------------------------------------------------
